﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

using PDFDouble = pdftron.Common.DoubleRef;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class AnnotationList : UserControl
    {
        public ViewModels.AnnotationListViewModel _ViewModel { get; private set; }
        public AnnotationList()
        {
            this.InitializeComponent();
        }

        public AnnotationList(PDFViewCtrl ctrl)
        {
            this.InitializeComponent();
            PDFViewCtrl = ctrl;
        }

        public PDFViewCtrl PDFViewCtrl
        {
            get { return _ViewModel != null ? _ViewModel.PDFViewCtrl : null; }
            set
            {
                SetPDFViewCtrl(value);
            }
        }

        private NavigationStack _NavigationStack = null;
        public NavigationStack NavigationStack
        {
            get { return _ViewModel != null ? _ViewModel.NavigationStack : _NavigationStack; }
            set
            {
                _NavigationStack = value;
                if (_ViewModel != null)
                {
                    _ViewModel.NavigationStack = _NavigationStack;
                }
            }
        }

        private void SetPDFViewCtrl(PDFViewCtrl ctrl)
        {
            if (ctrl != null)
            {
                _ViewModel = new ViewModels.AnnotationListViewModel(ctrl);
                _ViewModel.NavigationStack = _NavigationStack;
                _ViewModel.AnnotationGroups.CollectionChanged += AnnotationGroups_CollectionChanged;
                _ViewModel.PropertyChanged += ViewModel_PropertyChanged;
                AnnotationItemsListView.Loaded += AnnotationItemsListView_Loaded;
                AnnotationItemsListView.SelectionChanged += AnnotationItemsListView_SelectionChanged;
                this.DataContext = _ViewModel;
                InitAnimations();
            }
        }

        private void AnnotationItemsListView_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (AnnotationItemsListView.SelectedItem != null)
            {
                AnnotationItemsListView.ScrollIntoView(AnnotationItemsListView.SelectedItem);
            }
        }

        public event EventHandler<ItemClickEventArgs> ItemClicked;

        private void AnnotationList_ItemClick(object sender, ItemClickEventArgs e)
        {
            if (ItemClicked != null)
            {
                ItemClicked(this, e);
            }
        }


        #region Animation
        private SolidColorBrush _FlashBrush;
        public SolidColorBrush FlashBrush
        {
            get { return _FlashBrush; }
            set
            {
                _FlashBrush = value;
                if (_ViewModel != null)
                {
                    _ViewModel.FlashBrush = value;
                }
            }
        }

        private void InitAnimations()
        {
            _ViewModel.FlashAnimation = FlashTheRectangle;
            if (FlashBrush != null)
            {
                _ViewModel.FlashBrush = FlashBrush;
            }
            else
            {
                _ViewModel.FlashBrush = this.Resources["PrimaryHighlightBrighBrush"] as SolidColorBrush;
            }
        }
        #endregion Animation


        #region State

        public void RestoreState(double scrollPosition)
        {
            _CurrentRestoreProgress = RestoreProgressStates.Started;
            if (_ViewModel != null && _ViewModel.IsFinishedLoading)
            {
                ScrollRestoreOffsetIntoView(scrollPosition);
            }
            else
            {
                SetLoading();
                _RestoreScrollPosition = scrollPosition;
            }
        }
        public double GetState()
        {
            if (_RestoreScrollPosition > 0)
            {
                return _RestoreScrollPosition;
            }
            ScrollViewer sv = ListViewScroller;
            if (sv != null)
            {
                return sv.VerticalOffset;
            }

            return 0;
        }

        private bool _ViewIsVisible = true;
        /// <summary>
        /// This boolean can be set to inform the AnnotationList that it can not scroll.
        /// This happens for example when the AnnotationList is in a Pivot and not in view.
        /// Setting this to true will scroll the currently selected item into view
        /// and also affects restoring state.
        /// </summary>
        public bool ViewIsVisible
        {
            get { return _ViewIsVisible; }
            set
            {
                if (_ViewIsVisible != value)
                {
                    _ViewIsVisible = value;
                    if (_ViewIsVisible)
                    {
                        if (_ViewModel.SelectedItem != null)
                        {
                            AnnotationItemsListView.ScrollIntoView(_ViewModel.SelectedItem);
                            _CurrentRestoreProgress = RestoreProgressStates.Finished;
                        }
                        else if (_ViewModel.ToolManager != null && _ViewModel.ToolManager.SelectedAnnotation != null)
                        {
                            _ViewModel.ToolManager_SelectionChanged(_ViewModel.ToolManager.SelectedAnnotation, _ViewModel.ToolManager.SelectedAnnotationPageNumber);
                            _CurrentRestoreProgress = RestoreProgressStates.Finished;
                        }
                        else if (_CurrentRestoreProgress == RestoreProgressStates.LoadedButUnableToScroll)
                        {
                            ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
                        }
                    }
                }
            }
        }

        private enum RestoreProgressStates
        {
            None,
            Started,
            LoadedButUnableToScroll,
            Finished,
        }

        private RestoreProgressStates _CurrentRestoreProgress = RestoreProgressStates.None;

        private double _RestoreScrollPosition = -1; // negative number means don't restore

        private ScrollViewer _ListViewScroller = null;
        private ScrollViewer ListViewScroller
        {
            get
            {
                if (_ListViewScroller == null)
                {
                    _ListViewScroller = UtilityFunctions.FindVisualChild<ScrollViewer>(AnnotationItemsListView);
                }
                return _ListViewScroller;
            }
        }

        private async void ScrollRestoreOffsetIntoView(double scrollPosition)
        {
            AnnotationItemsListView.UpdateLayout();
            await this.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                ScrollViewer sv = ListViewScroller;
                if (sv != null)
                {
                    if (sv.ScrollableHeight >= scrollPosition - 1 ||
                        (_ViewModel != null && _ViewModel.IsFinishedLoading))
                    {
                        if (ViewIsVisible)
                        {
                            _CurrentRestoreProgress = RestoreProgressStates.Finished;
                            _RestoreScrollPosition = -1;
                            SetLoaded();
                        }
                        else
                        {
                            _CurrentRestoreProgress = RestoreProgressStates.LoadedButUnableToScroll;
                        }

                    }
                    sv.ChangeView(null, scrollPosition, null, true);
                }
            });
        }

        private void SetLoading()
        {
            AnnotationItemsListView.Opacity = 0;
            AnnotationItemsListView.IsHitTestVisible = false;
            LoadingProgressRing.Visibility = Visibility.Visible;
            LoadingProgressRing.IsActive = true;
        }

        private void SetLoaded()
        {
            AnnotationItemsListView.Opacity = 1;
            AnnotationItemsListView.IsHitTestVisible = true;
            LoadingProgressRing.Visibility = Visibility.Collapsed;
            LoadingProgressRing.IsActive = false;
        }


        private void ViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("IsFinishedLoading", StringComparison.OrdinalIgnoreCase))
            {
                if (_ViewModel.IsFinishedLoading)
                {
                    DelaySetLoaded();
                }
            }
        }

        private async void DelaySetLoaded()
        {
            int attempts = 5;
            while (attempts-- > 0 && _RestoreScrollPosition > 0)
            {
                if (_RestoreScrollPosition > 0)
                {
                    ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
                }
                if (attempts > 0)
                {
                    await Task.Delay(100);
                }
            }
            SetLoaded();
        }

        private void AnnotationItemsListView_Loaded(object sender, RoutedEventArgs e)
        {
            if (_RestoreScrollPosition > 0)
            {
                ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
            }
        }

        private void AnnotationGroups_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (_RestoreScrollPosition > 0)
            {
                ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
            }
        }

        public async Task WaitForAnnotationListToFinish()
        {
            if (_ViewModel != null)
            {
                _ViewModel.ClearViewModel();
                await _ViewModel.WaitForAnnotationListToFinish();
            }
        }


        #endregion State


        /// <summary>
        /// Will go back if possible.
        /// </summary>
        /// <returns>Return true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            if (_ViewModel != null)
            {
                return _ViewModel.GoBack();
            }
            return false;
        }
    }
}
