﻿using pdftron.PDF.Tools.Controls.Resources;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;

// This partial definition can be shared between Phone and Windows, allowing them to each specify their own styles and extra code behind

namespace pdftron.PDF.Tools.Controls
{
    public class BookmarkItem
    {
        public Bookmark SourceBookmark { get; set; }
        public string Title { get; set; }
        public bool HasChildren { get; set; }
        public int ChildIndex { get; set; }

        public BookmarkItem(Bookmark bookmark, int index = 0)
        {
            SourceBookmark = bookmark;
            Title = bookmark.GetTitle();
            HasChildren = bookmark.HasChildren();
            ChildIndex = index;
        }
    }

    public class AnimationCollection
    {
        public Storyboard EnterFromLeft { get; set; }
        public Storyboard LeaveToTheLeft { get; set; }
        public Storyboard EnterFromRight { get; set; }
        public Storyboard LeaveToTheRight { get; set; }

        public AnimationCollection()
        { }

        public void Refresh(double newWidth)
        {
            DoubleAnimation da = EnterFromLeft.Children[0] as DoubleAnimation;
            da.From = -newWidth;
            da = LeaveToTheLeft.Children[0] as DoubleAnimation;
            da.To = -newWidth;
            da = EnterFromRight.Children[0] as DoubleAnimation;
            da.From = newWidth;
            da = LeaveToTheRight.Children[0] as DoubleAnimation;
            da.To = newWidth;
        }
    }

    partial class Outline
    {
        private PDFViewCtrl _PDFViewCtrl;
        private NavigationStack _NavigationStack;
        private List<Grid> _BookmarkPages;
        private List<ScrollViewer> _ScrollViewerPages;
        private List<AnimationCollection> _Animations;
        private int _CurrentIndex = -1;
        private TextBlock _RootTextBlock = null;

        private List<int> _CurrentLocation = null;
        private List<int> _RestorePath = null;
        private bool _InitialOutlineLoaded = false;

        private bool _Loaded = false;
        private bool _IsConverting = false;
        private OnConversionEventHandler _OnConversionHandlder = null;

        private SemaphoreSlim OutlineReadingSemaphore { get; set; }

        private Size _CurrentSize;

        /// <summary>
        /// The PDFViewCtrl form which the outline should get the document with PDFs.
        /// </summary>
        public PDFViewCtrl PDFViewCtrl
        {
            get { return _PDFViewCtrl; }
            set
            {
                _PDFViewCtrl = value;
                if (_PDFViewCtrl != null)
                {

                    _BookmarkPages = new List<Grid>();
                    _ScrollViewerPages = new List<ScrollViewer>();
                    _Animations = new List<AnimationCollection>();
                    _CurrentLocation = new List<int>();
                    Show();
                }
            }
        }

        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        private string _DocumentTitle = string.Empty;
        public string DocumentTitle
        {
            get { return _DocumentTitle; }
            set
            {
                if (!string.Equals(value, _DocumentTitle, StringComparison.Ordinal))
                {
                    _DocumentTitle = value;
                    if (_RootTextBlock != null)
                    {
                        _RootTextBlock.Text = _DocumentTitle;
                    }
                }
            }
        }

        public delegate void PageChangedDelegate(int prevPage, int page);
        public PageChangedDelegate PageChanged;

        /// <summary>
        /// Creates a new outline (bookmark) control without any association with a PDFViewCtrl.
        /// Use the PDFViewCtrl property to associated the outline control with a document and populate it.
        /// </summary>
        public Outline()
        {
            this.InitializeComponent();
            OutlineReadingSemaphore = new SemaphoreSlim(1);
            this.SizeChanged += Outline_SizeChanged;
            this.Loaded += Outline_Loaded;
            this.Unloaded += Outline_Unloaded;
        }



        /// <summary>
        /// Creates a new outline (bookmark) control associated with a PDFViewCtrl. The control will show the bookmarks
        /// in the document that is open in the PDFViewCtrl.
        /// </summary>
        public Outline(PDFViewCtrl ctrl)
        {
            this.InitializeComponent();
            OutlineReadingSemaphore = new SemaphoreSlim(1);
            this.SizeChanged += Outline_SizeChanged;
            this.Loaded += Outline_Loaded;
            this.Unloaded += Outline_Unloaded;
            this.PDFViewCtrl = ctrl;
        }

        private void Outline_Unloaded(object sender, RoutedEventArgs e)
        {
            _Loaded = false;
            if (_OnConversionHandlder != null)
            {
                _PDFViewCtrl.OnConversionChanged -= _OnConversionHandlder;
                _OnConversionHandlder = null;
            }
        }

        private void Outline_Loaded(object sender, RoutedEventArgs e)
        {
            _Loaded = true;
            if (_IsConverting)
            {
                RegisterForConversion();
            }
        }
        
        private void RegisterForConversion()
        {
            _OnConversionHandlder = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
            _PDFViewCtrl.OnConversionChanged += _OnConversionHandlder;
        }

        private void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            if (type == PDFViewCtrlConversionType.e_conversion_finished)
            {
                Show();
            }
        }

        void Outline_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            SetSize(e.NewSize);
        }

        private void SetSize(Size size)
        {
            _CurrentSize = size;
            RectangleGeometry clipRect = new RectangleGeometry();
            clipRect.Rect = new Windows.Foundation.Rect(0, 0, size.Width, size.Height);
            ContentPanel.Clip = clipRect;

            foreach(AnimationCollection collection in _Animations)
            {
                collection.Refresh(size.Width);
            }

            if (_ScrollViewerPages != null)
            {
                foreach (ScrollViewer scroller in _ScrollViewerPages)
                {
                    FrameworkElement scrollContent = scroller.Content as FrameworkElement;
                    if (scrollContent != null)
                    {
                        scrollContent.Width = size.Width;
                    }
                }
            }

            foreach (UIElement child in ContentPanel.Children)
            {
                Grid g = child as Grid;
                g.Width = size.Width;
            }
        }

        public event EventHandler<ItemClickEventArgs> ItemClicked;

        public List<int> GetState()
        {
            return _CurrentLocation;
        }

        public async void RestoreState(List<int> list)
        {
            _RestorePath = list;
            if (_RestorePath != null && _RestorePath.Count > 0 &&
                _InitialOutlineLoaded && _ScrollViewerPages != null && _ScrollViewerPages.Count > 0)
            {
                object child = _ScrollViewerPages[0].Content;
                StackPanel childStack = child as StackPanel;
                if (childStack != null && childStack.Children.Count > _RestorePath[0])
                {
                    UIElement stackChild = childStack.Children[_RestorePath[0]];
                    _RestorePath.RemoveAt(0);
                    Grid childGrid = stackChild as Grid;
                    if (childGrid != null && childGrid.Children.Count > 0)
                    {
                        Button button = childGrid.Children[0] as Button;
                        if (button != null)
                        {
                            BookmarkItem bookmark = button.Tag as BookmarkItem;
                            if (bookmark != null && bookmark.HasChildren)
                            {
                                _BookmarkPages[0].Visibility = Visibility.Collapsed;
                                await NavigateToChildAsync(bookmark);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Resets the Bookmark display, so that next time it will create bookmarks from scratch.
        /// </summary>
        public void Reset()
        {
            _BookmarkPages.Clear();
            _ScrollViewerPages.Clear();
            _Animations.Clear();
            _CurrentIndex = -1;
            ContentPanel.Children.Clear();
        }

        public async Task WaitForOutlineToLoadAsync()
        {
            await OutlineReadingSemaphore.WaitAsync();
            OutlineReadingSemaphore.Release();
        }

        /// <summary>
        /// Animates the Bookmarks panel onto the screen.
        /// </summary>
        public async void Show()
        {
            if (_PDFViewCtrl.IsConverting)
            {
                _IsConverting = true;
                if (_Loaded)
                {
                    RegisterForConversion();
                }
                return;
            }
            IList<BookmarkItem> items;
            NoBookmarksTextBlock.Visibility = Windows.UI.Xaml.Visibility.Collapsed;
            if (ContentPanel.Children.Count < 1)
            {
                items = await GetBookmarkListAsync(null, PDFViewCtrl.GetDoc());
                if (items.Count == 0)
                {
                    NoBookmarksTextBlock.Visibility = Windows.UI.Xaml.Visibility.Visible;
                }
                else
                {
                    Grid ContentGrid = CreateBookmarkGrid(null, items);
                    _BookmarkPages.Add(ContentGrid);
                    ContentPanel.Children.Add(ContentGrid);
                    _CurrentIndex = 0;

                    _InitialOutlineLoaded = true;
                    if (_RestorePath != null && _RestorePath.Count > 0)
                    {
                        if (items.Count > _RestorePath[0])
                        {
                            BookmarkItem nextItem = items[_RestorePath[0]];
                            _RestorePath.RemoveAt(0);
                            if (nextItem.HasChildren)
                            {
                                ContentGrid.Visibility = Visibility.Collapsed;
                                await NavigateToChildAsync(nextItem);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Animates the Bookmarks panel of the screen.
        /// </summary>
        public void Hide()
        {
            //SlideOutPanel.Begin();
        }

        private void SlideOutCompleted(object sender, object e)
        {
            ContentPanel.Opacity = 0;
        }

        private IAsyncOperation<IList<BookmarkItem>> GetBookmarkListAsync(Bookmark root, PDFDoc doc)
        {
            Task<IList<BookmarkItem>> t = new Task<IList<BookmarkItem>>(() =>
            {
                return GetBookmarkList(root, doc);
            });
            t.Start();
            return t.AsAsyncOperation<IList<BookmarkItem>>();
        }

        private IList<BookmarkItem> GetBookmarkList(Bookmark root, PDFDoc doc, bool gotLock = false)
        {
            IList<BookmarkItem> items = new List<BookmarkItem>();

            try
            {
                doc.LockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore.Wait();
                }

                Bookmark current;
                if (root == null)
                {
                    current = doc.GetFirstBookmark();
                }
                else
                {
                    current = root;
                }

                int numbookmarks = 0;
                while (current.IsValid())
                {
                    BookmarkItem item = new BookmarkItem(current, numbookmarks);
                    items.Add(item);
                    current = current.GetNext();
                    numbookmarks++;
                }
                if (root == null && numbookmarks == 1)
                {
                    IList<BookmarkItem> nextLevel = GetBookmarkList(doc.GetFirstBookmark().GetFirstChild(), doc, true);
                    if (nextLevel != null && nextLevel.Count > 0)
                    {
                        return nextLevel;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
                if (!gotLock)
                {
                    OutlineReadingSemaphore.Release();
                }
            }

            return items;
        }

        private Grid CreateBookmarkGrid(Bookmark root, IList<BookmarkItem> items)
        {
            Grid hostGrid = new Grid();
            if (_CurrentSize != null && !double.IsNaN(_CurrentSize.Width) && _CurrentSize.Width > 0)
            {
                hostGrid.Width = _CurrentSize.Width;
            }

            RowDefinition topRow = new RowDefinition();
            topRow.Height = new GridLength(1, GridUnitType.Auto);
            hostGrid.RowDefinitions.Add(topRow);
            RowDefinition mainRow = new RowDefinition();
            mainRow.Height = new GridLength(1, GridUnitType.Star);
            hostGrid.RowDefinitions.Add(mainRow);
            hostGrid.RenderTransform = new TranslateTransform();

            Grid titleGrid = new Grid();
            titleGrid.Style = this.Resources["TitleGridStyle"] as Style;
            titleGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Auto) });
            titleGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
            hostGrid.Children.Add(titleGrid);

            TextBlock title = new TextBlock();
            title.Style = this.Resources["HeaderTextBlockStyle"] as Style;
            title.Tapped += Title_Tapped;
            if (root == null)
            {
                try
                {
                    _RootTextBlock = title;
                    if (string.IsNullOrWhiteSpace(_DocumentTitle))
                    {
                        _PDFViewCtrl.DocLockRead();
                        try
                        {
                            PDFDoc doc = _PDFViewCtrl.GetDoc();
                            if (doc != null)
                            {
                                string fileName = doc.GetFileName();
                                _DocumentTitle = System.IO.Path.GetFileName(doc.GetFileName());
                                if (string.IsNullOrWhiteSpace(_DocumentTitle))
                                {
                                    PDFDocInfo docInfo = doc.GetDocInfo();
                                    if (docInfo != null)
                                    {
                                        _DocumentTitle = docInfo.GetTitle();
                                    }
                                }
                            }
                        }
                        finally
                        {
                            _PDFViewCtrl.DocUnlockRead();
                        }
                    }
                    if (string.IsNullOrWhiteSpace(DocumentTitle))
                    {
                        _DocumentTitle = ResourceHandler.GetString("Outline_DefaultHeader");
                    }
                    title.Text = DocumentTitle;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            else
            {
                title.Text = root.GetTitle();
                Button backButton = new Button();
                backButton.Style = this.Resources["BackwardNavigationButtonStyle"] as Style;
                backButton.Click += backButton_Click;
                titleGrid.Children.Add(backButton);
            }
            title.SetValue(Grid.ColumnProperty, 1);
            titleGrid.Children.Add(title);

            ScrollViewer itemScroller = new ScrollViewer();
            itemScroller.Content = CreateBookmarkStack(items);
            itemScroller.HorizontalScrollBarVisibility = ScrollBarVisibility.Hidden;
            itemScroller.HorizontalScrollMode = ScrollMode.Disabled;
            itemScroller.VerticalScrollBarVisibility = ScrollBarVisibility.Auto;
            itemScroller.VerticalScrollMode = ScrollMode.Auto;
            itemScroller.ZoomMode = ZoomMode.Disabled;
            _ScrollViewerPages.Add(itemScroller);

            // Animations
            AnimationCollection animations = new AnimationCollection();
            animations.EnterFromRight = EnterFromRight(hostGrid);
            animations.EnterFromLeft = EnterFromLeft(hostGrid);
            animations.LeaveToTheRight = LeaveToTheRight(hostGrid);
            animations.LeaveToTheLeft = LeaveToTheLeft(hostGrid);
            _Animations.Add(animations);
            itemScroller.SetValue(Grid.RowProperty, 1);
            hostGrid.Children.Add(itemScroller);

            return hostGrid;

        }

        private void Title_Tapped(object sender, Windows.UI.Xaml.Input.TappedRoutedEventArgs e)
        {
            TextBlock tb = e.OriginalSource as TextBlock;
            double aw = tb.ActualWidth;
            System.Diagnostics.Debug.WriteLine("Actual width is " + aw);
        }

        private StackPanel CreateBookmarkStack(IList<BookmarkItem> items)
        {
            StackPanel bookmarkStack = new StackPanel();
            if (_CurrentSize != null && !double.IsNaN(_CurrentSize.Width) && _CurrentSize.Width > 0)
            {
                bookmarkStack.Width = _CurrentSize.Width;
            }
            bookmarkStack.Padding = new Thickness(0, 0, 8, 0);
            foreach (BookmarkItem item in items)
            {
                Grid itemGrid = new Grid();
                itemGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
                itemGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Auto) });
                itemGrid.HorizontalAlignment = Windows.UI.Xaml.HorizontalAlignment.Stretch;
                itemGrid.Height = (double)this.Resources["ToolControlsItemHeight"];
                Button itemButton = new Button();
                TextBlock title = new TextBlock();
                title.TextTrimming = TextTrimming.CharacterEllipsis;
                title.Text = item.Title;
                itemButton.Content = title;

                itemButton.Tag = item;
                itemButton.Click += itemButton_Click;
                itemGrid.Children.Add(itemButton);
                if (item.HasChildren)
                {
                    Button childButton = new Button();
                    childButton.Tag = item;
                    childButton.Style = this.Resources["ForwardNavigationButtonStyle"] as Style;
                    childButton.SetValue(Grid.ColumnProperty, 1);
                    itemGrid.Children.Add(childButton);
                    childButton.Click += childButton_Click;
                }
                bookmarkStack.Children.Add(itemGrid);
            }

            return bookmarkStack;
        }

        private void itemButton_Click(object sender, RoutedEventArgs e)
        {
            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("Bookmark", "Navigated by Outline List");

            Button button = sender as Button;
            BookmarkItem item = button.Tag as BookmarkItem;
            Bookmark bookmark = item.SourceBookmark;
            pdftron.PDF.Action action = bookmark.GetAction();
            if (action.IsValid())
            {
                ActionParameter action_param = new ActionParameter(action);
                ActionUtils.ExecuteAction(action_param, this.PDFViewCtrl, NavigationStack);
            }
            if (ItemClicked != null)
            {
                ItemClicked(this, new ItemClickEventArgs());
            }
        }

        private async void childButton_Click(object sender, RoutedEventArgs e)
        {
            Button button = sender as Button;
            BookmarkItem item = button.Tag as BookmarkItem;
            await NavigateToChildAsync(item);
        }

        private async Task NavigateToChildAsync(BookmarkItem item)
        {
            Bookmark bookmark = item.SourceBookmark;
            _CurrentLocation.Add(item.ChildIndex);
            IList<BookmarkItem> items = await GetBookmarkListAsync(bookmark.GetFirstChild(), PDFViewCtrl.GetDoc());

            Grid ContentGrid = CreateBookmarkGrid(bookmark, items);
            _BookmarkPages.Add(ContentGrid);
            ContentPanel.Children.Add(ContentGrid);
            _Animations[_CurrentIndex].LeaveToTheLeft.Begin();
            _Animations.Last().EnterFromRight.Begin();
            _CurrentIndex++;
            if (_RestorePath != null && _RestorePath.Count > 0)
            {
                if (items.Count > _RestorePath[0])
                {
                    BookmarkItem nextItem = items[_RestorePath[0]];
                    _RestorePath.RemoveAt(0);
                    if (nextItem.HasChildren)
                    {
                        ContentGrid.Visibility = Visibility.Collapsed;
                        await NavigateToChildAsync(nextItem);
                    }
                }
            }
        }

        private void backButton_Click(object sender, RoutedEventArgs e)
        {
            _BookmarkPages.RemoveAt(_CurrentIndex);
            _ScrollViewerPages.RemoveAt(_CurrentIndex);
            _ScrollViewerPages.Last().VerticalScrollBarVisibility = ScrollBarVisibility.Hidden;
            _Animations.Last().LeaveToTheRight.Begin();
            _Animations.RemoveAt(_CurrentIndex);
            _Animations.Last().EnterFromLeft.Begin();
            _CurrentIndex--;
            _BookmarkPages[_CurrentIndex].Visibility = Windows.UI.Xaml.Visibility.Visible;
            if (_CurrentLocation.Count > 0)
            {
                _CurrentLocation.RemoveAt(_CurrentLocation.Count - 1);
            }
        }

        private Storyboard EnterFromRight(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = _CurrentSize.Width;
            double to = 0;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "EnterFromRightAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Completed += EnterLeftCompleted;
            sb.Children.Add(translate);

            return sb;
        }

        private void EnterLeftCompleted(object sender, object e)
        {
        }

        private Storyboard EnterFromLeft(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = -_CurrentSize.Width;
            double to = 0;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "EnterFromLeftAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Completed += EnterRightCompleted;
            sb.Children.Add(translate);

            return sb;
        }

        private void EnterRightCompleted(object sender, object e)
        {
            _ScrollViewerPages.Last().VerticalScrollBarVisibility = ScrollBarVisibility.Auto;
        }

        private Storyboard LeaveToTheRight(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = 0;
            double to = _CurrentSize.Width;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "LeaveToTheRightAnimation", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Children.Add(translate);
            sb.Completed += LeaveRightCompleted;

            return sb;
        }

        private void LeaveRightCompleted(object sender, object e)
        {
            if (ContentPanel.Children.Count > 0)
            {
                ContentPanel.Children.RemoveAt(ContentPanel.Children.Count - 1);
            }
        }

        private Storyboard LeaveToTheLeft(FrameworkElement target)
        {
            Storyboard sb = new Storyboard();
            String ID = target.Name.ToString();
            double from = 0;
            double to = -_CurrentSize.Width;
            DoubleAnimation translate = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(300)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseOut },
            };

            target.Resources.Add(ID + "LeaveToTheLeft", sb);
            Storyboard.SetTarget(translate, target);
            Storyboard.SetTargetName(translate, target.Name);
            Storyboard.SetTargetProperty(translate, "(UIElement.RenderTransform).(TranslateTransform.X)");
            sb.Children.Add(translate);
            sb.Completed += LeaveLeftCompleted;

            return sb;
        }

        private void LeaveLeftCompleted(object sender, object e)
        {
            if (_BookmarkPages.Count > 1)
            {
                _BookmarkPages[_BookmarkPages.Count - 2].Visibility = Windows.UI.Xaml.Visibility.Collapsed;
            }
        }

        /// <summary>
        /// Will go back if possible.
        /// </summary>
        /// <returns>Return true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            return false;
        }
    }
}
