﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml.Data;

namespace pdftron.PDF.Tools.Controls.Resources
{
    /// <summary>
    /// This class derives from ObservableCollection, and adds tracking of which items are currently visible,
    /// and also provides convenience for requesting and tracking a per-item resource that needs to be requested asynchronously.
    /// 
    /// Usage:
    /// Register for ResourceNeededForIndex and DiscardResourceForIndex event, which tells when a resource should be requested
    /// and when one needs to be discarded.
    /// SortedResourceDictionary needs to mirror the items that have resources. Use SortedResourceDictionary.TryAdd before adding a resource, and only add it if TryAdd returns true.
    /// Use MaxAllowedSize to limit the resources that will be stored and requested.
    /// </summary>
    /// <typeparam name="T">The type stored in the list, for binding to a view model.</typeparam>
    /// /// <typeparam name="U">Any type that implements IResourceHolder, used for tracking the size of the resource.</typeparam>
    public class ObservableCollectionWithAsyncResource<T, U> : ObservableCollection<T>, IItemsRangeInfo where U : ObservableCollectionWithAsyncResource<T, U>.IResourceHolder
    {
        public interface IResourceHolder
        {
            double TotalSize { get; }
        };

        public class SortedResourceDictionary : IEnumerable, IReadOnlyDictionary<int, U>
        {
            internal delegate bool HostSizeCheckDelegate(int index, U value);

            private HostSizeCheckDelegate _HostSizeCheck;
            private SortedDictionary<int, U> _InternalDict;
            internal SortedResourceDictionary(HostSizeCheckDelegate hostSizeCheck)
            {
                _HostSizeCheck = hostSizeCheck;
                _InternalDict = new SortedDictionary<int, U>();
            }

            public U this[int key]
            {
                get { return _InternalDict[key]; }
            }

            public int Count { get { return _InternalDict.Count; } }

            public double TotalSize
            {
                get
                {
                    double totalSize = 0;
                    foreach (U info in this.Values)
                    {
                        totalSize += (info.TotalSize);
                    }
                    return totalSize;
                }
            }

            public ICollection<int> Keys { get { return _InternalDict.Keys; } }

            public ICollection<U> Values { get { return _InternalDict.Values; } }

            IEnumerable<int> IReadOnlyDictionary<int, U>.Keys { get { return _InternalDict.Keys; } }

            IEnumerable<U> IReadOnlyDictionary<int, U>.Values { get { return _InternalDict.Values; } }

            public bool TryAdd(KeyValuePair<int, U> item)
            {
                return TryAdd(item.Key, item.Value);
            }

            public bool TryAdd(int key, U value)
            {
                if (_HostSizeCheck(key, value))
                {
                    _InternalDict[key] = value;
                    return true;
                }
                return false;
            }

            public void Clear()
            {
                _InternalDict.Clear();
            }

            public bool Contains(KeyValuePair<int, U> item)
            {
                return _InternalDict.Contains(item);
            }

            public bool ContainsKey(int key)
            {
                return _InternalDict.ContainsKey(key);
            }

            public IEnumerator<KeyValuePair<int, U>> GetEnumerator()
            {
                return _InternalDict.GetEnumerator();
            }

            public bool Remove(KeyValuePair<int, U> item)
            {
                return _InternalDict.Remove(item.Key);
            }

            public bool Remove(int key)
            {
                return _InternalDict.Remove(key);
            }

            public bool TryGetValue(int key, out U value)
            {
                return _InternalDict.TryGetValue(key, out value);
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return _InternalDict.GetEnumerator();
            }
        }


        public ObservableCollectionWithAsyncResource() : base()
        {
            _ItemsWithResource = new SortedResourceDictionary(HasRoomToAddItem);
        }

        private int _FirstIndex = 0;

        private int _LastIndex = 0;

        private ItemIndexRange _PreviousRange = new ItemIndexRange(0, 1);

        private bool _FirstRequest = true;

        public void Dispose() { } // We maintain no state, so this can be empty

        /// <summary>
        /// Required by the system, from IItemsRangeInfo. Not for other use.
        /// </summary>
        /// <param name="visibleRange"></param>
        /// <param name="trackedItems"></param>
        public void RangesChanged(ItemIndexRange visibleRange, IReadOnlyList<ItemIndexRange> trackedItems)
        {
            if (!_FirstRequest && _PreviousRange.FirstIndex == visibleRange.FirstIndex && _PreviousRange.LastIndex == visibleRange.LastIndex)
            {
                return;
            }
            _FirstRequest = false;
            _FirstIndex = visibleRange.FirstIndex;
            _LastIndex = visibleRange.LastIndex;
            if (visibleRange.FirstIndex < _PreviousRange.FirstIndex || visibleRange.LastIndex < _PreviousRange.LastIndex)
            {
                Direction = Directions.Backwards;
            }
            else
            {
                Direction = Directions.Forwards;
            }
            VisibleRangeChanged?.Invoke(this, visibleRange);
            OnVisibleRangechanged();
            _PreviousRange = visibleRange;
        }

        private void ResetStoredValues()
        {
            _FirstIndex = 0;
            _LastIndex = 0;
            _PreviousRange = new ItemIndexRange(0, 1);
        }

        protected override void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            if (Items.Count == 0)
            {
                ResetStoredValues();
            }
            base.OnCollectionChanged(e);
        }

        #region Public API

        /// <summary>
        /// This enum specifies if request are done using FIFO or FILO
        /// </summary>
        public enum RequestBehaviours
        {
            FIFO = 0,
            FILO,
        }

        private RequestBehaviours _RequestBehaviour = RequestBehaviours.FIFO;
        /// <summary>
        /// Set this to indicate the desired ordering of request.
        /// </summary>
        public RequestBehaviours RequestBehaviour
        {
            get { return _RequestBehaviour; }
            set { _RequestBehaviour = value; }
        }

        /// <summary>
        /// Indicates whether the view is scrolling forwards or backwards.
        /// </summary>
        public enum Directions
        {
            Forwards = 0,
            Backwards,
        }
        private Directions _Direction = Directions.Forwards;
        /// <summary>
        /// Gets whether the ObservableCollectionWithAsyncResource is currently scrolling forwards or backwards.
        /// </summary>
        public Directions Direction
        {
            get { return _Direction; }
            private set
            {
                if (_Direction != value)
                {
                    _Direction = value;
                }
            }
        }

        private int MaximumConcurrentRequests = 2;

        /// <summary>
        /// Gets or sets whether items that have no resource should be re-requested when the visible range changes.
        /// </summary>
        public bool ResetRequestsWhenViewChanges
        {
            get { return _ResetRequestsWhenViewChanges; }
            set { _ResetRequestsWhenViewChanges = value; }
        }

        /// <summary>
        /// Gets or sets whether the list should pause it's work
        /// </summary>
        public bool Pause { get; set; }

        /// <summary>
        /// List of items that have their resource.
        /// </summary>
        public SortedResourceDictionary ItemsWithResource { get { return _ItemsWithResource; } }

        /// <summary>
        /// Total size that all added IResourceHolder are allowed to be
        /// </summary>
        public double MaxAllowedSize
        {
            get { return _MaxAllowedSize; }
            set { _MaxAllowedSize = value; }
        }

        /// <summary>
        /// Limits the number of items with a resource to those which are near the screen.
        /// </summary>
        public bool LimitResourcesToItemsNearScreen
        {
            get { return _LimitResourcesToItemsNearScreen; }
            set { _LimitResourcesToItemsNearScreen = value; }
        }

        public int FirstVisibleIndex { get { return _FirstIndex; } }

        public int LastVisibleIndex { get { return Math.Min(_LastIndex, Count - 1); } }

        public ItemIndexRange PreviousRange { get { return _PreviousRange; } }

        public event EventHandler<ItemIndexRange> VisibleRangeChanged;

        public event EventHandler<object> AllRequestsCancelled;

        /// <summary>
        /// Raised when an item is getting close to or into the view that has no resource.
        /// </summary>
        public event EventHandler<int> ResourceNeededForIndex;

        /// <summary>
        /// Raised when an item far from the view needs to discard it's resource to make room for an item closer to the view.
        /// </summary>
        public event EventHandler<int> DiscardResourceForIndex;

        /// <summary>
        /// Requests Resources
        /// </summary>
        public void RequestResources()
        {
            if (Pause)
            {
                return;
            }
            if ((RequestBehaviour == RequestBehaviours.FIFO && Direction == Directions.Forwards)
                || (RequestBehaviour == RequestBehaviours.FILO && Direction == Directions.Backwards))
            {
                for (int i = FirstVisibleIndex; i <= LastVisibleIndex && _RequestedIndexesSinceViewChanged.Count < MaximumConcurrentRequests; ++i)
                {
                    RequestResourceIfNeeded(i);
                }
            }
            else
            {
                for (int i = LastVisibleIndex; i >= FirstVisibleIndex && _RequestedIndexesSinceViewChanged.Count < MaximumConcurrentRequests; --i)
                {
                    RequestResourceIfNeeded(i);
                }
            }

            RequestResourcesOutOfRange();
        }

        /// <summary>
        /// Clears all resources and re-requests resources for the current view.
        /// </summary>
        public void Refresh()
        {
            ItemsWithResource.Clear();
            _RequestedIndexesSinceViewChanged.Clear();
            RequestResources();
        }

        /// <summary>
        /// Refreshes a particualr reseource, for example if it has been updated.
        /// </summary>
        /// <param name="index"></param>
        public void RefreshResource(int index)
        {
            if (ItemsWithResource.ContainsKey(index))
            {
                ItemsWithResource.Remove(index);
            }
            if (_RequestedIndexesSinceViewChanged.Contains(index))
            {
                _RequestedIndexesSinceViewChanged.Remove(index);
            }
            RequestResources();
        }

        /// <summary>
        /// Use this to report that a request for resources has failed on your end.
        /// </summary>
        /// <param name="index"></param>
        public void ReportRequestFailed(int index)
        {
            if (_RequestedIndexesSinceViewChanged.Contains(index))
            {
                _RequestedIndexesSinceViewChanged.Remove(index);
            }
        }

        #endregion Public API


        #region PreviewRanges

        private bool _ResetRequestsWhenViewChanges = true;
        private double _MaxAllowedSize = 0;
        private bool _LimitResourcesToItemsNearScreen = false;
        private ISet<int> _RequestedIndexesSinceViewChanged = new HashSet<int>();
        private SortedResourceDictionary _ItemsWithResource;
        private bool HasRoomToAddItem(int index, U value)
        {
            if (_RequestedIndexesSinceViewChanged.Contains(index))
            {
                _RequestedIndexesSinceViewChanged.Remove(index);
            }
            bool hasRoom = true;
            if (_MaxAllowedSize > 0)
            {
                while (hasRoom && _MaxAllowedSize < _ItemsWithResource.TotalSize)
                {
                    int furthestIndex = GetIndexFurthestFromView();
                    if (furthestIndex == -1)
                    {
                        return false;
                    }
                    int furthestDistance = GetDistanceFromView(furthestIndex);
                    int thisDistance = GetDistanceFromView(index);
                    if (thisDistance < furthestDistance)
                    {
                        DiscardResourceForIndex?.Invoke(this, furthestIndex);
                        ItemsWithResource.Remove(furthestIndex);
                    }
                    else
                    {
                        hasRoom = false;
                    }
                }
            }
            return hasRoom;
        }

        private void OnVisibleRangechanged()
        {
            int oldStart = PreviousRange.FirstIndex;
            int oldEnd = PreviousRange.LastIndex;
            int newStart = FirstVisibleIndex;
            int newEnd = LastVisibleIndex;
            if (ResetRequestsWhenViewChanges && (oldStart != newStart || oldEnd != newEnd))
            {
                _RequestedIndexesSinceViewChanged.Clear();
                AllRequestsCancelled?.Invoke(this, null);
                if (LimitResourcesToItemsNearScreen)
                {
                    List<Tuple<int, int, int, int>> previewRanges = GetPreviewRanges(4);
                    if (previewRanges.Count > 0)
                    {
                        int lowest = previewRanges[previewRanges.Count - 1].Item2;
                        int highest = previewRanges[previewRanges.Count - 1].Item4;
                        RemoveItemsOutsideRange(lowest, highest);
                    }
                }
            }
            RequestResourcesInRange(newStart, newEnd, oldStart, oldEnd);
        }

        private void RequestResourcesInRange(int firstVisibleIndex, int lastVisibleIndex, int previousFirstIndex, int previousLastIndex)
        {
            if (Pause)
            {
                return;
            }
            if (ResetRequestsWhenViewChanges ||
                previousFirstIndex - previousLastIndex == 0 ||
                firstVisibleIndex > previousLastIndex ||
                lastVisibleIndex < previousFirstIndex)
            {
                RequestResources();
                return;
            }

            if ((RequestBehaviour == RequestBehaviours.FIFO && Direction == Directions.Forwards)
                || (RequestBehaviour == RequestBehaviours.FILO && Direction == Directions.Backwards))
            {
                for (int i = previousFirstIndex - 1; i >= firstVisibleIndex; --i)
                {
                    RequestResourceIfNeeded(i);
                }
                for (int i = previousLastIndex + 1; i <= lastVisibleIndex; ++i)
                {
                    RequestResourceIfNeeded(i);
                }
            }
            else
            {
                for (int i = firstVisibleIndex; i <= previousFirstIndex - 1; ++i)
                {
                    RequestResourceIfNeeded(i);
                }
                for (int i = lastVisibleIndex; i >= previousLastIndex + 1; --i)
                {
                    RequestResourceIfNeeded(i);
                }
            }

            RequestResourcesOutOfRange();
        }

        private void RequestResourcesOutOfRange()
        {
            if (Pause)
            {
                return;
            }
            for (int i = FirstVisibleIndex; i <= LastVisibleIndex; ++i)
            {
                if (!_ItemsWithResource.ContainsKey(i) || _RequestedIndexesSinceViewChanged.Contains(i))
                {
                    return;
                }
            }

            List<Tuple<int, int, int, int>> previewRanges = GetPreviewRanges();

            foreach (Tuple<int, int, int, int> range in previewRanges)
            {
                bool hasRequested = false;
                for (int i = range.Item2; i <= range.Item1 && _RequestedIndexesSinceViewChanged.Count < MaximumConcurrentRequests; ++i)
                {
                    hasRequested |= RequestResourceIfNeeded(i);
                }
                for (int i = range.Item4; i >= range.Item3 && _RequestedIndexesSinceViewChanged.Count < MaximumConcurrentRequests; --i)
                {
                    hasRequested |= RequestResourceIfNeeded(i);
                }

                if (hasRequested)
                {
                    return;
                }
            }
        }

        /// <summary>
        /// Returns ranges for requesting resources for items out of range. Each range is a tuple 
        /// as follows (previous lowest - 1) (new lowest) (previous highest + 1) (new highest)
        /// Note that item 1 and 3 of each range (tuples are 1-indexed) may be out of range, but items
        /// 2 and 4 are limited to the 
        /// </summary>
        /// <returns></returns>
        private List<Tuple<int, int, int, int>> GetPreviewRanges(int total_ranges = 3)
        {
            List<Tuple<int, int, int, int>> ranges = new List<Tuple<int, int, int, int>>();
            int itemsOnScreen = LastVisibleIndex - FirstVisibleIndex;
            if (itemsOnScreen >= 0)
            {
                int stride = Math.Max(itemsOnScreen / 2, 3);
                int lowBound = FirstVisibleIndex - 1;
                int highBound = LastVisibleIndex + 1;
                for (int rounds = 0; rounds < total_ranges; ++rounds)
                {
                    int lowIndex = Math.Max(0, lowBound - stride);
                    int highIndex = Math.Min(highBound + stride, this.Count - 1);
                    ranges.Add(new Tuple<int, int, int, int>(lowBound, lowIndex, highBound, highIndex));
                    lowBound = lowBound - stride - 1;
                    highBound = highBound + stride + 1;
                }
            }

            return ranges;
        }

        private bool RequestResourceIfNeeded(int index)
        {
            if (!_ItemsWithResource.ContainsKey(index) && index < this.Count)
            {
                if (!_RequestedIndexesSinceViewChanged.Contains(index))
                {
                    try
                    {
                        ResourceNeededForIndex?.Invoke(this, index);
                        _RequestedIndexesSinceViewChanged.Add(index);
                    }
                    catch (Exception) { }
                }
                return true;
            }
            return false;
        }

        private int GetIndexFurthestFromView()
        {
            if (FirstVisibleIndex - LastVisibleIndex == 0 ||
                _ItemsWithResource.Count == 0)
            {
                return -1;
            }

            KeyValuePair<int, U> first = _ItemsWithResource.First();
            KeyValuePair<int, U> last = _ItemsWithResource.Last();
            int firstDist = GetDistanceFromView(first.Key - 1);
            int lastDist = GetDistanceFromView(last.Key - 1);
            if (lastDist > firstDist)
            {
                return last.Key;
            }
            return first.Key;
        }

        private int GetDistanceFromView(int index)
        {
            if (index >= FirstVisibleIndex && index <= LastVisibleIndex)
            {
                return 0;
            }
            int firstDist = Math.Abs(index - FirstVisibleIndex);
            int lastDist = Math.Abs(index - LastVisibleIndex);
            return Math.Min(firstDist, lastDist);
        }

        private void RemoveItemsOutsideRange(int lowest, int highest)
        {
            while (ItemsWithResource.Count > 0 && ItemsWithResource.First().Key < lowest)
            {
                DiscardResourceForIndex?.Invoke(this, ItemsWithResource.First().Key);
                ItemsWithResource.Remove(ItemsWithResource.First().Key);
            }
            while (ItemsWithResource.Count > 0 && ItemsWithResource.Last().Key > highest)
            {
                DiscardResourceForIndex?.Invoke(this, ItemsWithResource.Last().Key);
                ItemsWithResource.Remove(ItemsWithResource.Last().Key);
            }
        }

        #endregion PreviewRanges
    }
}
