﻿using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Graphics.Display;
using Windows.UI;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Navigation;


namespace pdftron.PDF.Tools.Controls
{
    /// <summary>
    /// Control that displays thumbnails in a grid pattern, and allows rearranging, adding, and removing of pages.
    /// </summary>
    /// <remarks>
    /// This control seem to have problems when a debugger is a attached in terms of memory usage
    /// as if garbage collection doesn't work. If you find it running up the memory, try it without a debugger attached.</remarks>
    public sealed partial class ThumbnailViewer : UserControl, pdftron.PDF.Tools.Controls.ControlBase.ICloseableControl
    {

        #region Dependency Properties

        static DependencyProperty NavigationOnlyProperty =
          DependencyProperty.Register("NavigationOnly", typeof(bool),
          typeof(ThumbnailViewer), new PropertyMetadata(false, OnNavigationOnlyChanged));

        private static void OnNavigationOnlyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ThumbnailViewer viewer = d as ThumbnailViewer;
            if (viewer != null)
            {
                viewer.UpdateNavigationOnly();
            }
        }

        public bool NavigationOnly
        {
            get { return (bool)GetValue(NavigationOnlyProperty); }
            set { SetValue(NavigationOnlyProperty, value); }
        }

        /// <summary>
        /// Dependency property to set whether Items should be automatically fitted to the width of the ThumbnailViewer,
        /// accoriding to NumberOfColumns
        /// </summary>
        static DependencyProperty FitItemsToWidthProperty =
          DependencyProperty.Register("FitItemsToWidth", typeof(bool),
          typeof(AnnotationList), new PropertyMetadata(false, OnFitItemsToWidthChanged)); // TODO UPDATE

        /// <summary>
        /// Gets or sets whether Items should be automatically fitted to the width of the ThumbnailViewer,
        /// according to NumberOfColumns
        /// </summary>
        public bool FitItemsToWidth
        {
            get { return (bool)(GetValue(FitItemsToWidthProperty)); }
            set { SetValue(FitItemsToWidthProperty, value); }
        }
        private static void OnFitItemsToWidthChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ThumbnailViewer viewer = d as ThumbnailViewer;
            if (viewer != null && e.NewValue is bool)
            {
                viewer._LastWidthCalculated = 0;
                if (viewer.ActualWidth > 0)
                {
                    viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
                }
            }
        }

        /// <summary>
        /// Dependency property to set the number of columns of thumbnails in the ThumbnailsView
        /// </summary>
        static DependencyProperty NumberOfColumnsProperty =
         DependencyProperty.Register("NumberOfColumns", typeof(int),
         typeof(ThumbnailViewer), new PropertyMetadata(null, OnNumberOfColumnsChanged));

        private static void OnNumberOfColumnsChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ThumbnailViewer viewer = d as ThumbnailViewer;
            if (viewer != null)
            {
                if (!double.IsNaN(viewer.MainListView.ActualWidth))
                {
                    viewer._LastWidthCalculated = 0;
                    viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
                }
            }
        }

        /// <summary>
        /// Gets or sets the number of columns of thumbnails in the ThumbnailsView
        /// Null or any number less than 1 means that the View will calculate an appropriate number of
        /// columns based on its size
        /// </summary>
        public int NumberOfColumns
        {
            get { return (int)GetValue(NumberOfColumnsProperty); }
            set { SetValue(NumberOfColumnsProperty, value); }
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        static DependencyProperty ListViewItemWidthProperty =
          DependencyProperty.Register("ListViewItemWidth", typeof(double),
          typeof(AnnotationList), new PropertyMetadata(10.0));

        public double ListViewItemWidth
        {
            get { return (double)(GetValue(ListViewItemWidthProperty)); }
            set { SetValue(ListViewItemWidthProperty, value); }
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        static DependencyProperty ListViewItemHeightProperty =
          DependencyProperty.Register("ListViewItemHeight", typeof(double),
          typeof(AnnotationList), new PropertyMetadata(10.0));

        public double ListViewItemHeight
        {
            get { return (double)(GetValue(ListViewItemHeightProperty)); }
            set { SetValue(ListViewItemHeightProperty, value); }
        }

        /// <summary>
        /// This property is used only by the control itself, when it dynamically calculates the size of elements
        /// to fit the width of the ListView
        /// </summary>
        static DependencyProperty ItemPaddingProperty =
          DependencyProperty.Register("ItemPadding", typeof(Thickness),
          typeof(AnnotationList), new PropertyMetadata(new Thickness(8), OnItemPaddingChanged));

        /// <summary>
        /// Sets the padding between items in the ThumbnailsView
        /// </summary>
        public Thickness ItemPadding
        {
            get { return (Thickness)(GetValue(ItemPaddingProperty)); }
            set { SetValue(ItemPaddingProperty, value); }
        }
        private static void OnItemPaddingChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ThumbnailViewer viewer = d as ThumbnailViewer;
            if (viewer != null && e.NewValue is Thickness)
            {
                viewer._LastWidthCalculated = 0;
                viewer.SetItemPadding();
                if (!double.IsNaN(viewer.MainListView.ActualWidth))
                {
                    viewer._LastWidthCalculated = 0;
                    viewer.SetItemTemplateAndSize(new Size(viewer.ActualWidth, viewer.ActualHeight));
                }
            }
        }

        private void SetItemPadding()
        {
            Thickness padding = (Thickness)GetValue(ItemPaddingProperty);
            Style itemContainerStyle = Resources["ListViewItemWihtoutExtraPadding"] as Style;
            Setter marginSetter = null;
            foreach (Setter setter in itemContainerStyle.Setters)
            {
                if (setter.Property == MarginProperty)
                {
                    marginSetter = setter;
                    break;
                }
            }
            if (marginSetter != null)
            {
                itemContainerStyle.Setters.Remove(marginSetter);
            }
            marginSetter = new Setter(MarginProperty, padding);
            itemContainerStyle.Setters.Add(marginSetter);
            MainListView.ItemContainerStyle = itemContainerStyle;
        }

        /// <summary>
        /// Dependency property for max width of items when automatically sizing items.
        /// This property only has an effect if NumberOfColumns is NaN or 0
        /// </summary>
        static DependencyProperty AutoSizingMaxWidthProperty =
          DependencyProperty.Register("AutoSizingMaxWidth", typeof(double),
          typeof(AnnotationList), new PropertyMetadata(250.0, OnAutoSizingMaxWidthChanged));

        /// <summary>
        /// The max width of items when automatically sizing items.
        /// This property only has an effect if NumberOfColumns is NaN or 0
        /// </summary>
        public double AutoSizingMaxWidth
        {
            get { return (double)(GetValue(AutoSizingMaxWidthProperty)); }
            set { SetValue(AutoSizingMaxWidthProperty, value); }
        }
        private static void OnAutoSizingMaxWidthChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ThumbnailViewer viewer = d as ThumbnailViewer;
            if (viewer != null && e.NewValue is double)
            {
                if (!double.IsNaN(viewer.MainListView.ActualWidth))
                {
                    viewer._LastWidthCalculated = 0;
                    viewer.SetItemTemplateAndSize(new Size(viewer.MainListView.ActualWidth, viewer.MainListView.ActualHeight));
                }
            }
        }

        #endregion Dependency Properties

        private ViewModels.ThumbnailsViewViewModel _ViewModel;
        public ViewModels.ThumbnailsViewViewModel ViewModel { get { return _ViewModel; } }

        private PDFViewCtrl _PDFViewCtrl;
        public PDFViewCtrl PDFViewCtrl
        {
            private get { return _PDFViewCtrl; }
            set
            {
                if (!string.IsNullOrWhiteSpace(_DocumentTag) && _PDFViewCtrl != null)
                {
                    _ViewModel.CleanUp();
                }

                _PDFViewCtrl = value;
                if (!string.IsNullOrWhiteSpace(_DocumentTag) && _PDFViewCtrl != null)
                {
                    SetPDFViewCtrl(_PDFViewCtrl);
                }
            }
        }

        private ToolManager _ToolManager;

        public ToolManager ToolManager
        {
            private get { return _ToolManager; }
            set
            {
                _ToolManager = value;
                if (_ViewModel != null)
                {                    
                    _ViewModel.ToolManager = _ToolManager;
                }
            }
        }

        private string _DocumentTag;
        public string DocumentTag
        {
            get { return _DocumentTag; }
            set
            {
                _DocumentTag = value;
                if (!string.IsNullOrWhiteSpace(_DocumentTag) && _PDFViewCtrl != null)
                {
                    SetPDFViewCtrl(_PDFViewCtrl);
                }
            }
        }

        public Color BlankPageDefaultColor
        {
            get
            {
                return _ViewModel.PageDefaultColor;
            }
            set
            {
                if (value != _ViewModel.PageDefaultColor)
                {
                    _ViewModel.PageDefaultColor = value;
                    if (ViewIsVisible)
                    {
                        _ViewModel.Refresh();
                    }
                    else
                    {
                        _ShouldRefreshWhenComingIntoView = true;
                    }
                }
            }
        }

        public new Brush Background
        {
            get { return BackgroundGrid.Background; }
            set { BackgroundGrid.Background = value; }
        }

        private NavigationStack _NavigationStack;
        public NavigationStack NavigationStack
        {
            get { return _ViewModel != null ? _ViewModel.NavigationStack : _NavigationStack; }
            set
            {
                _NavigationStack = value;
                if (_ViewModel != null)
                {
                    _ViewModel.NavigationStack = _NavigationStack;
                }
            }
        }

        private bool _IsActive = true;
        private bool IsActive
        {
            get { return _IsActive; }
            set
            {
                if (_IsActive != value)
                {
                    _IsActive = value;
                }
            }
        }

        private ThumbnailItem _CurrentItemWhenGoingOutOfView = null;
        private ISet<int> _RequestedIndexesSinceViewChanged = new HashSet<int>();
        private bool _ShouldRefreshWhenComingIntoView = false;

        private bool _ViewIsVisible = true;
        /// <summary>
        /// This boolean can be set to inform the AnnotationList that it can not scroll.
        /// This happens for example when the AnnotationList is in a Pivot and not in view.
        /// Setting this to true will scroll the currently selected item into view
        /// and also affects restoring state.
        /// </summary>
        public bool ViewIsVisible
        {
            get { return _ViewIsVisible; }
            set
            {
                if (_ViewIsVisible != value)
                {
                    _ViewIsVisible = value;
                    if (ViewModel != null)
                    {
                        ViewModel.Pause = !_ViewIsVisible;
                        if (_ViewIsVisible)
                        {
                            if (ViewModel.CurrentItem != null && ViewModel.CurrentItem != _CurrentItemWhenGoingOutOfView)
                            {
                                MainListView.ScrollIntoView(ViewModel.CurrentItem);
                            }
                            if (_ShouldRefreshWhenComingIntoView)
                            {
                                ViewModel.Refresh();
                                _ShouldRefreshWhenComingIntoView = false;
                                _RequestedIndexesSinceViewChanged.Clear();
                            }
                            foreach (int modifiedPage in _RequestedIndexesSinceViewChanged)
                            {
                                ViewModel.PageModified(modifiedPage);
                            }
                            _RequestedIndexesSinceViewChanged.Clear();
                        }
                        else
                        {
                            _CurrentItemWhenGoingOutOfView = ViewModel.CurrentItem;
                        }
                    }
                }
            }
        }

        public ThumbnailViewer()
        {
            this.InitializeComponent();
            Init();
        }

        public ThumbnailViewer(PDFViewCtrl ctrl, string docTag)
        {
            this.InitializeComponent();
            Init();

            PDFViewCtrl = ctrl;
            DocumentTag = docTag;
            _ViewModel.PageAdded += _ViewModel_PageAdded;
        }

        void DebugButton_Click(object sender, RoutedEventArgs e)
        {
            ControlClosed();
        }

        public ThumbnailViewer(PDFViewCtrl ctrl, string docTag, Color background)
        {
            this.InitializeComponent();
            Init();

            Background = new SolidColorBrush(background);
            PDFViewCtrl = ctrl;
            DocumentTag = docTag;
        }

        private void Init()
        {
            SetItemPadding();
            this.SizeChanged += ThumbnailViewer_SizeChanged;
            MainListView.SizeChanged += MainListView_SizeChanged;
            MainListView.Loaded += MainListView_Loaded;
        }

        private void MainListView_Loaded(object sender, RoutedEventArgs e)
        {
            if (ViewIsVisible)
            {
                if (ViewModel != null && ViewModel.CurrentItem != null)
                {
                    MainListView.ScrollIntoView(ViewModel.CurrentItem);
                }
            }
            else
            {
                _CurrentItemWhenGoingOutOfView = null;
            }
        }

        private void ThumbnailViewer_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            // Show/hide the excess Command Bar elements depending on screen width
            if (e.NewSize.Width < 450)
            {
                UtilityFunctions.HideElement(CancelSelectionButton);
                UtilityFunctions.HideElement(CancelSelectionButtonSeparator);
            }
            else
            {
                UtilityFunctions.ShowElement(CancelSelectionButton, CancelSelectionButton.Height);
                UtilityFunctions.ShowElement(CancelSelectionButtonSeparator, CancelSelectionButtonSeparator.Height);
            }
        }

        private async void _ViewModel_PageAdded(int pageNumber, ViewModels.ThumbnailItem item)
        {
            int pageCount = PDFViewCtrl.GetPageCount();
            await System.Threading.Tasks.Task.Delay(200); // Hack to ensure the ListView's layout is updated.
            if (item != null && ViewModel != null && ViewModel.ThumbnailsList != null && ViewModel.ThumbnailsList.Contains(item) && MainListView != null)
            {
                
                MainListView.ScrollIntoView(item, ScrollIntoViewAlignment.Default);
            }
        }

        void MainListView_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            SetItemTemplateAndSize(e.NewSize);
        }

        private enum ItemTemplates
        {
            None,
            Small,
            Medium,
            Large,
            Variable,
        }
        private ItemTemplates _LastItemTemplate = ItemTemplates.None;

        private void SetItemTemplateAndSize(Size listSize)
        {
            if (FitItemsToWidth)
            {
                CalculateItemWidth(listSize.Width);
            }
            else if (listSize.Width < 500 || listSize.Height < 500)
            {
                if (_LastItemTemplate != ItemTemplates.Small)
                {
                    MainListView.ItemTemplate = (DataTemplate)this.Resources["ThumbnailTemplateSmall"];
                    _LastItemTemplate = ItemTemplates.Small;
                }
                double w = (double)this.Resources["ThumbnailItemWidthSmall"];
                double h = (double)this.Resources["ThumbnailItemHeightSmall"];
                Thickness t = (Thickness)this.Resources["ThumbnailMarginSmall"];
                if (_ViewModel != null)
                {
                    _ViewModel.ImageMaxDimensions = new Windows.Foundation.Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
                }
            }
            else
            {
                double w = (double)this.Resources["ThumbnailItemWidthMedium"];
                double h = (double)this.Resources["ThumbnailItemWidthMedium"];
                double rows = listSize.Height / w;
                double columns = listSize.Width / h;

                if (rows * columns <= 15)
                {
                    if (_LastItemTemplate != ItemTemplates.Medium)
                    {
                        MainListView.ItemTemplate = (DataTemplate)this.Resources["ThumbnailTemplateMedium"];
                        _LastItemTemplate = ItemTemplates.Medium;
                    }
                    Thickness t = (Thickness)this.Resources["ThumbnailMarginMedium"];
                    if (_ViewModel != null)
                    {
                        _ViewModel.ImageMaxDimensions = new Windows.Foundation.Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
                    }
                }
                else
                {
                    if (_LastItemTemplate != ItemTemplates.Large)
                    {
                        MainListView.ItemTemplate = (DataTemplate)this.Resources["ThumbnailTemplateLarge"];
                        _LastItemTemplate = ItemTemplates.Large;
                    }
                    w = (double)this.Resources["ThumbnailItemWidthLarge"];
                    h = (double)this.Resources["ThumbnailItemHeightLarge"];
                    Thickness t = (Thickness)this.Resources["ThumbnailMarginLarge"];
                    if (_ViewModel != null)
                    {
                        _ViewModel.ImageMaxDimensions = new Windows.Foundation.Size(w - t.Left - t.Right, h - t.Top - t.Bottom);
                    }
                }
            }
        }

        private void SetPDFViewCtrl(PDFViewCtrl ctrl)
        {
            _ViewModel = new ViewModels.ThumbnailsViewViewModel(ctrl, _DocumentTag, this);
            _ViewModel.ReadOnly = NavigationOnly;
            _ViewModel.PropertyChanged += ViewModel_PropertyChanged;
            _ViewModel.ThumbnailReceived += _ViewModel_ThumbnailReceived;
            _ViewModel.ThumbnailsList.VisibleRangeChanged += ThumbnailsList_VisibleRangeChanged;
            _ViewModel.NavigationStack = _NavigationStack;
            this.DataContext = _ViewModel;
            MainListView.DataContext = _ViewModel;
        }

        private void _ViewModel_ThumbnailReceived(int pageNumber)
        {
            int index = pageNumber - 1;
            if (index >= _ViewModel.ThumbnailsList.FirstVisibleIndex && index <= _ViewModel.ThumbnailsList.LastVisibleIndex)
            {
                FrameworkElement container = (FrameworkElement)MainListView.ContainerFromIndex(index);
                Windows.UI.Xaml.Controls.Image image = UtilityFunctions.FindVisualChild<Windows.UI.Xaml.Controls.Image>(container);
                if (image != null)
                {
                    FrameworkElement parent = image.Parent as FrameworkElement;
                    if (parent != null)
                    {
                        parent.Opacity = 0;
                        parent.Name = "image" + index;
                        Storyboard opacityAnim = OpacityAnimation(parent, index);
                        opacityAnim.Begin();
                    }
                }
            }
        }

        private Storyboard OpacityAnimation(FrameworkElement target, int index)
        {
            Storyboard sb = new Storyboard();
            string ID = "anim" + index + "OpacityAnimation";
            if (target.Resources.ContainsKey(ID))
            {
                return target.Resources[ID] as Storyboard;
            }
            double from = 0;
            double to = 1;
            DoubleAnimation fadeIn = new DoubleAnimation
            {
                From = from,
                To = to,
                Duration = new Duration(TimeSpan.FromMilliseconds(200)),
                RepeatBehavior = new RepeatBehavior { Count = 1 },
                EasingFunction = new QuadraticEase() { EasingMode = EasingMode.EaseIn },
            };

            target.Resources.Add(ID, sb);
            Storyboard.SetTarget(fadeIn, target);
            Storyboard.SetTargetName(fadeIn, target.Name);
            Storyboard.SetTargetProperty(fadeIn, "(UIElement.Opacity)");
            sb.Children.Add(fadeIn);

            return sb;
        }

        private void ViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("CurrentItem", StringComparison.OrdinalIgnoreCase))
            {
                if (ViewModel.CurrentItem != null)
                {
                    MainListView.ScrollIntoView(ViewModel.CurrentItem);
                }
            }
        }

        /// <summary>
        /// Display the flyout (defined in resources) when a ThumbnailItem is right tapped 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ThumbnailItem_RightTapped(object sender, RightTappedRoutedEventArgs e)
        {
            if (NavigationOnly || _ViewModel == null ||  ViewModel.IsConverting)
            {
                return;
            }
            ThumbnailItem selectedItem = (sender as FrameworkElement).DataContext as ThumbnailItem;
 
            ListViewItem item = MainListView.ContainerFromItem(selectedItem) as ListViewItem;

            // Align the flyout
            var pos = e.GetPosition(this);
            MarginTextBlock.Margin = new Thickness(pos.X, pos.Y, 0, 0);
            RightClickButtonFlyout.ShowAt(MarginTextBlock);

            IList<MenuFlyoutItemBase> buttons = RightClickButtonFlyout.Items;

            // Since flyout is defined in resources, set the CommandParameter programatically 
            (buttons[0] as MenuFlyoutItem).CommandParameter = selectedItem;
            (buttons[1] as MenuFlyoutItem).CommandParameter = selectedItem;
            (buttons[2] as MenuFlyoutItem).CommandParameter = selectedItem;

            e.Handled = true;
        }
        private void MainListView_ItemClick(object sender, ItemClickEventArgs e)
        {
            ItemClicked?.Invoke(sender, e);
        }

        public event pdftron.PDF.Tools.Controls.ControlBase.ControlClosedDelegate ControlClosed = delegate { };
        public event EventHandler<ItemClickEventArgs> ItemClicked;

        public void CloseControl()
        {
            _ViewIsVisible = false;
            _ViewModel.CleanUp();
            ControlClosed();
        }

        private void UpdateNavigationOnly()
        {
            if (NavigationOnly)
            {
                MainCommandBar.Visibility = Visibility.Collapsed;
                BackgroundGrid.Margin = new Thickness(0);
                FabButton.Visibility = Visibility.Collapsed;
            }
            else
            {
                MainCommandBar.Visibility = Visibility.Visible;
                BackgroundGrid.Margin = new Thickness(0, 48, 0, 0);
                FabButton.Visibility = Visibility.Visible;
            }
            if (_ViewModel != null)
            {
                _ViewModel.ReadOnly = NavigationOnly;
            }
        }
        
        /// <summary>
        /// Will perform a back operation if applicable.
        /// Use this when the back button is pressed to give this control a chance to handle it.
        /// </summary>
        /// <returns>true if the control could go back</returns>
        public bool GoBack()
        {
            return _ViewModel.GoBack();
        }

        public void PageModified(int pageNumber)
        {
            if (ViewIsVisible)
            {
                _ViewModel.PageModified(pageNumber);
            }
            else if (!_RequestedIndexesSinceViewChanged.Contains(pageNumber))
            {
                _RequestedIndexesSinceViewChanged.Add(pageNumber);
            }
        }

        public void SelectPages(IList<int> pagesToSelect)
        {
            if (_ViewModel != null)
            {
                _ViewModel.EditCommand.Execute(null);
                MainListView.SelectedItems.Clear();
                foreach (int i in pagesToSelect)
                {
                    if (i >= 1 && i <= MainListView.Items.Count)
                    {
                        MainListView.SelectedItems.Add(MainListView.Items[i - 1]);
                    }
                }
            }
        }

        #region Calculate Thumbnail Size

        private double _LastWidthCalculated = double.NaN;

        private int _ColumnsCalculcated = 1;

        private int CalculateNumberOfcolumns(double listWidth)
        {
            Thickness itemPadding = (Thickness)GetValue(ItemPaddingProperty);
            double perItemWidth = AutoSizingMaxWidth + itemPadding.Left + itemPadding.Right;
            double numItems = listWidth / perItemWidth;
            int idealColumnNumber = (int)Math.Ceiling(numItems);
            if (idealColumnNumber < 0)
            {
                idealColumnNumber = 1;
            }
            return idealColumnNumber;
        }

        private async void CalculateItemWidth(double listWidth)
        {
            if (double.IsNaN(listWidth) || listWidth <= 0 || listWidth == _LastWidthCalculated)
            {
                return;
            }
            _LastWidthCalculated = listWidth;
            _ColumnsCalculcated = 1;
            if (NumberOfColumns <= 0)
            {
                _ColumnsCalculcated = CalculateNumberOfcolumns(listWidth);
            }
            else
            {
                _ColumnsCalculcated = NumberOfColumns;
            }
            double itemWidth = 10;

            if (_ColumnsCalculcated > 0)
            {
                Thickness itemPadding = (Thickness)GetValue(ItemPaddingProperty);
                double perGapSize = itemPadding.Left + itemPadding.Right;
                double gapSize = perGapSize * _ColumnsCalculcated;
                double remainingwidth = listWidth - gapSize - ListViewMarginAndPadding;
                itemWidth = remainingwidth / _ColumnsCalculcated;
                double rawPixels = DisplayInformation.GetForCurrentView().RawPixelsPerViewPixel;

                int numRawPixels = (int)(itemWidth * rawPixels);
                itemWidth = numRawPixels / rawPixels;
            }

            int previousFirst = _PreviousLowest;// _ViewModel.ThumbnailsList.FirstVisibleIndex;
            int previousLast = _PreviousHighest; //_ViewModel.ThumbnailsList.LastVisibleIndex;

            ListViewItemWidth = Math.Max(itemWidth, 10);
            double widthToHeightRatio = (double)Resources["ThumbnailItemWidthToHeightRatio"];
            ListViewItemHeight = ListViewItemWidth / widthToHeightRatio;

            // Need to set this again, so that we re-use the binding.
            MainListView.ItemTemplate = (DataTemplate)this.Resources["ThumbnailTemplateVariableSize"];


            if (_ViewModel != null)
            {
                Thickness t = (Thickness)this.Resources["TemplateImageMargin"];
                _ViewModel.ImageMaxDimensions = new Windows.Foundation.Size(ListViewItemWidth - t.Left - t.Right, ListViewItemHeight - t.Top - t.Bottom);
            }

            await this.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                if (_ViewModel.ThumbnailsList.Count >= 1 && previousFirst >= 0 && previousLast >= 0)
                {
                    int newfirst = _ViewModel.ThumbnailsList.FirstVisibleIndex;
                    int newlast = _ViewModel.ThumbnailsList.LastVisibleIndex;
                    int oldCenter = (previousFirst + previousLast) / 2;
                    int newCenter = (newfirst + newlast) / 2;
                    if (oldCenter != newCenter)
                    {
                        int newSpan = newlast - newfirst + 1;
                        int scrollToIndex = oldCenter;
                        if (newCenter > oldCenter)
                        {
                            scrollToIndex -= (newSpan / 2);
                        }
                        else
                        {
                            scrollToIndex += (newSpan / 2);
                        }
                        scrollToIndex = Math.Max(0, Math.Min(scrollToIndex, _ViewModel.ThumbnailsList.Count - 1));
                        MainListView.ScrollIntoView(_ViewModel.ThumbnailsList[scrollToIndex]);
                    }
                }
            });
            
        }

        int _PreviousLowest = -1;
        int _CurrentLowest = -1;
        int _PreviousHighest = -1;
        int _CurrentHighest = -1;
        private void ThumbnailsList_VisibleRangeChanged(object sender, ItemIndexRange e)
        {
            _PreviousLowest = _CurrentLowest;
            _CurrentLowest = ViewModel.ThumbnailsList.FirstVisibleIndex;
            _PreviousHighest = _CurrentHighest;
            _CurrentHighest = ViewModel.ThumbnailsList.LastVisibleIndex;
        }



        private double ListViewMarginAndPadding
        {
            get { return MainListView.Margin.Left + MainListView.Margin.Right + MainListView.Padding.Left + MainListView.Padding.Right; }
        }

        #endregion Calculate Thumbnail Size
    }
}
