﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;

// This partial definition can be shared between Phone and Windows, allowing them to each specify their own styles and extra code behind

namespace pdftron.PDF.Tools.Controls
{
    partial class UserBookmarkControl
    {
        #region Show Add Button

        static DependencyProperty AddButtonVisibilityProperty =
              DependencyProperty.Register("AddButtonVisibility", typeof(Visibility),
              typeof(UserBookmarkControl), new PropertyMetadata(Visibility.Visible));

        public Visibility AddButtonVisibility
        {
            get { return (Visibility)(GetValue(AddButtonVisibilityProperty)); }
            set { SetValue(AddButtonVisibilityProperty, value); }
        }

        static DependencyProperty CanAwalysEditProperty =
              DependencyProperty.Register("CanAwalysEdit", typeof(bool),
              typeof(UserBookmarkControl), new PropertyMetadata(true, OnCanAwalysEditChanged));

        private static void OnCanAwalysEditChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            UserBookmarkControl ctrl = d as UserBookmarkControl;
            if (ctrl != null)
            {
                ctrl._ViewModel.CanAlwaysEdit = (bool)e.NewValue;
            }
        }

        public bool CanAwalysEdit
        {
            get { return (bool)(GetValue(CanAwalysEditProperty)); }
            set { SetValue(CanAwalysEditProperty, value); }
        }

        #endregion

        public ViewModels.UserBookmarksViewModel _ViewModel { get; private set; }

        private NavigationStack _NavigationStack;
        public NavigationStack NavigationStack
        {
            get { return _ViewModel != null ? _ViewModel.NavigationStack : _NavigationStack; }
            set
            {
                _NavigationStack = value;
                if (_ViewModel != null)
                {
                    _ViewModel.NavigationStack = _NavigationStack;
                }
            }
        }

        public delegate void DocumentModifiedEventHandler(PDFDoc doc);
        public DocumentModifiedEventHandler DocumentModified;
        public delegate void BookmarksEditedEventHandler(PDFDoc doc);
        public event BookmarksEditedEventHandler BookmarksEdited;

        private void RaiseDocumentModified(PDFDoc doc)
        {
            if (DocumentModified != null)
            {
                DocumentModified(doc);
            }
        }

        public UserBookmarkControl()
        {
            this.InitializeComponent();
            this.BookmarksListView.GotFocus += BookmarksListView_GotFocus;
            this.BookmarksListView.LostFocus += BookmarksListView_LostFocus;
        }

        private void BookmarksListView_LostFocus(object sender, RoutedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("BookmarksListView lost focus");
        }

        private void BookmarksListView_GotFocus(object sender, RoutedEventArgs e)
        {
            System.Diagnostics.Debug.WriteLine("BookmarksListView got focus");
        }

        public void SaveBookmarks()
        {
            _ViewModel.SaveBookmarks();
        }

        public async Task WaitForBookmarkSavingAsync()
        {
            if (_ViewModel != null)
            {
                await _ViewModel.WaitForBookmarkSavingAsync();
            }
        }

        public UserBookmarkControl(PDFViewCtrl control, string docTag)
        {
            this.InitializeComponent();
            SetPDFViewCtrl(control, docTag);
        }

        public void SetPDFViewCtrl(PDFViewCtrl control, string docTag)
        {
            _ViewModel = new ViewModels.UserBookmarksViewModel();
            _ViewModel.CanAlwaysEdit = CanAwalysEdit;
            _ViewModel.Init(control, docTag);
            _ViewModel.NavigationStack = NavigationStack;
            this.DataContext = _ViewModel;
            _ViewModel.FocusOnSelectedItemRequested += ViewModelFocusOnSelectedItemRequested;
            _ViewModel.BookmarkList.CollectionChanged += BookmarkList_CollectionChanged;
            _ViewModel.DocumentModified += ViewModelDocumentModified;
            _ViewModel.BookmarksEdited += ViewModel_BookmarksEdited;
            _ViewModel.PropertyChanged += _ViewModel_PropertyChanged;
            BookmarksListView.Loaded += BookmarksListView_Loaded;
            BookmarksListView.ItemClick += BookmarksListView_ItemClick;
        }

        private void BookmarksListView_Loaded(object sender, RoutedEventArgs e)
        {
            if (_RestoreScrollPosition > 0)
            {
                ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
            }
        }

        private double _RestoreScrollPosition = -1; // negative number means don't restore
        public void RestoreState(double scrollPosition)
        {
            if (_ViewModel != null && _ViewModel.IsListPopulated)
            {
                ScrollRestoreOffsetIntoView(scrollPosition);
            }
            else
            {
                _RestoreScrollPosition = scrollPosition;
            }
        }

        private async void ScrollRestoreOffsetIntoView(double scrollPosition)
        {
            BookmarksListView.UpdateLayout();
            await this.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                ScrollViewer sv = UtilityFunctions.FindVisualChild<ScrollViewer>(BookmarksListView);
                if (sv != null)
                {
                    sv.ChangeView(null, scrollPosition, null);
                    _RestoreScrollPosition = -1;
                }
            });
        }

        public double GetState()
        {
            ScrollViewer sv = UtilityFunctions.FindVisualChild<ScrollViewer>(BookmarksListView);
            if (sv != null)
            {
                return sv.VerticalOffset;
            }
            return 0;
        }

        private void _ViewModel_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (_RestoreScrollPosition >= 0 && e.PropertyName.Equals("IsListPopulated", StringComparison.OrdinalIgnoreCase))
            {
                ScrollRestoreOffsetIntoView(_RestoreScrollPosition);
            }
        }

        void BookmarkList_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (_ViewModel.IsListPopulated && e.NewItems != null && e.NewItems.Count > 0)
            {
                BookmarksListView.ScrollIntoView(e.NewItems[0]);
            }
        }

        void BookmarksListView_ItemClick(object sender, ItemClickEventArgs e)
        {
            if (ItemClicked != null)
            {
                ItemClicked(this, e);
            }
        }

        private KeyEventHandler _KeyDownHandler = null;
        private RoutedEventHandler _FocusLostHandler = null;
        private string _OriginalText = "";

        public event EventHandler<ItemClickEventArgs> ItemClicked;

        private void ViewModelFocusOnSelectedItemRequested(ViewModels.UserBookmarksViewModel.UserBookmarkItem item)
        {
            ShowTextBox(item);
        }

        private void ViewModelDocumentModified(PDFDoc doc)
        {
            RaiseDocumentModified(doc);
        }
        private void ViewModel_BookmarksEdited(PDFDoc doc)
        {
            BookmarksEdited?.Invoke(doc);
        }

        private void ShowTextBox(ViewModels.UserBookmarksViewModel.UserBookmarkItem item)
        {
            FrameworkElement containerElement = BookmarksListView.ContainerFromItem(item) as FrameworkElement;
            if (containerElement != null)
            {
                BookmarksListView.ScrollIntoView(item);
                TextBox box = GetChildTextbox(containerElement);
                if (box != null)
                {
                    SetFocus(box);
                }
            }
        }

        private void SetFocus(TextBox box)
        {
            box.LostFocus += TextBox_LostFocus;
            box.KeyDown += TextBox_KeyDown;
            box.GotFocus += box_GotFocus;
            box.Focus(FocusState.Programmatic);
            box.SelectAll();
            _OriginalText = box.Text;
        }

        void box_GotFocus(object sender, RoutedEventArgs e)
        {
            //_ViewModel.TextBoxGotFocus();
        }

        void TextBox_KeyDown(object sender, KeyRoutedEventArgs e)
        {
            if (e.Key == Windows.System.VirtualKey.Enter)
            {
                _ViewModel.FinishEditingText();
                e.Handled = true;
            }
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                _ViewModel.CancelEditingText();
                TextBox box = sender as TextBox;
                box.Text = _OriginalText;
                e.Handled = true;
            }
        }

        void TextBox_LostFocus(object sender, RoutedEventArgs e)
        {
            //_ViewModel.TextBoxLostFocus();
            _ViewModel.FinishEditingText();
            TextBox box = sender as TextBox;
            CloseTextBox(box);
        }

        private void CloseTextBox(TextBox box)
        {
            if (_KeyDownHandler != null)
            {
                box.KeyDown -= _KeyDownHandler;
            }
            if (_FocusLostHandler != null)
            {
                box.LostFocus -= _FocusLostHandler;
            }
        }

        private TextBox GetChildTextbox(FrameworkElement containerElement)
        {
            DependencyObject box = FindChildControl<TextBox>(containerElement);
            return box as TextBox;
        }

        private DependencyObject FindChildControl<T>(DependencyObject control)
        {
            int childNumber = VisualTreeHelper.GetChildrenCount(control);
            for (int i = 0; i < childNumber; i++)
            {
                DependencyObject child = VisualTreeHelper.GetChild(control, i);
                FrameworkElement fe = child as FrameworkElement;
                // Not a framework element or is null
                if (fe == null) return null;

                if (child is T)
                {
                    // Found the control so return
                    return child;
                }
                else
                {
                    // Not found it - search children
                    DependencyObject nextLevel = FindChildControl<T>(child);
                    if (nextLevel != null)
                        return nextLevel;
                }
            }
            return null;
        }

        /// <summary>
        /// Will go back if possible.
        /// </summary>
        /// <returns>Return true if it went back, false otherwise</returns>
        public bool GoBack()
        {
            if (_ViewModel != null)
            {
                return _ViewModel.GoBack();
            }
            return false;
        }
    }
}
