﻿using pdftron.Common;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class ReflowItem : ViewModelBase
    {
        private string _ReflowedContent = "";
        public string ReflowedContent
        {
            get { return _ReflowedContent; }
            internal set
            {
                if (Set(ref _ReflowedContent, value))
                {
                    RaisePropertyChanged("HasReflowedContent");
                }
            }
        }

        private int _PageNumber = 0;
        public int PageNumber
        {
            get { return _PageNumber; }
            internal set { Set(ref _PageNumber, value); }
        }

        public bool HasReflowedContent { get { return !string.IsNullOrEmpty(_ReflowedContent); } }

        public ReflowItem(int pageNumber)
        {
            PageNumber = pageNumber;
        }

        public void Reset()
        {
            ReflowedContent = string.Empty;
        }
    }

    public class ReflowViewModel : ViewModelBase
    {
        public const string KEYDOWN_STRING = "<script> document.onkeydown = function (e) {window.external.notify(\"key: \" + e.keyCode); } </script>";
        // Ensures paragraphs don't span to the right edge of the display, to allow for tapping to bring up command bar.
        public const string PARAGRAPH_STYLE_STRING = "<style>p{display:inline;} p:not(:last-of-type):after{content:\".\";display:block;height:0;clear:both;visibility:hidden;margin-bottom:1em;} p:not(:first-of-type):before{content:\".\";display:block;height:0;clear:both;visibility:hidden;margin-top:1em;}</style>";
        public const string TAP_STRING = "<script> document.onclick = function (e) { if (e.pointerType === 'touch' && (e.target === document.body || e.target.tagName == 'DIV')) window.external.notify(\"tap\"); }</script>";
        private const int PAGE_FETCHING_RANGE = 3;
        private const string LINK_PREFIX = "LaunchLink:";
        public delegate void PageChangedDelegate(int pageNumber, int totalPages);
        public event PageChangedDelegate PageChanged;

        public delegate bool PageChangeIntendedDelegate(int pageNumber);
        /// <summary>
        /// Gets raised when the ReflowViewModel is about to change pages through a link being clicked
        /// Handle this and return true to make the ReflowViewModel abort.
        /// </summary>
        public event PageChangeIntendedDelegate PageChangeIntended;

        public ReflowViewModel()
        {
            Init();
        }

        public ReflowViewModel(PDFDoc doc)
        {
            Init();
            Doc = doc;
        }

        #region Properties

        private Dictionary<int, ReflowItem> _ReflowedPages;

        public Dictionary<int, ReflowItem> ReflowedPages
        {
            get { return _ReflowedPages; }
        }

        private PDFDoc _Doc = null;
        public PDFDoc Doc
        {
            get { return _Doc; }
            set
            {
                PDFDoc oldDoc = _Doc;
                if (Set(ref _Doc, value))
                {
                    if (oldDoc != null)
                    {
                        ReflowProcessor.ReflowProcessorResponse -= ReflowProcessor_ReflowProcessorResponse;
                    }
                    else if (_Doc != null)
                    {
                        ReflowProcessor.ReflowProcessorResponse += ReflowProcessor_ReflowProcessorResponse;
                    }
                }
            }
        }

        public int TotalPageNumbers
        {
            get
            {
                if (Doc != null)
                {
                    return Doc.GetPageCount();
                }
                return 0;
            }
        }

        private int _CurrentPageNumber = 0;
        public int CurrentPageNumber
        {
            get { return _CurrentPageNumber;  }
            private set
            {
                if (Doc != null && Set(ref _CurrentPageNumber, value))
                {
                    if (!ReflowedPages.ContainsKey(_CurrentPageNumber))
                    {
                        ReflowedPages[_CurrentPageNumber] = new ReflowItem(_CurrentPageNumber);
                    }
                    CurrentPage = ReflowedPages[_CurrentPageNumber];
                    PageChanged?.Invoke(CurrentPageNumber, TotalPageNumbers);
                    RaisePropertyChanged("IsPrevButtonEnabled");
                    RaisePropertyChanged("IsNextButtonEnabled");
                    RequestNextPage();
                }
            }
        }

        private ReflowItem _CurrentPage = new ReflowItem(0);
        public ReflowItem CurrentPage
        {
            get { return _CurrentPage; }
            private set
            {
                if(Set(ref _CurrentPage, value))
                {
                    UpdateReflowProperties();
                }
            }
        }

        public bool HasReflowedContent
        {
            get
            {
                if (CurrentPage == null)
                {
                    return false;
                }
                return CurrentPage.HasReflowedContent;
            }
        }

        public string ReflowedContent
        {
            get
            {
                if (CurrentPage == null)
                {
                    return "";
                }
                return CurrentPage.ReflowedContent;
            }
        }

        public string GetReflowProgressText
        {
            get
            {
                int pageNumber = 1;
                if (CurrentPage != null)
                {
                    pageNumber = CurrentPage.PageNumber;
                }
                return string.Format(ResourceHandler.GetString("Reflow_WorkingOnPage"), pageNumber);
            }
        }

        public bool IsPrevButtonEnabled { get { return Doc != null && CurrentPageNumber > 1; } }
        public bool IsNextButtonEnabled { get { return Doc != null && CurrentPageNumber < TotalPageNumbers; } }

        
        public PDFRasterizerColorPostProcessMode PostProcessingMode { get { return _PostProcessBundle.PostProcessMode; } }
        public Windows.UI.Color PostProcessingWhiteColor { get { return _PostProcessBundle.PostProcessWhiteColor; } }
        public Windows.UI.Color PostProcessingBlackColor { get { return _PostProcessBundle.PostProcessBlackColor; } }

        public PDFViewCtrl PostProcessPDFView
        {
            get
            {
                if (_PostProcessPDFView == null)
                {
                    _PostProcessPDFView = new PDFViewCtrl();
                }
                return _PostProcessPDFView;
            }
        }

        private void UpdateReflowProperties()
        {
            RaisePropertyChanged("HasReflowedContent");
            RaisePropertyChanged("ReflowedContent");
            RaisePropertyChanged("GetReflowProgressText");
        }

        private void UpdateReflowColors()
        {
            RaisePropertyChanged("PostProcessingMode");
            RaisePropertyChanged("PostProcessingWhiteColor");
            RaisePropertyChanged("PostProcessingBlackColor");
        }

        #endregion Properties


        #region Public Functions

        /// <summary>
        /// Sets the current page to pageNumber
        /// </summary>
        /// <param name="pageNumber">the page number to navigate to</param>
        /// <returns>True if the pageNumber is a valid page within range</returns>
        public bool SetPageNumber(int pageNumber)
        {
            if (Doc != null && pageNumber > 0 && pageNumber <= TotalPageNumbers)
            {
                CurrentPageNumber = pageNumber;
                return true;
            }
            return false;
        }

        public void ShutDown()
        {
            try
            {
                ReflowProcessor.CancelAllRequests();
                Doc = null;
            }
            catch (Exception) { }
        }

        public async void SetColorPostProcessingMode(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
        {
            await _ReflowSemaphore.WaitAsync();
            try
            {
                PostProcessBundle newPostProcess = new PostProcessBundle(mode, whiteColor, blackColor);
                if (!newPostProcess.Equals(_PostProcessBundle))
                {
                    _PostProcessBundle = newPostProcess;
                    if (mode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                        mode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                        mode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                    {
                        PostProcessPDFView.SetColorPostProcessMode(mode);
                    }
                    else
                    {
                        PostProcessPDFView.SetColorPostProcessColors(whiteColor, blackColor);
                    }

                    foreach (ReflowItem item in ReflowedPages.Values)
                    {
                        item.Reset();
                    }
                    UpdateReflowProperties();
                    UpdateReflowColors();
                    RequestNextPage();
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _ReflowSemaphore.Release();
            }
        }

        /// <summary>
        /// Will try to navigate to an internal link. Returns true if successful
        /// </summary>
        /// <param name="link"></param>
        /// <returns>true if link was a link and it could sucessfully navigate to it</returns>
        public bool NavigateInternalLink(string link)
        {
            int linkLength = LINK_PREFIX.Length;
            if (link.Length < linkLength || Doc == null)
            {
                return false;
            }
            if (link.Substring(0, linkLength).Equals(LINK_PREFIX))
            {
                string htmlLink = link.Substring(linkLength);
                if (htmlLink.StartsWith("file:///") && htmlLink.EndsWith(".html"))
                {
                    try
                    {
                        int slashPos = htmlLink.LastIndexOf('/');
                        string objNumString = htmlLink.Substring(slashPos + 1, htmlLink.Length - 5 - (slashPos + 1));
                        int objNum = int.Parse(objNumString);
                        int pageNum = 0;
                        Doc.LockRead();
                        try
                        {
                            for (PageIterator itr = Doc.GetPageIterator(); itr.HasNext(); itr.Next())
                            {
                                if (itr.Current().GetSDFObj().GetObjNum() == objNum)
                                {
                                    pageNum = itr.GetPageNumber();
                                    break;
                                }
                            }
                        }
                        catch (Exception e)
                        {
                            PDFNetException pdfe = new PDFNetException(e.HResult);
                            if (!pdfe.IsPDFNetException)
                            {
                                throw;
                            }
                            else
                            {
                                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                            }
                        }
                        finally
                        {
                            Doc.UnlockRead();
                        }

                        if (pageNum > 0)
                        {
                            if (PageChangeIntended == null || !PageChangeIntended(pageNum))
                            {
                                CurrentPageNumber = pageNum;
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        PDFNetException pdfe = new PDFNetException(e.HResult);
                        if (!pdfe.IsPDFNetException)
                        {
                            throw;
                        }
                        else
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                        }
                    }
                }
                else
                {
                    ActionUtils.LaunchBrowser(htmlLink);
                }
            }

            return false;
        }

        public void UpdatePageNumberProperties()
        {
            RaisePropertyChanged("TotalPageNumbers");
            RaisePropertyChanged("IsPrevButtonEnabled");
            RaisePropertyChanged("IsNextButtonEnabled");
        }

        #endregion Public Functions


        #region Commands

        private void InitCommands()
        {
            NavigationCommand = new RelayCommand(NavigationCommandImpl);
        }

        public RelayCommand NavigationCommand { get; private set; }

        private void NavigationCommandImpl(object parameter)
        {
            string pageNumStr = parameter as string;
            if (!string.IsNullOrEmpty(pageNumStr))
            {
                if (pageNumStr.Equals("prev", StringComparison.OrdinalIgnoreCase))
                {
                    CurrentPageNumber = Math.Max(CurrentPageNumber - 1, 1);
                }
                else if (pageNumStr.Equals("next", StringComparison.OrdinalIgnoreCase))
                {
                    CurrentPageNumber = Math.Min(CurrentPageNumber + 1, TotalPageNumbers);
                }
            }
        }

        #endregion Commands


        #region Impl

        private int _CurrentJobPage = 0;
        private SemaphoreSlim _ReflowSemaphore = null;

        private class PostProcessBundle
        {
            public PDFRasterizerColorPostProcessMode PostProcessMode { get; private set; }
            public Windows.UI.Color PostProcessWhiteColor { get; private set; }
            public Windows.UI.Color PostProcessBlackColor { get; private set; }

            public PostProcessBundle(PDFRasterizerColorPostProcessMode mode)
            {
                PostProcessMode = mode;
                PostProcessWhiteColor = Windows.UI.Colors.White;
                PostProcessBlackColor = Windows.UI.Colors.Black;
        }

            public PostProcessBundle(PDFRasterizerColorPostProcessMode mode, Windows.UI.Color whiteColor, Windows.UI.Color blackColor)
            {
                PostProcessMode = mode;
                PostProcessWhiteColor = whiteColor;
                PostProcessBlackColor = blackColor;
            }

            public override bool Equals(object obj)
            {
                PostProcessBundle other = obj as PostProcessBundle;
                if (other != null)
                {
                    if (PostProcessMode != other.PostProcessMode)
                    {
                        return false;
                    }

                    if (PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                        PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                        PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                    {
                        return true;
                    }

                    return PostProcessWhiteColor == other.PostProcessWhiteColor &&
                        PostProcessBlackColor == other.PostProcessBlackColor;
                }
                
                return base.Equals(obj);
            }

            public override int GetHashCode()
            {
                if (PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_none ||
                       PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_invert ||
                       PostProcessMode == PDFRasterizerColorPostProcessMode.e_postprocess_night_mode)
                {
                    return (int)PostProcessMode;
                }

                return (int)PostProcessMode * PostProcessBlackColor.GetHashCode() * PostProcessWhiteColor.GetHashCode();
            }
        }

        private PostProcessBundle _PostProcessBundle;
        private PDFViewCtrl _PostProcessPDFView;

        private void Init()
        {
            if (!ReflowProcessor.IsInitialized)
            {
                ReflowProcessor.Initialize();
            }

            _ReflowedPages = new Dictionary<int, ReflowItem>();
            _ReflowSemaphore = new SemaphoreSlim(1);
            _PostProcessBundle = new PostProcessBundle(PDFRasterizerColorPostProcessMode.e_postprocess_none);
            InitCommands();
        }

        private async void ReflowProcessor_ReflowProcessorResponse(JobRequestResultState result, string resultFileName, object customData)
        {
            await _ReflowSemaphore.WaitAsync();
            try
            {
                int page = System.Convert.ToInt32(customData);
                await Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () =>
                {
                    if (page == _CurrentJobPage)
                    {
                        _CurrentJobPage = 0;
                    }
                    if (!ReflowedPages.ContainsKey(page))
                    {
                        ReflowedPages[page] = new ReflowItem(page);
                    }
                    if ((result == JobRequestResultState.e_success || result == JobRequestResultState.e_failure) && page > 0)
                    {
                        string str = System.IO.File.ReadAllText(resultFileName);
                        
                        // WebView steals all input gesture commands. Need to manually send it back out. 
                        string js = "<script> document.onmousedown = function (e) {if (e.which == 3) window.external.notify(\"rightClick\"); } </script>";
                        int index = str.IndexOf("</head>");
                        str = str.Insert(index, js);

                        str = str.Insert(index, TAP_STRING);
                        str = str.Insert(index, PARAGRAPH_STYLE_STRING);
                        str = str.Insert(index, KEYDOWN_STRING);

                        index = str.IndexOf("</html>");
                        js = string.Format("{0}{1}{2}",
                            "<script>for (var i = 0; i < document.links.length; i++) { document.links[i].onclick = function() { window.external.notify('",
                            LINK_PREFIX, "' + this.href); return false; } }</script>");
                        //js = string.Format("<script>for (var i = 0; i < document.links.length; i++) { document.links[i].onclick = function() { window.external.notify('{0}' + this.href); return false; } }</script>", LINK_PREFIX);
                        str = str.Insert(index, js);

                        bool oldHasReflow = HasReflowedContent;
                        bool postProcessed = false;
                        if (PostProcessingMode != PDFRasterizerColorPostProcessMode.e_postprocess_none)
                        {
                            try
                            {
                                ReflowedPages[page].ReflowedContent = PostProcessReflowContent(str);
                                postProcessed = true;
                            }
                            catch (Exception e)
                            {
                                PDFNetException pdfe = new PDFNetException(e.HResult);
                                if (!pdfe.IsPDFNetException)
                                {
                                    throw;
                                }
                                else
                                {
                                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(e));
                                }
                            }
                        }
                        if (!postProcessed)
                        {
                            ReflowedPages[page].ReflowedContent = str;
                        }
                        if (oldHasReflow != HasReflowedContent)
                        {
                            UpdateReflowProperties();
                        }
                        RequestNextPage();
                    }
                });
            }
            catch (Exception) { }
            finally
            {
                _ReflowSemaphore.Release();
            }
        }

        private void RequestNextPage()
        {
            if (CurrentPageNumber <= 0)
            {
                return;
            }
            try
            {
                if (!ReflowedPages.ContainsKey(CurrentPageNumber))
                {
                    ReflowedPages[CurrentPageNumber] = new ReflowItem(CurrentPageNumber);
                }
                // If current page is currently being processed, don't change anything
                if (_CurrentJobPage == CurrentPageNumber)
                {
                    return;
                }
                // If current page is not available, request it
                else if (!ReflowedPages[CurrentPageNumber].HasReflowedContent)
                {
                    ReflowProcessor.CancelAllRequests();
                    Page page = Doc.GetPage(CurrentPageNumber);
                    ReflowProcessor.GetReflow(page, CurrentPageNumber);
                    _CurrentJobPage = CurrentPageNumber;
                }
                else if (_CurrentJobPage < 1) // no current job
                {
                    // Pre-render in following order:
                    // 1 ahead, 1 behind, 2 ahead, 2 behind, etc
                    int nextJobPage = CurrentPageNumber;
                    bool lookingForJob = PAGE_FETCHING_RANGE > 0;
                    while (lookingForJob)
                    {
                        if (nextJobPage > CurrentPageNumber)
                        {
                            nextJobPage = CurrentPageNumber + CurrentPageNumber - nextJobPage;
                        }
                        else
                        {
                            nextJobPage = CurrentPageNumber + CurrentPageNumber - nextJobPage + 1;
                        }
                        if (Math.Abs(nextJobPage - CurrentPageNumber) > PAGE_FETCHING_RANGE)
                        {
                            return;
                        }
                        if (nextJobPage > 0 && nextJobPage <= TotalPageNumbers)
                        {
                            if (!ReflowedPages.ContainsKey(nextJobPage))
                            {
                                ReflowedPages[nextJobPage] = new ReflowItem(nextJobPage);
                            }
                            if (!ReflowedPages[nextJobPage].HasReflowedContent)
                            {
                                lookingForJob = false;
                            }
                        }
                    }
                    if (nextJobPage > 0 && nextJobPage != CurrentPageNumber)
                    {
                        Page page = Doc.GetPage(nextJobPage);
                        ReflowProcessor.GetReflow(page, nextJobPage);
                        _CurrentJobPage = nextJobPage;
                    }
                }
            }
            catch (Exception) { }
        }

        private string PostProcessReflowContent(string input)
        {
            StringBuilder sb = new StringBuilder();

            Regex tagStr = new Regex("<.*?>");
            Regex colorStr = new Regex("color:#[0-9|a-f|A-F]{6}");
            Regex backgroundColorStr = new Regex("background-color:#[0-9|a-f|A-F]{6}");
            Regex paragraphStr = new Regex("<p.*?>");
            Regex spanStr = new Regex("<span.*?>");

            MatchCollection matches = tagStr.Matches(input);
            int currStringIndex = 0;
            foreach (Match match in matches)
            {
                sb.Append(input.Substring(currStringIndex, match.Index - currStringIndex));
                currStringIndex = match.Index + match.Length;
                string matchedString = match.Value;
                int matchedStringIndex = 0;
                MatchCollection colorMatches = colorStr.Matches(matchedString);
                MatchCollection paraMatches = paragraphStr.Matches(matchedString);
                MatchCollection spanMatches = spanStr.Matches(matchedString);
                MatchCollection backgroundMatches = backgroundColorStr.Matches(matchedString);

                foreach (Match colorMatch in colorMatches)
                {
                    bool isInsideParahraph = IsInsideMatch(colorMatch.Index, paraMatches);
                    bool isInsideSpan = IsInsideMatch(colorMatch.Index, spanMatches);
                    bool isInsideBackgroundColor = IsInsideMatch(colorMatch.Index, backgroundMatches);

                    sb.Append(matchedString.Substring(matchedStringIndex, colorMatch.Index - matchedStringIndex));
                    matchedStringIndex = colorMatch.Index + colorMatch.Length;
                    string colorString = colorMatch.Value;
                    if ((isInsideParahraph || isInsideSpan) && !isInsideBackgroundColor)
                    {
                        sb.Append("color:#" + PostProcessColor(colorString.Substring(7, 6)));
                    }
                    else
                    {
                        sb.Append(colorString);
                    }
                }
                sb.Append(matchedString.Substring(matchedStringIndex));
            }
            sb.Append(input.Substring(currStringIndex));

            return sb.ToString();
        }

        private bool IsInsideMatch(int index, MatchCollection matches)
        {
            foreach (Match match in matches)
            {
                if (index > match.Index && index < match.Index + match.Length)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// Convers colorString into a color that can be PostProcessed by the PDFViewCtrl
        /// </summary>
        /// <param name="colorString">string of the form RRGGBB</param>
        /// <returns></returns>
        private string PostProcessColor(string colorString)
        {
            if (colorString.Length == 6)
            {
                double r = System.Convert.ToInt32(colorString.Substring(0, 2), 16);
                double g = System.Convert.ToInt32(colorString.Substring(2, 2), 16);
                double b = System.Convert.ToInt32(colorString.Substring(4, 2), 16);
                pdftron.PDF.ColorPt cp = new ColorPt(r / 255.0, g / 255.0, b / 255.0);
                cp = PostProcessPDFView.GetPostProcessedColor(cp);
                int ppr = (int)(cp.Get(0) * 255);
                int ppg = (int)(cp.Get(1) * 255);
                int ppb = (int)(cp.Get(2) * 255);
                string retString = "";
                if (ppr < 16)
                {
                    retString += "0";
                }
                retString += ppr.ToString("X");
                if (ppg < 16)
                {
                    retString += "0";
                }
                retString += ppg.ToString("X");
                if (ppb < 16)
                {
                    retString += "0";
                }
                retString += ppb.ToString("X");
                return retString;
            }
            return colorString;
        }

        #endregion Impl

    }
}
