﻿using System;
using System.Threading.Tasks;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ControlBase;
using Windows.UI.Xaml.Media;
using System.Collections.ObjectModel;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml;
using System.Collections.Generic;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Linq;

using UIRect = Windows.Foundation.Rect;
using System.IO;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI;
using Windows.Storage;
using Windows.Storage.Pickers;
using Windows.Storage.Streams;
using pdftron.Common;
using Windows.UI.Popups;
using Windows.UI.Xaml.Data;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class ThumbnailItem : ViewModelBase
    {
        private int _PageNumber;
        public int PageNumber
        {
            get { return _PageNumber; }
            set { Set(ref _PageNumber, value); }
        }

        private ImageSource _Thumbnail;
        public ImageSource Thumbnail
        {
            get { return _Thumbnail; }
            set
            {
                Set(ref _Thumbnail, value);
                if (_Thumbnail != null)
                {
                    IsDirty = false;
                }
                RaisePropertyChanged("HasThumbnail");
            }
        }

        private Windows.Foundation.Size _BitmapSize = new Windows.Foundation.Size(0, 0);
        public Windows.Foundation.Size BitmapSize
        {
            get { return _BitmapSize; }
            set
            {
                if (Set(ref _BitmapSize, value))
                {

                }
            }
        }

        private bool _IsCurrentPage = false;
        public bool IsCurrentPage
        {
            get { return _IsCurrentPage; }
            set { Set(ref _IsCurrentPage, value); }
        }

        public bool HasThumbnail { get { return Thumbnail != null; } }
        private bool _IsDirty = false;
        public bool IsDirty
        {
            get { return _IsDirty; }
            set
            {
                Set(ref _IsDirty, value);
            }
        }


        public SolidColorBrush DefaultBackground { get; set; }

        private int _Angle;
        public int Angle
        {
            get { return _Angle; }
            set { Set(ref _Angle, value); }

        }
        private double _UniformScale = 1;
        public double UniformScale
        {
            get { return _UniformScale; }
            set { Set(ref _UniformScale, value); }

        }

        private bool? _IsLandScape;

        public bool IsLandscape
        {
            get
            {
                if (_IsLandScape != null)
                {
                    return (bool)_IsLandScape;
                }
                var bitmap = Thumbnail as BitmapSource;
                _IsLandScape = bitmap != null && bitmap.PixelHeight < bitmap.PixelWidth;
                return (bool)_IsLandScape;
            }
        }

        private bool _IsDeleted = false;
        public bool IsDeleted
        {
            get { return _IsDeleted; }
            set { _IsDeleted = value; }
        }

        public ThumbnailItem(int pageNumber, Color color)
        {
            PageNumber = pageNumber;
            DefaultBackground = new SolidColorBrush(color);
        }
    }

    /// <summary>
    /// This class lets us keep track of which items are currently visible, so that we can know which items to request thumbnails for.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class ObservableCollectionWithVisibleItemsTracking<T> : ObservableCollection<T>, IItemsRangeInfo
    {
        private int _FirstIndex = 0;
        public int FirstVisibleIndex { get { return _FirstIndex; } }

        private int _LastIndex = 0;
        public int LastVisibleIndex { get { return Math.Min(_LastIndex, Count - 1); } }

        public EventHandler<ItemIndexRange> VisibleRangeChanged;

        private ItemIndexRange _PreviousRange = new ItemIndexRange(0, 1);
        public ItemIndexRange PreviousRange { get { return _PreviousRange; } }

        private bool _FirstRequest = true;

        public void Dispose() { } // We maintain no state, so this can be empty

        public void RangesChanged(ItemIndexRange visibleRange, IReadOnlyList<ItemIndexRange> trackedItems)
        {
            if (!_FirstRequest && _PreviousRange.FirstIndex == visibleRange.FirstIndex && _PreviousRange.LastIndex == visibleRange.LastIndex)
            {
                return;
            }
            _FirstRequest = false;
            _FirstIndex = visibleRange.FirstIndex;
            _LastIndex = visibleRange.LastIndex;
            VisibleRangeChanged?.Invoke(this, visibleRange);
            _PreviousRange = visibleRange;
        }
    }


    public class ThumbnailsViewViewModel : ViewModelBase
    {
        ICloseableControl _View;

        private bool _HasModifiedPageLayout = false;
        private List<int> _ThumbnailRequestQueue;
        private static IList<string> Valid_Image_Types = new List<string>() { ".jpg", ".bmp", ".jpeg", ".tif", ".tiff", ".png", ".gif", ".ico", ".jpeg-xr" };
        private static IList<string> Valid_PDF_Types = new List<string>() { ".pdf" };

        private const int MAX_SPACE_PC = 200000000;
        private const int MAX_SPACE_PHONE = 50000000;

        private bool _ClearingSelection = false;

        private bool _GotBookmarksFromDocument = false;

        private bool _FilePickingInProgress = false;

        private bool _WindowsEditingStyle = true;
        public bool WindowsEditingStyle
        {
            get { return _WindowsEditingStyle; }
            set
            {
                if (value != _WindowsEditingStyle)
                {
                    _WindowsEditingStyle = value;
                    RaisePropertyChanged();
                    if (_WindowsEditingStyle)
                    {
                        SelectionMode = ListViewSelectionMode.Single;
                    }
                    else
                    {
                        SelectionMode = ListViewSelectionMode.None;
                    }
                }
            }
        }

        private ThumbnailItem _CurrentItem = null;
        public ThumbnailItem CurrentItem
        {
            get { return _CurrentItem; }
            private set
            {
                if (_CurrentItem != value)
                {
                    if (_CurrentItem != null)
                    {
                        _CurrentItem.IsCurrentPage = false;
                    }
                    _CurrentItem = value;
                    if (_CurrentItem != null)
                    {
                        _CurrentItem.IsCurrentPage = true;
                    }
                    RaisePropertyChanged();
                }
            }
        }

        private OnThumbnailRequestCompletedEventHandler _ThumbnailGeneratedHandler;
        private OnPageNumberChangedEventHandler _PageNumberChangedHanlder;
        private OnConversionEventHandler _ConversionChangedHandler;
        private DateTime _LastConversionEvent = DateTime.Now;

        private System.Collections.Specialized.NotifyCollectionChangedEventHandler _CollectionChangedHandler;

        private Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> _ThumbnailsList;
        public Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> ThumbnailsList
        {
            get { return _ThumbnailsList; }
            private set
            {
                if (value != _ThumbnailsList)
                {
                    _ThumbnailsList = value;
                    RaisePropertyChanged();
                }
            }
        }

        private ListViewSelectionMode _SelectionMode = ListViewSelectionMode.Multiple;
        public ListViewSelectionMode SelectionMode
        {
            get { return _SelectionMode; }
            set
            {
                if (value != _SelectionMode)
                {
                    _SelectionMode = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged("IsInEditMode");
                }
            }
        }
        public bool IsInEditMode
        {
            get { return _SelectionMode != ListViewSelectionMode.None; }
        }
        private bool _IsSelectButtonVisible = true;
        public bool IsSelectButtonVisible
        {
            get { return _IsSelectButtonVisible; }
            set
            {
                _IsSelectButtonVisible = value;
                RaisePropertyChanged("IsSelectButtonVisible");
            }
        }
        private int _NumberOfSelectedItems = 0;
        public bool HasSelectedItems { get { return _NumberOfSelectedItems > 0; } }
        public int NumberOfSelectedItems
        {
            get { return _NumberOfSelectedItems; }
            set
            {
                if (value != _NumberOfSelectedItems)
                {
                    _NumberOfSelectedItems = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged("HasSelectedItems");
                }
            }
        }

        public enum NewPageType
        {
            Blank,
            Image,
            ExistingPDF
        }

        private Color _PageDefaultColor = Colors.White;
        public Color PageDefaultColor
        {
            get { return _PageDefaultColor; }
            set
            {
                _PageDefaultColor = value;
                CreateCollection();
            }
        }

        private bool _ReadOnly = false;
        public bool ReadOnly
        {
            get { return _ReadOnly; }
            set
            {
                if (value != _ReadOnly)
                {
                    _ReadOnly = value;
                    RaisePropertyChanged();
                    if (_ReadOnly)
                    {
                        SelectionMode = ListViewSelectionMode.None;
                    }
                }
            }
        }

        public bool IsConverting
        {
            get { return _PDFViewCtrl != null && _PDFViewCtrl.IsConverting; }
        }

        private Windows.Foundation.Size _ImageMaxDimensions = new Windows.Foundation.Size(0, 0);
        public Windows.Foundation.Size ImageMaxDimensions
        {
            get { return _ImageMaxDimensions; }
            set
            {
                if (Set(ref _ImageMaxDimensions, value))
                {
                    foreach (int key in ThumbnailsList.ItemsWithResource.Keys)
                    {
                        UpdateScaleFromRotation(ThumbnailsList[key]);
                    }
                }
            }
        }

        private NavigationStack _NavigationStack;
        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        private PDFViewCtrl _PDFViewCtrl;
        PDFViewCtrlThumbnailRequestHandler _ThumbHandler;
        private string _DocumentTag;

        private ToolManager _ToolManager;
        public ToolManager ToolManager
        {
            get { return _ToolManager; }
            set { _ToolManager = value; }
        }

        public ThumbnailsViewViewModel(PDFViewCtrl ctrl, string docTag, ICloseableControl view)
        {
            _PDFViewCtrl = ctrl;
            _ThumbHandler = ctrl.ThumbnailResponseHandler;
            _View = view;

            PDFDoc doc = null;
            if (_PDFViewCtrl != null)
            {
                _DocumentTag = docTag;
                doc = _PDFViewCtrl.GetDoc();
                if (doc == null && !_PDFViewCtrl.IsConverting)
                {
                    return;
                }
            }

            SelectionChangedCommand = new RelayCommand(SelectionChangedCommandImpl);
            DeleteCommand = new RelayCommand(DeleteCommandImpl);
            RotateCommand = new RelayCommand(RotateCommandImpl);
            CloseCommand = new RelayCommand(CloseCommandImpl);
            ClearSelectionCommand = new RelayCommand(ClearSelectionCommandImpl);
            ItemClickCommand = new RelayCommand(ItemClickCommandImpl);

            EditCommand = new RelayCommand(EditCommandImpl);
            DoneEditingCommand = new RelayCommand(DoneEditingCommandImpl);

            ExportSelectedPagesCommand = new RelayCommand(ExportSelectedPagesCommandImpl);
            AddPageCommand = new RelayCommand(AddPageCommandImpl);
            AddImageCommand = new RelayCommand(AddImageCommandImpl);
            AddPDFCommand = new RelayCommand(AddPDFCommandImpl);
            InsertPageCommand = new RelayCommand(InsertPageCommandImpl);
            InsertImageCommand = new RelayCommand(InsertImageCommandImpl);
            InsertPDFCommand = new RelayCommand(InsertPDFCommandImpl);

            if (_PDFViewCtrl == null)
            {
                return;
            }

            _ThumbnailRequestQueue = new List<int>();
            _ThumbnailGeneratedHandler = new OnThumbnailRequestCompletedEventHandler(ThumbHandler_OnThumbnailGenerated);
            _ThumbHandler.OnThumbnailGenerated += _ThumbnailGeneratedHandler;
            _PageNumberChangedHanlder = new OnPageNumberChangedEventHandler(PDFViewCtrl_OnPageNumberChanged);
            _PDFViewCtrl.OnPageNumberChanged += _PageNumberChangedHanlder;
            if (_PDFViewCtrl.IsConverting)
            {
                _ConversionChangedHandler = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
                _PDFViewCtrl.OnConversionChanged += _ConversionChangedHandler;
            }
            CreateCollection();

            if (!_PDFViewCtrl.IsConverting)
            {
                GetBookmarks(doc);
            }
            RaisePropertyChanged("IsConverting");

            SelectionMode = ListViewSelectionMode.None;
        }

        private async void GetBookmarks(PDFDoc doc)
        {
            if (doc != null)
            {
                try
                {
                    if (UserBookmarkManager.DoesDocHaveBookmarks(doc))
                    {
                        await UserBookmarkManager.LoadBookmarksListAsync(doc);
                        _GotBookmarksFromDocument = true;
                    }
                    else
                    {
                        await UserBookmarkManager.LoadBookmarksListAsync(_DocumentTag);
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        private void CreateCollection()
        {
            if (ThumbnailsList != null && _CollectionChangedHandler != null)
            {
                ThumbnailsList.CollectionChanged -= _CollectionChangedHandler;
                _CollectionChangedHandler = null;
            }
            Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> items = new Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>();
            items.RequestBehaviour = Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>.RequestBehaviours.FILO;
            int pageCount = _PDFViewCtrl.GetPageCount();
            if (_PDFViewCtrl.IsConverting)
            {
                pageCount = _PDFViewCtrl.ConvertedPages;
            }
            for (int i = 1; i <= pageCount; ++i)
            {
                items.Add(new ThumbnailItem(i, _PageDefaultColor));
            }
            if (pageCount > 0)
            {
                CurrentItem = items[_PDFViewCtrl.GetCurrentPage() - 1];
            }
            ThumbnailsList = items;

            if (!Windows.System.Profile.AnalyticsInfo.VersionInfo.DeviceFamily.Equals("Windows.Mobile", StringComparison.OrdinalIgnoreCase))
            {
                ThumbnailsList.MaxAllowedSize = MAX_SPACE_PC;
            }
            else
            {
                ThumbnailsList.MaxAllowedSize = MAX_SPACE_PHONE;
            }

            _CollectionChangedHandler = new System.Collections.Specialized.NotifyCollectionChangedEventHandler(_ThumbnailsList_CollectionChanged);
            _ThumbnailsList.CollectionChanged += _CollectionChangedHandler;
            _ThumbnailsList.ResourceNeededForIndex += ThumbnailsList_ResourceNeededForIndex;
            _ThumbnailsList.DiscardResourceForIndex += ThumbnailsList_DiscardResourceForIndex;
            _ThumbnailsList.AllRequestsCancelled += ThumbnailsList_AllRequestsCancelled;
        }

        public async void CleanUp()
        {
            if (_ThumbnailGeneratedHandler != null)
            {
                _PDFViewCtrl.CancelAllThumbRequests();
                _ThumbHandler.OnThumbnailGenerated -= _ThumbnailGeneratedHandler;
                _PDFViewCtrl.OnPageNumberChanged -= _PageNumberChangedHanlder;
                _ThumbnailGeneratedHandler = null;
                _PageNumberChangedHanlder = null;

                if (_ConversionChangedHandler != null)
                {
                    _PDFViewCtrl.OnConversionChanged -= _ConversionChangedHandler;
                    _ConversionChangedHandler = null;
                }
            }

            _ThumbHandler = null;
            if (_CollectionChangedHandler != null)
            {
                _ThumbnailsList.CollectionChanged -= _CollectionChangedHandler;
                _CollectionChangedHandler = null;
            }
            try
            {
                _UnloadingMutex.Wait();
                _UnloadingJobsPending.Clear();
            }
            catch (Exception) { }
            finally
            {
                _UnloadingMutex.Release();
            }
        }

        #region Events

        public delegate void PageDeletedDelegate(int pageNumber);
        public delegate void PageAddedDelegate(int pageNumber, ThumbnailItem item);
        public delegate void PageMovedDelegate(int pageNumber, int newLocation);
        public delegate void PageClickedDelegate(int prevPage, int newPage);
        public delegate void PageRotatedDelegate(int pageNumber);

        public delegate void ThumbnailReceivedDelegate(int pageNumber);

        public event PageDeletedDelegate PageDeleted = delegate { };
        public event PageAddedDelegate PageAdded = delegate { };
        public event PageMovedDelegate PageMoved = delegate { };
        public event PageClickedDelegate PageClicked = delegate { };
        public event PageRotatedDelegate PageRotated = delegate { };

        public event ThumbnailReceivedDelegate ThumbnailReceived;

        #endregion Events

        #region Commands

        public RelayCommand RotateCommand { get; private set; }
        public RelayCommand DeleteCommand { get; private set; }
        public RelayCommand CloseCommand { get; private set; }
        public RelayCommand SelectionChangedCommand { get; private set; }
        public RelayCommand ClearSelectionCommand { get; private set; }
        public RelayCommand ItemClickCommand { get; private set; }
        public RelayCommand EditCommand { get; private set; }
        public RelayCommand DoneEditingCommand { get; private set; }
        public RelayCommand ExportSelectedPagesCommand { get; private set; }
        public RelayCommand AddPageCommand { get; private set; }
        public RelayCommand AddImageCommand { get; private set; }
        public RelayCommand AddPDFCommand { get; private set; }
        public RelayCommand InsertPageCommand { get; private set; }
        public RelayCommand InsertImageCommand { get; private set; }
        public RelayCommand InsertPDFCommand { get; private set; }

        private async void DeleteCommandImpl(object sender)
        {
            Windows.UI.Popups.MessageDialog md = new Windows.UI.Popups.MessageDialog("");
            if (_SelectedItems.Count > 1 && sender == null)
            {
                md.Title = ResourceHandler.GetString("ThumbnailsView_Delete_Multiple_Message_Title");
                md.Content = String.Format(ResourceHandler.GetString("ThumbnailsView_Delete_Multiple_Message_Content"), _SelectedItems.Count);
            }
            else
            {
                md.Title = ResourceHandler.GetString("ThumbnailsView_Delete_Single_Message_Title");
                md.Content = ResourceHandler.GetString("ThumbnailsView_Delete_Single_Message_Content");
            }
            md.Commands.Add(new UICommand(ResourceHandler.GetString("ThumbnailsView_Delete_Message_Confirm"), (command) =>
            {
                DeleteItems(sender as ThumbnailItem);

                if ((_ToolManager != null) && (_ToolManager.UndoRedoAction != null))
                {
                    _ToolManager.UndoRedoAction.TakeSnapshot("Delete Page", true);
                }
            }));

            md.Commands.Add(new UICommand(ResourceHandler.GetString("ThumbnailsView_Delete_Message_Cancel"), (command) => { }));
            await md.ShowAsync();
        }

        /// <summary>
        /// Rotates all the selected pages. Applies XAML rotation to increase responsive first.
        /// After the rotated thumbnail is fetched, the XAML rotation is reset. 
        /// </summary>
        /// <param name="sender"></param>
        private void RotateCommandImpl(object sender)
        {
            _PDFViewCtrl.CancelAllThumbRequests();
            List<int> pagesToRotate = new List<int>();
            try
            {
                _PDFViewCtrl.DocLock(true);
                if (sender != null)
                {
                    HandleRotation(sender as ThumbnailItem);
                    pagesToRotate.Add(ThumbnailsList.IndexOf(sender as ThumbnailItem) + 1);
                }
                else
                {
                    foreach (ThumbnailItem item in _SelectedItems)
                    {
                        ThumbnailItem thumbItem = item as ThumbnailItem;
                        HandleRotation(thumbItem);
                        pagesToRotate.Add(ThumbnailsList.IndexOf(item) + 1);
                    }
                }

                // Call this to make the rotate take effect in the viewer page immediately
                _PDFViewCtrl.UpdatePageLayout();

                if ((_ToolManager != null) && (_ToolManager.UndoRedoAction != null))
                {
                    _ToolManager.UndoRedoAction.TakeSnapshot("Rotate Page", true);
                }
            }
            catch (Exception)
            {
                // send analytics
            }
            finally
            {
                _PDFViewCtrl.DocUnlock();
            }

            foreach (int pageNum in pagesToRotate)
            {
                PageDeleted(pageNum);
            }

            ThumbnailsList.RequestResources();
        }

        /// <summary>
        /// This method handles the XAML rotation of a thumbnailItem to increase responsiveness
        /// when rotating a thumbnail, rather than relying on the result of the Thumbnail_Requested event
        /// </summary>
        /// <param name="thumbItem"></param>
        private void HandleRotation(ThumbnailItem thumbItem)
        {
            Page pageToRotate = _PDFViewCtrl.GetDoc().GetPage(thumbItem.PageNumber);
            int pageRotation = ((int)pageToRotate.GetRotation() + 1) % 4;

            thumbItem.Angle = thumbItem.Angle == 270 ? 0 : thumbItem.Angle + 90;
            pageToRotate.SetRotation((PageRotate)pageRotation);

            UpdateScaleFromRotation(thumbItem);
        }

        private void CloseCommandImpl(object sender)
        {
            if (IsInEditMode)
            {
                SelectionMode = ListViewSelectionMode.None;
                IsSelectButtonVisible = true;
                return;
            }

            if (_PDFViewCtrl != null)
            {
                if (_HasModifiedPageLayout)
                {
                    try
                    {
                        _PDFViewCtrl.UpdatePageLayout();
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                }
                _PDFViewCtrl.RequestRendering();
                if (CurrentItem != null && CurrentItem.IsCurrentPage)
                {
                    int index = ThumbnailsList.IndexOf(CurrentItem);
                    _PDFViewCtrl.SetCurrentPage(index + 1);
                }
            }
            _View.CloseControl();
            CleanUp();
        }

        private void SelectionChangedCommandImpl(object changeArgs)
        {
            SelectionChangedEventArgs args = changeArgs as SelectionChangedEventArgs;
            if (args != null)
            {
                ResolveSelection(args);
            }
            if (_ClearingSelection)
            {
                return;
            }
        }

        private void ClearSelectionCommandImpl(object sender)
        {
            SelectionMode = ListViewSelectionMode.None;
        }

        private void ItemClickCommandImpl(object args)
        {
            ItemClickEventArgs eargs = args as ItemClickEventArgs;
            ThumbnailItem clickedItem = eargs.ClickedItem as ThumbnailItem;
            int index = ThumbnailsList.IndexOf(clickedItem);
            if (!ReadOnly)
            {
                _PDFViewCtrl.CancelAllThumbRequests();
                try
                {
                    _PDFViewCtrl.UpdatePageLayout();
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            PageClicked?.Invoke(_PDFViewCtrl.GetCurrentPage(), index + 1);
            if (NavigationStack != null)
            {
                NavigationStack.SetCurrentPage(index + 1);
            }
            else
            {
                _PDFViewCtrl.SetCurrentPage(index + 1);
            }

            _PDFViewCtrl.RequestRendering();

            if (!ReadOnly)
            {
                CleanUp();
                _View.CloseControl();
            }
        }

        private void EditCommandImpl(object sender)
        {
            SelectionMode = ListViewSelectionMode.Multiple;
            IsSelectButtonVisible = false;
        }

        private void DoneEditingCommandImpl(object sender)
        {
            SelectionMode = ListViewSelectionMode.None;
            IsSelectButtonVisible = true;
        }

        private async void ExportSelectedPagesCommandImpl(object sender)
        {
            await SaveSelectedPagesFromPicker(sender as ThumbnailItem);
        }

        private void AddPageCommandImpl(object sender)
        {
            AddNewPage(NewPageType.Blank);
        }

        private void AddImageCommandImpl(object sender)
        {
            AddNewPage(NewPageType.Image);
        }
        private void AddPDFCommandImpl(object sender)
        {
            AddNewPage(NewPageType.ExistingPDF);
        }
        private void InsertPageCommandImpl(object sender)
        {
            AddNewPage(NewPageType.Blank, false);
        }

        private void InsertImageCommandImpl(object sender)
        {
            AddNewPage(NewPageType.Image, false);
        }

        private void InsertPDFCommandImpl(object sender)
        {
            AddNewPage(NewPageType.ExistingPDF, false);
        }
        #endregion Commands

        #region Public Interface

        private bool _Pause = false;
        /// <summary>
        /// Gets or sets whether this ViewModel is paused.
        /// Settings the ViewModel to paused will not stop any current requests, but 
        /// will prevent it from generating new ones.
        /// </summary>
        public bool Pause
        {
            get { return _Pause; }
            set
            {
                if (Set(ref _Pause, value))
                {
                    if (!_Pause)
                    {
                        ThumbnailsList.Refresh();
                    }
                }
            }
        }

        public bool GoBack()
        {
            if (IsInEditMode)
            {
                SelectionMode = ListViewSelectionMode.None;
                return true;
            }
            _View.CloseControl();
            return true;
        }

        public void Refresh()
        {
            foreach (int index in ThumbnailsList.ItemsWithResource.Keys)
            {
                ThumbnailsList[index].Thumbnail = null;
            }
            ThumbnailsList.Refresh();
        }

        public void PageModified(int pageNumber)
        {
            int index = pageNumber - 1;
            if (ThumbnailsList.Count > index && ThumbnailsList[index].HasThumbnail)
            {
                ThumbnailsList.RefreshResource(index);
            }
        }

        #endregion Public Interface

        #region Manage View
        public class ThumbInfo : Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>.IResourceHolder
        {
            public int PageNumber;
            public int W;
            public int H;
            public int ByteSize { get; private set; }

            public double TotalSize { get { return ByteSize; } }

            public ThumbnailItem SourceItem;

            public ThumbInfo(int pageNumber, int w, int h, ThumbnailItem item)
            {
                PageNumber = pageNumber;
                W = w;
                H = h;
                ByteSize = W * H * 4;
                SourceItem = item;
            }
        }


        private void UpdatePageCount()
        {
            for (int i = 0; i < ThumbnailsList.Count; i++)
            {
                ThumbnailsList[i].PageNumber = i + 1;
            }
            List<ThumbInfo> newInfoList = new List<ThumbInfo>();
            foreach (ThumbInfo info in ThumbnailsList.ItemsWithResource.Values)
            {
                newInfoList.Add(info);
            }
            ThumbnailsList.ItemsWithResource.Clear();
            foreach (ThumbInfo info in newInfoList)
            {
                if (!info.SourceItem.IsDeleted)
                {
                    ThumbnailsList.ItemsWithResource.TryAdd(info.SourceItem.PageNumber, info);
                }
            }
        }

        private void ThumbnailsList_ResourceNeededForIndex(object sender, int index)
        {
            if ((_ThumbHandler == null) || (_PDFViewCtrl.GetDoc() == null))
                return;
            
            if (_PDFViewCtrl.IsConverting && _PDFViewCtrl.ConvertedPages <= 0)
                return;

            try
            {
                _ThumbHandler.GetThumbAsync(index + 1);
            }
            catch (Exception) { }
        }

        private void ThumbnailsList_DiscardResourceForIndex(object sender, int index)
        {
            ThumbnailsList[index].Thumbnail = null;
            ThumbnailsList[index].Angle = 0;
        }

        private void ThumbnailsList_AllRequestsCancelled(object sender, object e)
        {
            if (_ThumbHandler == null)
            {
                return;
            }
            if (_PDFViewCtrl.IsConverting && _PDFViewCtrl.ConvertedPages <= 0)
            {
                return;
            }
            _PDFViewCtrl.CancelAllThumbRequests();
            try
            {
                _UnloadingMutex.Wait();
                _UnloadingJobsPending.Clear();
            }
            catch (Exception) { }
            finally
            {
                _UnloadingMutex.Release();
            }
        }

        private bool _UnloadingThumbs = false;
        private System.Threading.SemaphoreSlim _UnloadingMutex = new System.Threading.SemaphoreSlim(1);
        private SortedDictionary<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> _UnloadingJobsPending = new SortedDictionary<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>>();

        private KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> FindFirstVisibleJob()
        {
            KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> job = _UnloadingJobsPending.Last();
            if (ThumbnailsList.Direction == Resources.ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>.Directions.Forwards)
            {
                int firstVisiblePageNumber = ThumbnailsList.FirstVisibleIndex + 1;
                if (job.Key > firstVisiblePageNumber)
                {
                    job = _UnloadingJobsPending.First(kvp => kvp.Key >= firstVisiblePageNumber);
                }
            }
            else
            {
                job = _UnloadingJobsPending.First();
                int lastVisiblePageNumber = ThumbnailsList.LastVisibleIndex + 1;
                if (job.Key < lastVisiblePageNumber)
                {
                    job = _UnloadingJobsPending.Last(kvp => kvp.Key <= lastVisiblePageNumber);
                }
            }
            return job;
        }

        private Tuple<ThumbInfo, InMemoryRandomAccessStream> GetNextJob()
        {
            Tuple<ThumbInfo, InMemoryRandomAccessStream> currentJob = null;
            try
            {
                _UnloadingMutex.Wait();
                if (_UnloadingJobsPending.Count > 0 && _ThumbHandler != null)
                {
                    KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> job = FindFirstVisibleJob();
                    currentJob = job.Value;
                    _UnloadingJobsPending.Remove(job.Key);
                }
                else
                {
                    _UnloadingThumbs = false;
                }
            }
            catch (Exception) { }
            finally
            {
                _UnloadingMutex.Release();
            }

            return currentJob;
        }


        private async void StartUnloadingTask()
        {
            Tuple<ThumbInfo, InMemoryRandomAccessStream> currentJob = GetNextJob();

            while (currentJob != null && _ThumbHandler != null)
            {
                int index = currentJob.Item1.PageNumber - 1;
                currentJob.Item1.SourceItem = ThumbnailsList[index];
                BitmapImage bmi = new BitmapImage();
                await bmi.SetSourceAsync(currentJob.Item2);
                if (_ThumbHandler != null)
                {
                    bool hadThumbnail = ThumbnailsList[index].Thumbnail != null;
                    ThumbnailsList[index].Thumbnail = bmi;
                    ThumbnailsList[index].BitmapSize = new Windows.Foundation.Size(bmi.PixelWidth, bmi.PixelHeight);
                    UpdateScaleFromRotation(ThumbnailsList[index]);
                    bool couldAdd = ThumbnailsList.ItemsWithResource.TryAdd(index, currentJob.Item1); // new ThumbInfo(.PageNumber, currentJob.Item1.W, currentJob.Item1.H, ThumbnailsList[index]));
                    if (!hadThumbnail && couldAdd)
                    {
                        ThumbnailReceived?.Invoke(currentJob.Item1.PageNumber);
                    }


                    int visibleItems = ThumbnailsList.LastVisibleIndex - ThumbnailsList.FirstVisibleIndex + 1;

                    await Task.Delay(1000 / visibleItems);
                }

                currentJob = GetNextJob();
            }
            if (_ThumbHandler != null)
            {
                ThumbnailsList.RequestResources();
            }
        }


        private async void ThumbHandler_OnThumbnailGenerated(Object sender, int pageNumber, bool wasThumbFound, byte[] thumb, int w, int h)
        {
            int index = pageNumber - 1;
            InMemoryRandomAccessStream iras = null;
            if (wasThumbFound)
            {
                try
                {
                    iras = await UtilityFunctions.ConvertToPngStreamAsync(thumb, w, h).ConfigureAwait(false);
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    wasThumbFound = false;
                }
            }

            if (wasThumbFound)
            {
                wasThumbFound = false;
                try
                {
                    _UnloadingMutex.Wait();
                    if (!_UnloadingJobsPending.ContainsKey(pageNumber))
                    {
                        ThumbInfo thumbInfo = new ThumbInfo(pageNumber, w, h, null);
                        _UnloadingJobsPending.Add(pageNumber, new Tuple<ThumbInfo, InMemoryRandomAccessStream>(thumbInfo, iras));
                    }
                    if (!_UnloadingThumbs)
                    {
                        _UnloadingThumbs = true;
                        wasThumbFound = true;
                    }
                }
                catch (Exception) { }
                finally
                {
                    _UnloadingMutex.Release();
                }
            }

            if (wasThumbFound)
            {
                await _PDFViewCtrl.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
                {
                    StartUnloadingTask();
                });
            }
        }

        private void PDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            if (ThumbnailsList != null && current_page > 0 && current_page <= ThumbnailsList.Count)
            {
                CurrentItem = ThumbnailsList[current_page - 1];
            }
        }

        private void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            bool update = false;
            if (type == PDFViewCtrlConversionType.e_conversion_finished)
            {
                update = true;
                RaisePropertyChanged("IsConverting");
            }
            else if (type == PDFViewCtrlConversionType.e_conversion_progress && DateTime.Now - _LastConversionEvent > TimeSpan.FromSeconds(1))
            {
                update = true;
                GetBookmarks(_PDFViewCtrl.GetDoc());
            }
            if (update)
            {
                _LastConversionEvent = DateTime.Now;
                int totalPages = _PDFViewCtrl.GetPageCount();
                for (int i = ThumbnailsList.Count; i < totalPages; ++i)
                {
                    ThumbnailItem item = new ThumbnailItem(i + 1, _PageDefaultColor);
                    ThumbnailsList.Add(item);
                }
            }
        }

        private void UpdateScaleFromRotation(ThumbnailItem item)
        {
            if (!item.HasThumbnail || ImageMaxDimensions.IsEmpty || ImageMaxDimensions.Width == 0 || ImageMaxDimensions.Height == 0
                || item.BitmapSize.IsEmpty || item.BitmapSize.Width == 0 || item.BitmapSize.Height == 0)
            {
                item.UniformScale = 1.0;
                return;
            }

            if (item.Angle == 0 || item.Angle == 180)
            {
                item.UniformScale = 1.0;
                return;
            }

            double widthRatio = ImageMaxDimensions.Width / item.BitmapSize.Width;
            double heightRatio = ImageMaxDimensions.Height / item.BitmapSize.Height;

            double defaultScale = Math.Min(widthRatio, heightRatio);
            double visualImageWidth = defaultScale * item.BitmapSize.Width;
            double visualImageHeight = defaultScale * item.BitmapSize.Height;

            double rotatedWidthGrowth = ImageMaxDimensions.Width / visualImageHeight;
            double rotatedHeightGrowth = ImageMaxDimensions.Height / visualImageWidth;

            item.UniformScale = Math.Min(rotatedWidthGrowth, rotatedHeightGrowth);
        }

        #endregion Manage View


        #region Selection

        private List<ThumbnailItem> _SelectedItems = new List<ThumbnailItem>();

        private void ResolveSelection(SelectionChangedEventArgs args)
        {
            int oldSelectionCount = _SelectedItems.Count;
            foreach (ThumbnailItem item in args.RemovedItems)
            {
                if (_SelectedItems.Contains(item))
                {
                    _SelectedItems.Remove(item);
                }
            }
            foreach (ThumbnailItem item in args.AddedItems)
            {
                if (!_SelectedItems.Contains(item))
                {
                    _SelectedItems.Add(item);
                }
            }

            NumberOfSelectedItems = _SelectedItems.Count;
            if (NumberOfSelectedItems == 0)
            {
                DoneEditingCommandImpl(null);
            }
        }

        #endregion Selection


        #region Editing Items

        object _ReorderItem;
        int _ReorderIndexFrom;

        void _ThumbnailsList_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case System.Collections.Specialized.NotifyCollectionChangedAction.Remove:
                    _ReorderItem = e.OldItems[0];
                    _ReorderIndexFrom = e.OldStartingIndex;
                    break;
                case System.Collections.Specialized.NotifyCollectionChangedAction.Add:
                    if (_ReorderItem == null)
                    {
                        return;
                    }
                    HandleReorder(_ReorderItem, _ReorderIndexFrom, e.NewStartingIndex);
                    _ReorderItem = null;
                    break;
            }
        }

        void HandleReorder(object item, int indexFrom, int indexTo)
        {
            _HasModifiedPageLayout = true;
            _PDFViewCtrl.CancelAllThumbRequests();
            ThumbnailItem thumbItem = item as ThumbnailItem;
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            try
            {
                _PDFViewCtrl.CancelRendering();
                doc.Lock();

                Page pageToMove = doc.GetPage(indexFrom + 1);
                int oldNum = pageToMove.GetSDFObj().GetObjNum();
                if (indexFrom < indexTo)
                {
                    PageIterator moveTo = doc.GetPageIterator(indexTo + 2);
                    doc.PageInsert(moveTo, pageToMove);
                    PageIterator createdPage = doc.GetPageIterator(indexTo + 2);
                    if (_GotBookmarksFromDocument)
                    {
                        UserBookmarkManager.MoveBookmarkDestination(doc, pageToMove, createdPage.Current());
                    }
                    PageIterator removeIter = doc.GetPageIterator(indexFrom + 1);
                    doc.PageRemove(removeIter);

                    Page newPage = doc.GetPage(indexTo + 1);
                    int newNum = newPage.GetSDFObj().GetObjNum();
                    if (!_GotBookmarksFromDocument)
                    {
                        UserBookmarkManager.PageMoved(_DocumentTag, oldNum, newNum, indexFrom + 1, indexTo + 1);
                    }
                }
                else
                {
                    PageIterator moveTo = doc.GetPageIterator(indexTo + 1);
                    doc.PageInsert(moveTo, pageToMove);
                    PageIterator createdPage = doc.GetPageIterator(indexTo + 2);
                    if (_GotBookmarksFromDocument)
                    {
                        UserBookmarkManager.MoveBookmarkDestination(doc, pageToMove, createdPage.Current());
                    }
                    PageIterator removeIter = doc.GetPageIterator(indexFrom + 2);
                    doc.PageRemove(removeIter);

                    Page newPage = doc.GetPage(indexTo + 1);
                    int newNum = newPage.GetSDFObj().GetObjNum();
                    if (_GotBookmarksFromDocument)
                    {
                        UserBookmarkManager.PageMoved(_DocumentTag, oldNum, newNum, indexFrom + 1, indexTo + 1);
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.Unlock();
            }
            _PDFViewCtrl.UpdatePageLayout();
            ThumbnailsList.RequestResources();
            PageMoved(indexFrom + 1, indexTo + 1);
            UpdatePageCount();
        }

        Windows.Foundation.IAsyncOperation<Windows.UI.Popups.IUICommand> _DeleteErrorMessage;
        private async void DeleteItems(ThumbnailItem thumbItem = null)
        {
            if (NumberOfSelectedItems == ThumbnailsList.Count && thumbItem == null)
            {
                if (_DeleteErrorMessage == null)
                {
                    Windows.UI.Popups.MessageDialog deleteError = new Windows.UI.Popups.MessageDialog(ResourceHandler.GetString("ThumbnailsView_Delete_Error_Title"), ResourceHandler.GetString("ThumbnailsView_Delete_Error_Info"));

                    _DeleteErrorMessage = deleteError.ShowAsync();
                    try
                    {
                        await _DeleteErrorMessage;
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    _DeleteErrorMessage = null;
                }
                return;
            }

            _HasModifiedPageLayout = true;

            _PDFViewCtrl.CancelAllThumbRequests();
            List<ThumbnailItem> itemsToDelete = new List<ThumbnailItem>();
            List<int> pagesToDelete = new List<int>();
            if (thumbItem != null)
            {
                itemsToDelete.Add(thumbItem);
                pagesToDelete.Add(ThumbnailsList.IndexOf(thumbItem) + 1);
            }
            else
            {
                foreach (ThumbnailItem item in _SelectedItems)
                {
                    itemsToDelete.Add(item);
                    pagesToDelete.Add(ThumbnailsList.IndexOf(item) + 1);
                }

                SelectionMode = ListViewSelectionMode.None;
            }
            ThumbnailItem currentItem = CurrentItem;
            foreach (ThumbnailItem item in itemsToDelete)
            {
                if (item == currentItem)
                {
                    currentItem = null;
                }
                item.IsDeleted = true;
                ThumbnailsList.Remove(item);
            }

            PDFDoc doc = _PDFViewCtrl.GetDoc();
            try
            {
                doc.Lock();
                pagesToDelete.Sort();
                for (int i = pagesToDelete.Count - 1; i >= 0; --i)
                {
                    PageIterator itr = doc.GetPageIterator(pagesToDelete[i]);
                    if (_GotBookmarksFromDocument)
                    {
                        UserBookmarkManager.PageDeleted(doc, itr.Current());
                    }
                    else
                    {
                        UserBookmarkManager.PageDeleted(_DocumentTag, itr.Current().GetSDFObj().GetObjNum(), pagesToDelete[i], _PDFViewCtrl.GetPageCount());
                    }
                    doc.PageRemove(itr);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.Unlock();
            }
            _PDFViewCtrl.UpdatePageLayout();
            if (currentItem != null && ThumbnailsList.Contains(currentItem))
            {
                int index = ThumbnailsList.IndexOf(currentItem);
                _PDFViewCtrl.SetCurrentPage(index + 1);
                CurrentItem = currentItem;
            }

            ThumbnailsList.RequestResources();
            foreach (int pageNum in pagesToDelete)
            {
                PageDeleted(pageNum);
            }

            // Update all page numbers after removing pages
            UpdatePageCount();
        }

        /// <summary>
        /// Saves all the selected PDF pages into a new PDF file 
        /// </summary>
        /// <param name="thumbItem"></param>
        /// <returns></returns>
        private async Task SaveSelectedPagesFromPicker(ThumbnailItem thumbItem = null)
        {
            FileSavePicker fileSavePicker = new FileSavePicker();
            fileSavePicker.FileTypeChoices.Add("PDF", Valid_PDF_Types);
            fileSavePicker.SuggestedFileName = _PDFViewCtrl.GetDoc().GetDocInfo().GetTitle();
            fileSavePicker.SuggestedStartLocation = Windows.Storage.Pickers.PickerLocationId.DocumentsLibrary;

            StorageFile pickedFile = await fileSavePicker.PickSaveFileAsync();
            if (pickedFile != null)
            {
                PDFDoc doc = new PDFDoc();
                if (thumbItem != null)
                {
                    PageIterator pageIterator = doc.GetPageIterator(thumbItem.PageNumber);
                    doc.PageInsert(pageIterator, _PDFViewCtrl.GetDoc().GetPage(thumbItem.PageNumber));
                }
                else
                {
                    var sortedSelectedItems = _SelectedItems.OrderBy(x => (x as ThumbnailItem).PageNumber);
                    foreach (ThumbnailItem item in sortedSelectedItems)
                    {
                        PageIterator pageIterator = doc.GetPageIterator(item.PageNumber);
                        doc.PageInsert(pageIterator, _PDFViewCtrl.GetDoc().GetPage(item.PageNumber));
                    }
                }
                await doc.SaveToNewLocationAsync(pickedFile, SDF.SDFDocSaveOptions.e_compatibility);
            }
        }

        private async Task<StorageFile> GetOpenFileFromPicker(IList<string> fileTypes)
        {
            StorageFile theFile = null;

            Windows.ApplicationModel.Resources.ResourceLoader loader = Windows.ApplicationModel.Resources.ResourceLoader.GetForCurrentView();

            FileOpenPicker fileOpenPicker = new FileOpenPicker();
            fileOpenPicker.ViewMode = PickerViewMode.Thumbnail;
            if (fileTypes == Valid_Image_Types)
            {
                fileOpenPicker.CommitButtonText = loader.GetString("DocumentCreation_ImageDocument_PickFileConform");
            }
            else
            {
                fileOpenPicker.CommitButtonText = loader.GetString("DocumentCreation_PDFDocument_PickFileConform");
            }
            foreach (string filetype in fileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(filetype);
            }

            theFile = await fileOpenPicker.PickSingleFileAsync();
            return theFile;
        }

        /// <summary>
        /// This method adds a new page (blank, image, PDF) into the current PDF
        /// </summary>
        /// <param name="PageType"></param>
        /// <param name="insertAtEnd"> Whether to insert at end. If false, insert immediately after last selected item.</param>
        private async void AddNewPage(NewPageType PageType, bool insertAtEnd = true)
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            _PDFViewCtrl.CancelAllThumbRequests();

            // Calculate where to place new pages
            int offset = insertAtEnd ? doc.GetPageCount() + 1 : _SelectedItems.Max(x => (x as ThumbnailItem).PageNumber) + 1;
            int pages = 1;

            switch (PageType)
            {
                case NewPageType.Blank:
                    InsertBlankPage(insertAtEnd);
                    break;
                case NewPageType.Image:
                    bool pickedImage = await InsertImagePage(insertAtEnd);
                    if (!pickedImage)
                    {
                        return;
                    }

                    break;
                case NewPageType.ExistingPDF:
                    pages = await InsertPDF(insertAtEnd);
                    if (pages == 0)
                    {
                        return;
                    }
                    break;
            }

            _HasModifiedPageLayout = true;
            // Add all the needed pages to thumbnail items 
            ThumbnailItem item = null;
            for (int i = 0; i < pages; i++)
            {
                item = new ThumbnailItem(offset + i, _PageDefaultColor);
                ThumbnailsList.Insert(offset + i - 1, item);
            }

            ThumbnailsList.Refresh();

            if (!insertAtEnd)
            {
                UpdatePageCount();
            }

            _PDFViewCtrl.UpdatePageLayout();
            ThumbnailsList.RequestResources();

            PageAdded(offset, item);

            if ((_ToolManager != null) && (_ToolManager.UndoRedoAction != null))
            {
                _ToolManager.UndoRedoAction.TakeSnapshot("Add new Page", true);
            }
        }

        private void InsertBlankPage(bool insertAtEnd)
        {
            bool needsUnlock = false;
            PDFDoc doc = _PDFViewCtrl.GetDoc();

            try
            {
                _PDFViewCtrl.DocLock(true);
                needsUnlock = true;
                int pageNumberToFolow = 0;
                if (insertAtEnd)
                {
                    pageNumberToFolow = doc.GetPageCount();
                }
                else
                {
                    // Insert right after the max selected page number
                    pageNumberToFolow = _SelectedItems.Max(x => (x as ThumbnailItem).PageNumber);
                }

                PageIterator pageIterator = doc.GetPageIterator(pageNumberToFolow);
                Rect mediaBox = pageIterator.Current().GetBox(PageBox.e_media);
                Rect cropBox = pageIterator.Current().GetBox(PageBox.e_crop);
                Rect userCrop = pageIterator.Current().GetBox(PageBox.e_user_crop);

                Page page = doc.PageCreate(mediaBox);
                page.SetBox(PageBox.e_crop, cropBox);
                page.SetBox(PageBox.e_user_crop, userCrop);
                page.SetRotation(pageIterator.Current().GetRotation());

                if (!insertAtEnd && pageIterator.HasNext())
                {
                    pageIterator.Next();
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (needsUnlock)
                {
                    doc.Unlock();
                }
            }
        }

        private async Task<bool> InsertImagePage(bool insertAtEnd)
        {
            if (_FilePickingInProgress)
            {
                return false;
            }

            bool needsUnlock = false;
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            Page page = doc.PageCreate();
            try
            {
                _PDFViewCtrl.DocLock(true);
                needsUnlock = true;
                _FilePickingInProgress = true;

                StorageFile file = await GetOpenFileFromPicker(Valid_Image_Types);
                if (file == null)
                    return false;

                Stream stream = await file.OpenStreamForReadAsync();
                Byte[] bytes = new byte[(int)stream.Length];
                stream.Read(bytes, 0, (int)stream.Length);

                Image img = Image.Create(doc.GetSDFDoc(), bytes);
                ElementBuilder f = new ElementBuilder();
                Element element = f.CreateImage(img, new Matrix2D(img.GetImageWidth(), 0, 0, img.GetImageHeight(), 0, 0));

                page.SetMediaBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));
                page.SetCropBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));

                ElementWriter writer = new ElementWriter();
                writer.Begin(page);
                writer.WritePlacedElement(element);
                writer.End();

                if (!insertAtEnd)
                {
                    // Insert right after the max selected page number
                    int offset = _SelectedItems.Max(x => (x as ThumbnailItem).PageNumber) + 1;
                    PageIterator pageIterator = doc.GetPageIterator(offset);
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _FilePickingInProgress = false;
                if (needsUnlock)
                {
                    doc.Unlock();
                }
            }

            return true;
        }

        private async Task<int> InsertPDF(bool insertAtEnd)
        {
            if (_FilePickingInProgress)
            {
                return 0;
            }

            bool needsUnlock = false;
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            int pagesAdded = 0;

            try
            {
                _PDFViewCtrl.DocLock(true);
                needsUnlock = true;
                _FilePickingInProgress = true;

                StorageFile file = await GetOpenFileFromPicker(Valid_PDF_Types);
                if (file == null)
                    return 0;

                PDFDoc docTemp = new PDFDoc(file);
                pagesAdded = docTemp.GetPageCount();
                // Check where to place new pages 
                int offset = insertAtEnd ? doc.GetPageCount() + 1 : _SelectedItems.Max(x => (x as ThumbnailItem).PageNumber) + 1;
                doc.InsertPages(offset, docTemp, 1, pagesAdded, PDFDocInsertFlag.e_none);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _FilePickingInProgress = false;
                if (needsUnlock)
                {
                    doc.Unlock();
                }
            }

            return pagesAdded;
        }
        #endregion Editing Items
    }
}