﻿using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.SDF;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Serialization;
using Windows.Foundation;
using Windows.Storage;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class UserBookmarkManager
    {
        public class KeyValuePair // We're using this since using the built in KeyValuePair<string, string> seems to serialize to null values
        {
            public string Key { get; set; }
            public string Value { get; set; }

            public KeyValuePair() { }

            public KeyValuePair(string key, string value)
            {
                Key = key;
                Value = value;
            }
        }



        private const string USER_BOOKMARK_SETTING_KEY_STRING_PREFIX = "User_Bookmarks_";
        private const string USER_BOOKMARK_FILE_NAME = "UserBookmarkMap";
        private const string USER_BOOKMARK_OBJ_TITLE = "pdftronUserBookmarks";

        private static List<UserBookmarksViewModel.UserBookmarkItem> _CurrentBookmarkList = null;
        private static Dictionary<string, string> _FileDictionary = null;
        private static StorageFolder _BookmarksFolder = null;

        #region Embedded user bookmarks

        /// <summary>
        /// Returns true if the PDFDoc has use bookmarks embedded
        /// </summary>
        /// <param name="doc"></param>
        /// <returns>true if the PDFDoc has embedded bookmarks, false otherwise</returns>
        public static bool DoesDocHaveBookmarks(PDFDoc doc)
        {
            try
            {
                doc.LockRead();
                Bookmark bookmark = GetRootPDFBookmark(doc, false);
                if (bookmark != null)
                {
                    return true;
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.UnlockRead();
            }

            return false;
        }

        /// <summary>
        /// Gets the root bookmarks for user bookmarks
        /// </summary>
        /// <param name="pdfDoc"></param>
        /// <param name="createNew">whether or not to create a root bookmarks if non exists</param>
        /// <returns>the created bookmark</returns>
        public static Bookmark GetRootPDFBookmark(PDFDoc pdfDoc, bool createNew)
        {
            Bookmark bookmark = null;
            if (null != pdfDoc)
            {
                try
                {
                    Obj catalog = pdfDoc.GetRoot();
                    Obj bookmarkObj = catalog.FindObj(USER_BOOKMARK_OBJ_TITLE);
                    if (null != bookmarkObj)
                    {
                        // found existing bookmark obj
                        bookmark = new Bookmark(bookmarkObj);
                    }
                    else
                    {
                        if (createNew)
                        {
                            // create new bookmark obj
                            bookmark = Bookmark.Create(pdfDoc, USER_BOOKMARK_OBJ_TITLE);
                            pdfDoc.GetRoot().Put(USER_BOOKMARK_OBJ_TITLE, bookmark.GetSDFObj());
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    bookmark = null;
                }
            }
            return bookmark;
        }

        public static IAsyncOperation<List<UserBookmarksViewModel.UserBookmarkItem>> GetBookmarkListAsync(PDFDoc doc)
        {
            Task<List<UserBookmarksViewModel.UserBookmarkItem>> t = new Task<List<UserBookmarksViewModel.UserBookmarkItem>>(() =>
            {
                List<UserBookmarksViewModel.UserBookmarkItem> items = GetBookmarkList(doc);
                return items;
            });
            t.Start();
            return t.AsAsyncOperation<List<UserBookmarksViewModel.UserBookmarkItem>>();
        }

        public static async Task LoadBookmarksListAsync(PDFDoc doc)
        {
            _CurrentBookmarkList = await GetBookmarkListAsync(doc);
        }

        public static List<UserBookmarksViewModel.UserBookmarkItem> GetBookmarkList(PDFDoc doc)
        {
            List<UserBookmarksViewModel.UserBookmarkItem> data = new List<UserBookmarksViewModel.UserBookmarkItem>();
            if (doc != null)
            {
                try
                {
                    doc.LockRead();
                    Bookmark rootBookmark = GetRootPDFBookmark(doc, false);
                    Bookmark item = rootBookmark.GetFirstChild();
                    while (item.IsValid())
                    {
                        UserBookmarksViewModel.UserBookmarkItem bookmarkItem = new UserBookmarksViewModel.UserBookmarkItem();
                        bookmarkItem.BookmarkName = item.GetTitle();
                        bookmarkItem.OriginalName = bookmarkItem.BookmarkName;
                        Action action = item.GetAction();
                        if (action != null && action.IsValid())
                        {
                            if (action.GetType() == ActionType.e_GoTo)
                            {
                                Destination dest = action.GetDest();
                                if (dest != null && dest.IsValid())
                                {
                                    bookmarkItem.PageNumber = dest.GetPage().GetIndex();
                                    bookmarkItem.ObjectNumber = dest.GetPage().GetSDFObj().GetObjNum();
                                    bookmarkItem.PDFBookmark = item;
                                    data.Add(bookmarkItem);
                                }
                            }
                        }
                        bookmarkItem.IsBookmarkEdited = false;
                        item = item.GetNext();
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.UnlockRead();
                }
            }

            return data;
        }

        public static IAsyncAction SaveBookmarkListAsync(PDFDoc doc, List<UserBookmarksViewModel.UserBookmarkItem> items)
        {
            Task t = new Task(() =>
            {
                SaveBookmarkList(doc, items);
            });
            t.Start();
            return t.AsAsyncAction();
        }



        public static void SaveBookmarkList(PDFDoc doc, List<UserBookmarksViewModel.UserBookmarkItem> items)
        {
            if (items == null)
            {
                items = _CurrentBookmarkList;
            }

            if (items == null)
            {
                return;
            }

            try
            {
                System.Diagnostics.Debug.WriteLine("Saving bookmarks to doc");
                doc.Lock();

                if (items.Count > 0)
                {
                    Bookmark rootBookmark = GetRootPDFBookmark(doc, true);
                    Bookmark firstBookmark = null;
                    if (rootBookmark.HasChildren())
                    {
                        firstBookmark = rootBookmark.GetFirstChild();
                    }
                    Bookmark currentBookmark = null;
                    if (rootBookmark != null)
                    {
                        foreach (UserBookmarksViewModel.UserBookmarkItem item in items)
                        {
                            try
                            {
                                if (item.PDFBookmark == null)
                                {
                                    if (currentBookmark == null) // No items in the list above this on are currently in the document
                                    {
                                        if (firstBookmark == null) // this means there are no bookmarks at all, so create one.
                                        {
                                            currentBookmark = rootBookmark.AddChild(item.BookmarkName);
                                            currentBookmark.SetAction(Action.CreateGoto(Destination.CreateFit(doc.GetPage(item.PageNumber))));
                                            firstBookmark = currentBookmark;
                                        }
                                        else // there already are bookmarks, so the new bookamrk needs to be inserted in front of the first one
                                        {
                                            currentBookmark = firstBookmark.AddPrev(item.BookmarkName);
                                            currentBookmark.SetAction(Action.CreateGoto(Destination.CreateFit(doc.GetPage(item.PageNumber))));
                                            firstBookmark = currentBookmark;
                                        }
                                    }
                                    else // at least one item in the list above the current item was in the list
                                    {
                                        currentBookmark = currentBookmark.AddNext(item.BookmarkName);
                                        currentBookmark.SetAction(Action.CreateGoto(Destination.CreateFit(doc.GetPage(item.PageNumber))));
                                    }
                                    item.PDFBookmark = currentBookmark;
                                }
                                else
                                {
                                    currentBookmark = item.PDFBookmark;
                                    if (item.IsBookmarkEdited)
                                    {
                                        Action action = item.PDFBookmark.GetAction();
                                        Destination dest = action.GetDest();
                                        dest.SetPage(doc.GetPage(item.PageNumber));
                                        item.PDFBookmark.SetTitle(item.BookmarkName);
                                    }
                                }
                            }
                            catch (Exception ex)
                            {
                                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                            }
                        }
                    }
                }
                else
                {
                    RemoveRootPDFBookmark(doc);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.Unlock();
            }
        }

        public static bool RemoveRootPDFBookmark(PDFDoc doc)
        {
            bool success = true;
            try
            {
                System.Diagnostics.Debug.WriteLine("Removing root PDF bookmark");
                
                doc.Lock();

                Obj catalog = doc.GetRoot();
                Obj bookmarkObj = catalog.FindObj(USER_BOOKMARK_OBJ_TITLE);
                if (bookmarkObj != null)
                {
                    catalog.Erase(USER_BOOKMARK_OBJ_TITLE);
                }
            }
            catch (Exception ex) 
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                success = false; 
            }
            finally
            {
                doc.Unlock();
            }
            return success;
        }

        /// <summary>
        /// Adds a bookmark to a PDFDoc, excepts that the Document is already locked for writing
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="rootBookmark"></param>
        /// <param name="item"></param>
        private static void AddChildPDFBookmark(PDFDoc doc, Bookmark rootBookmark, UserBookmarksViewModel.UserBookmarkItem item)
        {
            if (rootBookmark != null)
            {
                try
                {
                    Bookmark child = rootBookmark.AddChild(item.BookmarkName);
                    child.SetAction(Action.CreateGoto(Destination.CreateFit(doc.GetPage(item.PageNumber))));
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        #endregion Embedded user bookmarks



        public static async Task<List<UserBookmarksViewModel.UserBookmarkItem>> GetBookmarkListAsync(string docTag)
        {
            await LoadBookmarksListAsync(docTag);
            return _CurrentBookmarkList;
        }

        public static async Task LoadBookmarksListAsync(string docTag)
        {
            _CurrentBookmarkList = null;
            if (_FileDictionary == null)
            {
                try
                {
                    _BookmarksFolder = await ApplicationData.Current.LocalFolder.CreateFolderAsync("UserBookmarks", CreationCollisionOption.OpenIfExists);
                    if (_BookmarksFolder != null)
                    {
                        StorageFile file = await _BookmarksFolder.CreateFileAsync(USER_BOOKMARK_FILE_NAME, CreationCollisionOption.OpenIfExists);
                        if (file != null)
                        {
                            string dictString = await FileIO.ReadTextAsync(file);
                            _FileDictionary = DeSerializeFileDict(dictString);
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }

            if (_FileDictionary == null)
            {
                return;
            }

            if (!string.IsNullOrWhiteSpace(docTag))
            {
                if (_FileDictionary.ContainsKey(docTag))
                {
                    string bookmarkFileName = _FileDictionary[docTag];
#if WINDOWS_PHONE_APP
                    IStorageItem fileItem = null;
                    try
                    {
                        fileItem = await _BookmarksFolder.GetItemAsync(bookmarkFileName);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
#else
                    IStorageItem fileItem = await _BookmarksFolder.TryGetItemAsync(bookmarkFileName);
#endif
                    if (fileItem != null)
                    {
                        StorageFile file = fileItem as StorageFile;
                        if (file != null)
                        {
                            string bookmarkString = await FileIO.ReadTextAsync(file);
                            if (!string.IsNullOrWhiteSpace(bookmarkString))
                            {
                                _CurrentBookmarkList = DeSerializeString(bookmarkString);
                            }
                        }
                    }
                }
            }

            return;
        }


        public static async Task SaveBookmarkListAsync(string docTag, List<UserBookmarksViewModel.UserBookmarkItem> items)
        {
            if (_FileDictionary == null)
            {
                return;
            }

            if (items == null)
            {
                items = _CurrentBookmarkList;
            }

            if (items == null)
            { 
                return;
            }

            if (!string.IsNullOrWhiteSpace(docTag) && items != null)
            {
                bool shouldSaveFiledict = false;
                if (!_FileDictionary.ContainsKey(docTag))
                {
                    Guid guid = Guid.NewGuid();
                    string fileName = guid.ToString();
                    _FileDictionary[docTag] = fileName;
                    shouldSaveFiledict = true;
                }
                try
                {
                    StorageFile bookmarkFile = await _BookmarksFolder.CreateFileAsync(_FileDictionary[docTag], CreationCollisionOption.OpenIfExists);
                    string bookmarkString = SerializeList(items);
                    await FileIO.WriteTextAsync(bookmarkFile, bookmarkString);
                    if (shouldSaveFiledict)
                    {
                        StorageFile file = await _BookmarksFolder.GetFileAsync(USER_BOOKMARK_FILE_NAME);
                        string dictString = SerializeFileDictionary(_FileDictionary);
                        await FileIO.WriteTextAsync(file, dictString);
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        public static async Task EraseBookmarkListAsync(string docTag)
        {
            if (!string.IsNullOrWhiteSpace(docTag))
            {
                if (_FileDictionary.ContainsKey(docTag))
                {
                    try
                    {
                        _FileDictionary.Remove(docTag);
                        StorageFile file = await _BookmarksFolder.GetFileAsync(USER_BOOKMARK_FILE_NAME);
                        string dictString = SerializeFileDictionary(_FileDictionary);
                        await FileIO.WriteTextAsync(file, dictString);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                }
            }
        }

        public static bool CanGetKeyFromDoc(PDFDoc doc)
        {
            return !string.IsNullOrWhiteSpace(GetKeyFromDoc(doc));
        }

        private static string GetKeyFromDoc(PDFDoc doc)
        {
            string docName = doc.GetFileName();
            if (string.IsNullOrWhiteSpace(docName))
            {
                return "";
            }
            return USER_BOOKMARK_SETTING_KEY_STRING_PREFIX + docName;
        }

        /// <summary>
        /// Indicates that the old page, previously being at oldPageNumber with object number objNumber,
        /// now is at page newPageNumber and has new object number newObjNumber
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="objNumber"></param>
        /// <param name="newObjNumber"></param>
        /// <param name="oldPageNumber"></param>
        /// <param name="newPageNumber"></param>
        public static void PageMoved(string docTag, int objNumber, int newObjNumber, int oldPageNumber, int newPageNumber)
        {
            if (_CurrentBookmarkList == null)
            {
                return;
            }
            if (!string.IsNullOrWhiteSpace(docTag))
            {
                if (_CurrentBookmarkList.Count > 0)
                {
                    foreach (UserBookmarksViewModel.UserBookmarkItem item in _CurrentBookmarkList)
                    {
                        if (item.ObjectNumber == objNumber)
                        {
                            item.ObjectNumber = newObjNumber;
                            item.PageNumber = newPageNumber;
                        }
                    }
                    if (oldPageNumber < newPageNumber)
                    {
                        UpdateBookmarksAfterRearranging(docTag, _CurrentBookmarkList, oldPageNumber + 1, newPageNumber, false, newObjNumber);
                    }
                    else
                    {
                        UpdateBookmarksAfterRearranging(docTag, _CurrentBookmarkList, newPageNumber, oldPageNumber - 1, true, newObjNumber);
                    }
                }
            }
        }

        /// <summary>
        /// Page pageNumber with object number objNumber has been deleted.
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="objNumber"></param>
        /// <param name="pageNumber"></param>
        public static void PageDeleted(string docTag, int objNumber, int pageNumber, int pageCount)
        {
            if (_CurrentBookmarkList == null)
            {
                return;
            }
            if (!string.IsNullOrWhiteSpace(docTag))
            {
                if (_CurrentBookmarkList.Count > 0)
                {
                    List<UserBookmarksViewModel.UserBookmarkItem> newBookmarkList = new List<UserBookmarksViewModel.UserBookmarkItem>();
                    foreach (UserBookmarksViewModel.UserBookmarkItem item in _CurrentBookmarkList)
                    {
                        if (item.ObjectNumber != objNumber)
                        {
                            newBookmarkList.Add(item);
                        }
                    }
                    UpdateBookmarksAfterRearranging(docTag, newBookmarkList, pageNumber, pageCount, false);
                    _CurrentBookmarkList = newBookmarkList;
                }
            }
        }

        /// <summary>
        /// Updates all the bookmarks that pointed to oldPage to now point to newPage
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="oldPage"></param>
        /// <param name="newPage"></param>
        public static void MoveBookmarkDestination(PDFDoc doc, Page oldPage, Page newPage)
        {
            if (_CurrentBookmarkList != null && doc != null && oldPage != null && newPage != null)
            {
                try
                {
                    doc.Lock();

                    SDF.Obj oldObj = oldPage.GetSDFObj();
                    SDF.Obj newObj = newPage.GetSDFObj();

                    foreach (UserBookmarksViewModel.UserBookmarkItem item in _CurrentBookmarkList)
                    {
                        if (item.PDFBookmark != null)
                        {
                            Destination dest = item.PDFBookmark.GetAction().GetDest();
                            Page bmPage = dest.GetPage();
                            if (bmPage.GetSDFObj().IsEqual(oldObj))
                            {
                                dest.SetPage(newPage);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                }
            }
        }

        /// <summary>
        /// Deletes all bookmarks pointing to the page 
        /// Note that CurrentBookmarkList must be a list with actual PDF bookmarks. 
        /// That is, the last call to LoadBookmarksListAsync should have been with a PDFDoc
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="page"></param>
        public static void PageDeleted(PDFDoc doc, Page page)
        {
            if (_CurrentBookmarkList != null && doc != null && page != null)
            {
                try
                {
                    doc.Lock();
                    List<UserBookmarksViewModel.UserBookmarkItem> bookmarksToDelete = new List<UserBookmarksViewModel.UserBookmarkItem>();
                    foreach (UserBookmarksViewModel.UserBookmarkItem item in _CurrentBookmarkList)
                    {
                        if (item.PDFBookmark != null)
                        {
                            Page bmPage = item.PDFBookmark.GetAction().GetDest().GetPage();
                            if (bmPage != null && bmPage.GetSDFObj().IsEqual(page.GetSDFObj()))
                            {
                                bookmarksToDelete.Add(item);
                            }
                        }
                    }
                    foreach (UserBookmarksViewModel.UserBookmarkItem item in bookmarksToDelete)
                    {
                        item.Delete();
                        _CurrentBookmarkList.Remove(item);
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    doc.Unlock();
                }
            }
        }

        /// <summary>
        /// Updates all the bookmarks whose page number falls in the range fromPage to toPage. They will either be
        /// incremented by 1 or decremented by 1.
        /// 
        /// fromPage and toPage are inclusive.
        /// </summary>
        /// <param name="doc"></param>
        /// <param name="fromPage"></param>
        /// <param name="toPage"></param>
        /// <param name="increment"></param>
        public static void UpdateBookmarksAfterRearranging(string docTag, int fromPage, int toPage, bool increment, int ignoreObjNumber = -1)
        {
            if (_CurrentBookmarkList == null)
            {
                return;
            }
            if (!string.IsNullOrWhiteSpace(docTag))
            {
                if (_CurrentBookmarkList.Count > 0)
                {
                    UpdateBookmarksAfterRearranging(docTag, _CurrentBookmarkList, fromPage, toPage, increment, ignoreObjNumber);
                }
            }
        }

        private static void UpdateBookmarksAfterRearranging(string docTag, List<UserBookmarksViewModel.UserBookmarkItem> bookmarkList,
            int fromPage, int toPage, bool increment, int ignoreObjNumber = -1)
        {
            if (_CurrentBookmarkList == null)
            {
                return;
            }
            if (fromPage > toPage)
            {
                int temp = fromPage;
                fromPage = toPage;
                toPage = temp;
            }
            int change = -1;
            if (increment)
            {
                change = 1;
            }
            foreach (UserBookmarksViewModel.UserBookmarkItem item in bookmarkList)
            {
                if (item.PageNumber >= fromPage && item.PageNumber <= toPage && item.ObjectNumber != ignoreObjNumber)
                {
                    item.PageNumber += change;
                }
            }
        }


        private static string SerializeList(List<UserBookmarksViewModel.UserBookmarkItem> items)
        {
            try
            {
                XmlSerializer xmlIzer = new XmlSerializer(typeof(List<UserBookmarksViewModel.UserBookmarkItem>));
                System.IO.StringWriter writer = new System.IO.StringWriter();
                xmlIzer.Serialize(writer, items);
                return writer.ToString();
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return string.Empty;
        }

        private static List<UserBookmarksViewModel.UserBookmarkItem> DeSerializeString(string itemsString)
        {
            List<UserBookmarksViewModel.UserBookmarkItem> items = new List<UserBookmarksViewModel.UserBookmarkItem>();
            try
            {
                XmlSerializer xmlIzer = new XmlSerializer(typeof(List<UserBookmarksViewModel.UserBookmarkItem>));
                using (System.IO.StringReader sr = new System.IO.StringReader(itemsString))
                {
                    items = (xmlIzer.Deserialize(sr)) as List<UserBookmarksViewModel.UserBookmarkItem>;
                    if (items != null)
                    {
                        foreach (UserBookmarksViewModel.UserBookmarkItem item in items)
                        {
                            item.OriginalName = item.BookmarkName;
                            item.IsBookmarkEdited = false;
                        }

                        return items;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return new List<UserBookmarksViewModel.UserBookmarkItem>();
        }

        private static string SerializeFileDictionary(Dictionary<string, string> fileDict)
        {
            try
            {
                List<KeyValuePair> tuples = new List<KeyValuePair>();
                foreach (KeyValuePair<string, string> keyValPair in fileDict)
                {
                    tuples.Add(new KeyValuePair(keyValPair.Key, keyValPair.Value));
                }
                XmlSerializer xmlIzer = new XmlSerializer(typeof(List<KeyValuePair>));
                System.IO.StringWriter writer = new System.IO.StringWriter();
                xmlIzer.Serialize(writer, tuples);
                return writer.ToString();
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return string.Empty;
        }

        private static Dictionary<string, string> DeSerializeFileDict(string fileDictString)
        {
            if (!string.IsNullOrWhiteSpace(fileDictString))
            {
                Dictionary<string, string> fileDict = new Dictionary<string, string>();
                try
                {
                    List<KeyValuePair> tuples = new List<KeyValuePair>();
                    XmlSerializer xmlIzer = new XmlSerializer(typeof(List<KeyValuePair>));
                    using (System.IO.StringReader sr = new System.IO.StringReader(fileDictString))
                    {
                        tuples = (xmlIzer.Deserialize(sr)) as List<KeyValuePair>;
                        if (tuples != null)
                        {
                            foreach (KeyValuePair tuple in tuples)
                            {
                                fileDict[tuple.Key] = tuple.Value;
                            }
                            return fileDict;
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            return new Dictionary<string, string>();
        }
    }

    public class UserBookmarksViewModel : ViewModelBase
    {
        public class UserBookmarkItem : ViewModelBase
        {
            private string _OriginalName = "";
            [XmlIgnoreAttribute]
            public string OriginalName { get { return _OriginalName; } set { _OriginalName = value; } }

            private string _OldName = null;
            private bool _IsEditing = false;
            [XmlIgnoreAttribute]
            public bool IsEditing
            {
                get { return _IsEditing; }
                set
                {
                    if (Set(ref _IsEditing, value))
                    {
                        RaisePropertyChanged();
                        RaisePropertyChanged("IsSelectedButNotEditing");
                        if (IsEditing)
                        {
                            _OldName = BookmarkName;
                        }
                    }
                }
            }
            [XmlIgnoreAttribute]
            private bool _IsSelected = false;
            public bool IsSelected
            {
                get { return _IsSelected; }
                set
                {
                    if (Set(ref _IsSelected, value))
                    {
                        RaisePropertyChanged("IsSelectedButNotEditing");
                    }
                }
            }
            [XmlIgnoreAttribute]
            public bool IsSelectedButNotEditing
            {
                get { return _IsSelected && !_IsEditing; }
            }

            private bool _IsBookmarkEdited = false;
            public bool IsBookmarkEdited
            {
                get { return _IsBookmarkEdited; }
                set { _IsBookmarkEdited = value; }
            }

            private string _BookmarkName = "";
            public string BookmarkName
            {
                get { return _BookmarkName; }
                set
                {
                    if (value != null)
                    {
                        if (!value.Equals(_BookmarkName))
                        {
                            _BookmarkName = value;
                            IsBookmarkEdited = true;
                            RaisePropertyChanged();
                        }
                    }
                }
            }

            private int _PageNumber = 0;
            public int PageNumber
            {
                get { return _PageNumber; }
                set
                {
                    if (value != _PageNumber)
                    {
                        _PageNumber = value;
                        IsBookmarkEdited = true;
                        RaisePropertyChanged();
                    }
                }
            }

            private int _ObjectNumber = -1;
            public int ObjectNumber
            {
                get { return _ObjectNumber; }
                set { _ObjectNumber = value; }
            }

            private Bookmark _PDFBookmark = null;
            [XmlIgnoreAttribute]
            public Bookmark PDFBookmark
            {
                get { return _PDFBookmark; }
                set { _PDFBookmark = value; }
            }

            [XmlIgnoreAttribute]
            public RelayCommand TextChangedCommand { get; private set; }
            [XmlIgnoreAttribute]
            public RelayCommand BookmarkItemButtonClickedCommand { get; private set; }
            public enum ItemRequestUpdates
            {
                DeleteRequested = 0,
                FocusRequested,
            }
            public delegate void ItemRequestDelegate(UserBookmarkItem item, ItemRequestUpdates updateType);
            public event ItemRequestDelegate ItemRequest;

            private void TextChangedCommandImpl(object parameter)
            {
                System.Diagnostics.Debug.WriteLine("Text changed!!!");
                string text = parameter as string;
                if (text != null)
                {
                    BookmarkName = text;
                }
            }
            private void BookmarkItemButtonClickedCommandImpl(object parameter)
            {
                string paramText = parameter as string;
                if (!string.IsNullOrEmpty(paramText))
                {
                    if (paramText.Equals("delete", StringComparison.OrdinalIgnoreCase))
                    {
                        ItemRequest?.Invoke(this, ItemRequestUpdates.DeleteRequested);
                    }
                    if (paramText.Equals("edit", StringComparison.OrdinalIgnoreCase))
                    {
                        IsEditing = true;
                        ItemRequest?.Invoke(this, ItemRequestUpdates.FocusRequested);
                    }
                    if (paramText.Equals("done", StringComparison.OrdinalIgnoreCase))
                    {
                        IsEditing = false;
                    }
                }
            }

            public UserBookmarkItem()
            {
                InitCommands();
            }

            public UserBookmarkItem(string itemName, int pageNumber, int objNum)
            {
                _BookmarkName = itemName;
                OriginalName = itemName;
                _PageNumber = pageNumber;
                ObjectNumber = objNum;
                InitCommands();
            }

            private void InitCommands()
            {
                TextChangedCommand = new RelayCommand(TextChangedCommandImpl);
                BookmarkItemButtonClickedCommand = new RelayCommand(BookmarkItemButtonClickedCommandImpl);
            }

            public void Delete()
            {
                if (PDFBookmark != null)
                {
                    PDFBookmark.Delete();
                }
            }
        }

        public delegate void FocusOnSelectedItemEventHandler(UserBookmarkItem item);
        public event FocusOnSelectedItemEventHandler FocusOnSelectedItemRequested = delegate { };
        private PDFViewCtrl _PDFViewCtrl;
        private string _DocumentTag;

        public delegate void DocumentModifiedEventHandler(PDFDoc doc);
        public DocumentModifiedEventHandler DocumentModified;

        public delegate void PageChangedDelegate(int prevPage, int newPage);
        public event PageChangedDelegate PageChanged;

        public delegate void BookmarksEditedEventHandler(PDFDoc doc);
        public event BookmarksEditedEventHandler BookmarksEdited;

        private void RaiseDocumentModified(PDFDoc doc)
        {
            if (DocumentModified != null)
            {
                DocumentModified(doc);
            }
        }

        private void NotifyBookmarksEdited(PDFDoc doc)
        {
            BookmarksEdited?.Invoke(doc);
            HasUnsavedBookmarks = true;
        }

        private enum SaveLocations
        {
            ToLocalStorage = 0,
            ToDoc,
        }
        private SaveLocations _SaveLocation = SaveLocations.ToLocalStorage;
        private SemaphoreSlim BookmarksIOSemaphore { get; set; }

        private ObservableCollection<UserBookmarkItem> _BookmarkList;
        public ObservableCollection<UserBookmarkItem> BookmarkList
        {
            get { return _BookmarkList; }
            set 
            { 
                _BookmarkList = value; 
                RaisePropertyChanged(); 
            }
        }

        private UserBookmarkItem _SelectedBookmark;
        public UserBookmarkItem SelectedBookmark
        {
            get { return _SelectedBookmark; }
            set
            {
                if (value != _SelectedBookmark)
                {
                    DeselectCurrentBookmark(_SelectedBookmark);
                    _SelectedBookmark = value;
                    RaisePropertyChanged();
                    if (_SelectedBookmark != null)
                    {
                        _SelectedBookmark.IsSelected = true;
                    }
                }
            }
        }

        private NavigationStack _NavigationStack;
        public NavigationStack NavigationStack
        {
            get { return _NavigationStack; }
            set { _NavigationStack = value; }
        }

        private bool _CanAlwaysEdit = true;
        public bool CanAlwaysEdit
        {
            get { return _CanAlwaysEdit; }
            set
            {
                if (Set(ref _CanAlwaysEdit, value))
                {
                    if (BookmarkList != null)
                    {
                        foreach (UserBookmarkItem item in BookmarkList)
                        {
                            item.IsSelected = _CanAlwaysEdit || item == SelectedBookmark;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Due to some strange problem on Windows 10 (or VS 2015) we have removed
        /// UserBookmarksViewModel(PDFViewCtrl ctr, string docTag)
        /// and replaced it with the empty constructor and an Init function.
        /// 
        /// Before, trying to include the tools in another project would cause some very strange
        /// error messages when building in release mode.
        /// </summary>
        public UserBookmarksViewModel()
        {
            BookmarkList = new ObservableCollection<UserBookmarkItem>();
            BookmarksIOSemaphore = new SemaphoreSlim(1);
            CreateCommands();
        }

        public void Init(PDFViewCtrl ctrl, string docTag)
        {
            _PDFViewCtrl = ctrl;
            _DocumentTag = docTag;
            PopulateList();
        }

        /// <summary>
        /// Deselects all bookmarks
        /// </summary>
        public void DeselectAll()
        {
            SelectedBookmark = null;
        }

        private async void PopulateList()
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            if (doc != null)
            {
                IList<UserBookmarkItem> items = null;
                if (UserBookmarkManager.DoesDocHaveBookmarks(doc))
                {
                    await BookmarksIOSemaphore.WaitAsync();
                    await UserBookmarkManager.GetBookmarkListAsync(""); // ensures the bookmark list is loaded if it exists.
                    items = await UserBookmarkManager.GetBookmarkListAsync(doc);
                    BookmarksIOSemaphore.Release();
                }
                if (items == null)
                {
                    await BookmarksIOSemaphore.WaitAsync();
                    items = await UserBookmarkManager.GetBookmarkListAsync(_DocumentTag);
                    BookmarksIOSemaphore.Release();
                }

                if (items != null && items.Count > 0)
                {
                    foreach (UserBookmarkItem item in items)
                    {
                        item.IsSelected = CanAlwaysEdit;
                        BookmarkList.Add(item);
                        item.ItemRequest += UserBookmarkItem_ItemRequest;
                    }
                }
                IsListPopulated = true;
                BookmarkList.CollectionChanged += BookmarkList_CollectionChanged;
                _SaveLocation = SaveLocations.ToDoc;
            }
            RaisePropertyChanged("ListHasItems");
        }

        #region Properties

        public bool ListHasItems
        {
            get { return BookmarkList.Count > 0; }
        }

        private bool _IsListPopulated = false;
        public bool IsListPopulated
        {
            get { return _IsListPopulated; }
            set
            {
                if (value != _IsListPopulated)
                {
                    _IsListPopulated = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasUnsavedBookmarks = false;
        public bool HasUnsavedBookmarks
        {
            get { return _HasUnsavedBookmarks; }
            set { Set(ref _HasUnsavedBookmarks, value); }
        }

        #endregion Properties


        #region Commands

        public RelayCommand AddBookmarkCommand { get; private set; }
        public RelayCommand RemoveBookmarkCommand { get; private set; }
        public RelayCommand RenameBookmarkCommand { get; private set; }
        public RelayCommand DoneEditingCommand { get; private set; }

        public RelayCommand ListViewItemClickCommand { get; private set; }

        private void CreateCommands()
        {
            AddBookmarkCommand = new RelayCommand(AddBookmarkCommandImpl);
            RemoveBookmarkCommand = new RelayCommand(RemoveBookmarkCommandImpl);
            RenameBookmarkCommand = new RelayCommand(RenameBookmarkCommandImpl);
            DoneEditingCommand = new RelayCommand(DoneEditingCommandImpl);

            ListViewItemClickCommand = new RelayCommand(ListViewItemClickCommandImpl);
        }

        private void AddBookmarkCommandImpl(object sender)
        {
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            if (doc != null)
            {
                try
                {
                    _PDFViewCtrl.DocLockRead();
                    int pgNum = _PDFViewCtrl.GetCurrentPage();
                    string pgNumString = string.Format(ResourceHandler.GetString("DefaultBookmarkFormat"), pgNum);

                    Page page = doc.GetPage(pgNum);
                    int objNum = page.GetSDFObj().GetObjNum();

                    UserBookmarkItem bookmark = new UserBookmarkItem(pgNumString, pgNum, objNum);
                    bookmark.IsSelected = CanAlwaysEdit;
                    BookmarkList.Add(bookmark);
                    bookmark.ItemRequest += UserBookmarkItem_ItemRequest;
                    _SaveLocation = SaveLocations.ToDoc;
                    NotifyBookmarksEdited(doc);
                    RaisePropertyChanged("ListHasItems");
                    SelectedBookmark = bookmark;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    _PDFViewCtrl.DocUnlockRead();
                }
            }
        }

        //private void EditBookmarksCommandImpl(object sender)
        //{
        //    IsInEditMode = true;
        //}

        private void RemoveBookmarkCommandImpl(object sender)
        {
            if (SelectedBookmark != null)
            {
                RemoveBookmarkImpl(SelectedBookmark);
            }
        }

        private void RemoveBookmarkImpl(UserBookmarkItem bookmark)
        {
            if (bookmark != null)
            {
                bookmark.Delete();
                BookmarkList.Remove(bookmark);
                _ReorderItem = null;
                NotifyBookmarksEdited(_PDFViewCtrl.GetDoc());
                RaisePropertyChanged("ListHasItems");
                _SaveLocation = SaveLocations.ToDoc;
            }
        }

        private void RenameBookmarkCommandImpl(object sender)
        {
            EditCurrentBookmark();
        }

        private void DoneEditingCommandImpl(object sender)
        {
            //_SpontaneousEditMode = false;
            //IsInEditMode = false;
            SelectedBookmark = null;
        }

        private void ListViewItemClickCommandImpl(object sentArgs)
        {
            Windows.UI.Xaml.Controls.ItemClickEventArgs args = sentArgs as Windows.UI.Xaml.Controls.ItemClickEventArgs;
            if (args != null)
            {
                UserBookmarkItem item = args.ClickedItem as UserBookmarkItem;
                if (item != null)
                {
                    if (item.PageNumber > 0 && item.PageNumber <= _PDFViewCtrl.GetPageCount())
                    {
                        PageChanged?.Invoke(_PDFViewCtrl.GetCurrentPage(), item.PageNumber);
                        NavigationState beforeState = NavigationState.TakeSnapshot(_PDFViewCtrl);
                        _PDFViewCtrl.SetCurrentPage(item.PageNumber);
                        _PDFViewCtrl.SetPageViewMode(PDFViewCtrlPageViewMode.e_fit_page);
                        NavigationState afterState = NavigationState.TakeSnapshot(_PDFViewCtrl);
                        if (NavigationStack != null)
                        {
                            NavigationStack.RegisterJump(beforeState, afterState);
                        }
                    }
                }
            }
        }


        #endregion Commands


        #region Behaviour

        private string _BeforeEditingBookmarkName = "";

        private void DeselectCurrentBookmark(UserBookmarkItem item)
        {
            if (item != null)
            {
                item.IsSelected = _CanAlwaysEdit;
                item.IsEditing = false;
            }
        }

        private void EditCurrentBookmark()
        {
            if (_SelectedBookmark != null)
            {
                _SelectedBookmark.IsEditing = true;
                _BeforeEditingBookmarkName = _SelectedBookmark.BookmarkName;
                FocusOnSelectedItemRequested(_SelectedBookmark);
            }
        }

        private void UserBookmarkItem_ItemRequest(UserBookmarkItem item, UserBookmarkItem.ItemRequestUpdates updateType)
        {
            if (updateType == UserBookmarkItem.ItemRequestUpdates.FocusRequested)
            {
                _SelectedBookmark = item;
                FocusOnSelectedItemRequested?.Invoke(item);
            }
            else if (updateType == UserBookmarkItem.ItemRequestUpdates.DeleteRequested)
            {
                RemoveBookmarkImpl(item);
                NotifyBookmarksEdited(_PDFViewCtrl.GetDoc());
            }
        }

        object _ReorderItem;
        int _ReorderIndexFrom;

        void BookmarkList_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            _SaveLocation = SaveLocations.ToDoc;

            switch (e.Action)
            {
                case System.Collections.Specialized.NotifyCollectionChangedAction.Remove:
                    _ReorderItem = e.OldItems[0];
                    _ReorderIndexFrom = e.OldStartingIndex;
                    break;
                case System.Collections.Specialized.NotifyCollectionChangedAction.Add:
                    if (_ReorderItem == null)
                    {
                        return;
                    }
                    HandleReorder(_ReorderItem, _ReorderIndexFrom, e.NewStartingIndex);
                    _ReorderItem = null;
                    break;
            }
        }

        void HandleReorder(object item, int indexFrom, int indexTo)
        {
            UserBookmarkItem movedBookmark = item as UserBookmarkItem;
            if (movedBookmark != null)
            {
                if (movedBookmark.PDFBookmark != null)
                {
                    movedBookmark.PDFBookmark.Delete();
                    movedBookmark.PDFBookmark = null;
                }
                movedBookmark.IsBookmarkEdited = true;
                NotifyBookmarksEdited(_PDFViewCtrl.GetDoc());
            }
        }


        #endregion Behaviour

        #region Public Functions for the View

        public void CancelEditingText()
        {
            _SelectedBookmark.BookmarkName = _BeforeEditingBookmarkName;
            if (_SelectedBookmark.BookmarkName.Equals(_SelectedBookmark.OriginalName))
            {
                _SelectedBookmark.IsBookmarkEdited = false;
            }
            FinishEditingText();
        }

        public void FinishEditingText()
        {
            if (_SelectedBookmark != null)
            {
                _SelectedBookmark.IsEditing = false;
                if (_SelectedBookmark.IsBookmarkEdited)
                {
                    NotifyBookmarksEdited(_PDFViewCtrl.GetDoc());
                }
            }
        }

        public async void SaveBookmarks()
        {
            FinishEditingText();
            PDFDoc doc = _PDFViewCtrl.GetDoc();
            await BookmarksIOSemaphore.WaitAsync();
            try
            {
                if (HasUnsavedBookmarks && doc != null && _BookmarkList != null)
                {
                    bool eraseFromLocalStorage = _SaveLocation == SaveLocations.ToDoc;

                    if (_BookmarkList.Count > 0 || _SaveLocation == SaveLocations.ToDoc)
                    {
                        bool hasUnsavedBookmarks = HasUnsavedBookmarks;
                        List<UserBookmarkItem> items = new List<UserBookmarkItem>();
                        items.AddRange(_BookmarkList);
                        try
                        {
                            if (_SaveLocation == SaveLocations.ToDoc)
                            {
                                await UserBookmarkManager.SaveBookmarkListAsync(doc, items);
                                RaiseDocumentModified(doc);
                            }
                            else if (!string.IsNullOrEmpty(_DocumentTag))
                            {
                                await UserBookmarkManager.SaveBookmarkListAsync(_DocumentTag, items);
                            }
                        }
                        catch (Exception ex)
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                        }
                        HasUnsavedBookmarks = HasUnsavedBookmarks;
                    }
                    else
                    {
                        eraseFromLocalStorage = true;
                    }

                    await UserBookmarkManager.EraseBookmarkListAsync(_DocumentTag);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                BookmarksIOSemaphore.Release();
            }
        }

        public async Task WaitForBookmarkSavingAsync()
        {
            await BookmarksIOSemaphore.WaitAsync();
            BookmarksIOSemaphore.Release();
            HasUnsavedBookmarks = false;
        }

        #endregion Public Functions for the View

        #region Back Key
        public bool GoBack()
        {
            if (SelectedBookmark != null && SelectedBookmark.IsEditing)
            {
                SelectedBookmark.IsEditing = false;
                return true;
            }
            return false;
        }
        #endregion Back Key

    }
}
