using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    public class FreehandCreate : SimpleShapeCreate
    {
        public class StrokeAppearance
        {
            public int PageNumber { get; private set; }
            public double Opacity { get; private set; }
            public double Thickness { get; private set; }
            public SolidColorBrush Brush { get; private set; }
            public bool UseSmoothing { get; private set; }

            public StrokeAppearance(int page, double opacity, double thickness, SolidColorBrush brush, bool useSmoothing)
            {
                PageNumber = page;
                Opacity = opacity;
                Thickness = thickness;
                Brush = brush;
                UseSmoothing = useSmoothing;
            }
            
            public StrokeAppearance(StrokeAppearance other)
            {
                PageNumber = other.PageNumber;
                Opacity = other.Opacity;
                Thickness = other.Thickness;
                Brush = other.Brush;
                UseSmoothing = other.UseSmoothing;
            }

            public bool IsEqual(StrokeAppearance other)
            {
                if (other == null)
                {
                    return false;
                }
                bool isEqual = PageNumber == other.PageNumber;
                isEqual &= Opacity == other.Opacity;
                isEqual &= Thickness == other.Thickness;
                isEqual &= UseSmoothing == other.UseSmoothing;
                isEqual &= Brush.Color == other.Brush.Color;
                return isEqual;
            }
        }

        public class PathFigureBuilder
        {
            private bool mHasFirstPoint = false;
            private uint mLastCubicIndex = 1;

            public bool mUseSmoothing = true;
            private PathFigure mFigure;

            private CubicCurveBuilder mCurveBuilder;

            public PathFigure Figure { get { return mFigure; } }

            public PathFigureBuilder(bool useSmoothing)
            {
                mFigure = new PathFigure();
                mCurveBuilder = Annots.Ink.GetCurveBuilder();
                mUseSmoothing = true;
            }

            public void AddPointToPath(UIPoint point)
            {
                uint pointsBeforeAddition = mCurveBuilder.NumCubicPoints();
                mCurveBuilder.AddSourcePoint(point.X, point.Y);
                if (!mHasFirstPoint)
                {
                    mHasFirstPoint = true;
                    mFigure.StartPoint = point;
                }
                else
                {
                    if (mUseSmoothing)
                    {
                        uint numCubicPoints = mCurveBuilder.NumCubicPoints();
                        while (mLastCubicIndex + 3 < mCurveBuilder.NumCubicPoints())
                        {
                            BezierSegment cbc = new BezierSegment();

                            cbc.Point1 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex), mCurveBuilder.GetCubicYCoord(mLastCubicIndex));
                            cbc.Point2 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex + 1), mCurveBuilder.GetCubicYCoord(mLastCubicIndex + 1));
                            cbc.Point3 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex + 2), mCurveBuilder.GetCubicYCoord(mLastCubicIndex + 2));
                            mFigure.Segments.Add(cbc);

                            mLastCubicIndex += 3;
                        }
                    }
                    else
                    {
                        LineSegment segment = new LineSegment();
                        segment.Point = point;
                        mFigure.Segments.Add(segment);
                    }
                }
            }

            public void ClosePath()
            {
                if (mUseSmoothing && mCurveBuilder.NumSourcePoints() > 1 && (int)mCurveBuilder.NumCubicPoints() > 3)
                {
                    uint numCubicPoints = mCurveBuilder.NumCubicPoints();
                    while (mLastCubicIndex + 2 < mCurveBuilder.NumCubicPoints())
                    {
                        BezierSegment cbc = new BezierSegment();

                        cbc.Point1 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex), mCurveBuilder.GetCubicYCoord(mLastCubicIndex));
                        cbc.Point2 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex + 1), mCurveBuilder.GetCubicYCoord(mLastCubicIndex + 1));
                        cbc.Point3 = new UIPoint(mCurveBuilder.GetCubicXCoord(mLastCubicIndex + 2), mCurveBuilder.GetCubicYCoord(mLastCubicIndex + 2));
                        mFigure.Segments.Add(cbc);

                        mLastCubicIndex += 3;
                    }
                }
            }
        }

        public class PenStroke
        {
            private List<PDFPoint> mStrokePoints;
            public IReadOnlyList<PDFPoint> StrokePoints
            {
                get { return mStrokePoints; }
            }

            private PDFRect mBounds;
            public PDFRect Bounds { get { return mBounds; } }
            
            public void AddPoint(PDFPoint point)
            {
                mStrokePoints.Add(point);
                if (mBounds == null)
                {
                    mBounds = new PDFRect(point.x, point.y, point.x, point.y);
                }
                else
                {
                    mBounds.x1 = Math.Min(point.x, mBounds.x1);
                    mBounds.y1 = Math.Min(point.y, mBounds.y1);
                    mBounds.x2 = Math.Max(point.x, mBounds.x2);
                    mBounds.y2 = Math.Max(point.y, mBounds.y2);
                }
            }

            public PenStroke()
            {
                mStrokePoints = new List<PDFPoint>();
            }
        }


        public class StrokeCollection
        {
            public class StrokeBundle
            {
                public Path mShape;
                public PenStroke mPenStroke;
            }

            public StrokeAppearance mStrokeAppearance;
            public List<StrokeBundle> mStrokeList;
            public List<Path> mShapes;
            public PDFViewCtrl mPDFView;
            public PDFRect mPageBoundaries;

            private Canvas mStrokeCanvas;

            public PenStroke CurrentPenStroke { get { return mStrokeList.Count > 0 ? mStrokeList[mStrokeList.Count - 1].mPenStroke : null; } }
            public Path CurrentShape { get { return mStrokeList.Count > 0 ? mStrokeList[mStrokeList.Count - 1].mShape : null; } }
            /// <summary>
            /// Attach this to the tool.
            /// </summary>
            public FrameworkElement StrokeWidget { get { return mStrokeCanvas; } }

            public StrokeCollection(StrokeAppearance strokeAppearance, PDFViewCtrl pdfView)
            {
                mStrokeList = new List<StrokeCollection.StrokeBundle>();
                mStrokeAppearance = strokeAppearance;
                mPDFView = pdfView;
                mStrokeCanvas = new Canvas();
            }

            /// <summary>
            /// Creates the necessary structures to create a new path, making 
            /// them available through CurrentPenStroke and CurrentShape.
            /// </summary>
            /// <returns>The PathFigureCollection which belongs to CurrentShape</returns>
            public PathFigureCollection StartNewPath()
            {
                StrokeBundle bundle = new StrokeBundle();
                bundle.mShape = CreateNewShape();
                bundle.mPenStroke = new PenStroke();
                mStrokeList.Add(bundle);
                mStrokeCanvas.Children.Add(bundle.mShape);
                return (bundle.mShape.Data as PathGeometry).Figures;
            }

            public void AddPath(PenStroke penStroke)
            {
                StrokeBundle bundle = new StrokeBundle();
                bundle.mPenStroke = penStroke;
                mStrokeList.Add(bundle);
                RestorePath(bundle);
            }

            public void CancelCurrentPath()
            {
                if (mStrokeList.Count > 0)
                {
                    mStrokeCanvas.Children.Remove(mStrokeList[mStrokeList.Count - 1].mShape);
                    mStrokeList.RemoveAt(mStrokeList.Count - 1);
                }
            }

            /// <summary>
            /// The StrokeBundle needs a penStroke, but doesn't need a path.
            /// </summary>
            /// <param name="bundle"></param>
            private void RestorePath(StrokeBundle bundle)
            {
                Path path = CreateNewShape();
                if (bundle.mShape != null && mStrokeCanvas.Children.Contains(bundle.mShape))
                {
                    mStrokeCanvas.Children.Remove(bundle.mShape);
                }

                if (mPageBoundaries != null)
                {
                    PathFigureBuilder builder = new PathFigureBuilder(mStrokeAppearance.UseSmoothing);
                    PathFigure figure = builder.Figure;
                    PDFDouble x = new PDFDouble();
                    PDFDouble y = new PDFDouble();
                    foreach (PDFPoint point in bundle.mPenStroke.StrokePoints)
                    {
                        x.Value = point.x;
                        y.Value = point.y;
                        mPDFView.ConvPagePtToScreenPt(x, y, mStrokeAppearance.PageNumber);
                        mPDFView.ConvScreenPtToAnnotationCanvasPt(x, y);
                        builder.AddPointToPath(new UIPoint(x.Value - mPageBoundaries.x1, y.Value - mPageBoundaries.y1));
                    }
                    builder.ClosePath();
                    PathFigureCollection figures = (path.Data as PathGeometry).Figures;
                    figures.Add(figure);
                    bundle.mShape = path;

                    mStrokeCanvas.Children.Add(bundle.mShape);
                }
            }

            public void RedrawAllPaths()
            {
                foreach (StrokeBundle stroke in mStrokeList)
                {
                    RestorePath(stroke);
                }
            }

            public bool Erase(List<PDFPoint> eraseList, PDFRect inkRekt, double halfWidth, ref List<PenStroke> addedStrokes, ref List<PenStroke> removedStrokes)
            {
                bool somethingWasModified = false;
                if (eraseList.Count < 2)
                {
                    return somethingWasModified;
                }
                List<StrokeBundle> removedBundles = new List<StrokeBundle>();
                foreach (StrokeBundle strokeBundle in mStrokeList)
                {
                    SDF.ObjSet set = new SDF.ObjSet();
                    SDF.Obj listOfStrokes = set.CreateArray();
                    SDF.Obj existingStrokes = listOfStrokes.PushBackArray();
                    foreach (PDFPoint point in strokeBundle.mPenStroke.StrokePoints)
                    {
                        existingStrokes.PushBackNumber(point.x);
                        existingStrokes.PushBackNumber(point.y);
                    }
                    bool wasModified = false;
                    PDFPoint p1 = eraseList[0];
                    for (int i = 1; i < eraseList.Count; ++i)
                    {
                        PDFPoint p2 = eraseList[i];
                        if (Annots.Ink.ErasePoints(listOfStrokes, inkRekt, p1, p2, halfWidth))
                        {
                            wasModified = true;
                            somethingWasModified = true;
                        }
                        p1 = p2;
                    }

                    if (wasModified)
                    {
                        removedBundles.Add(strokeBundle);
                        for (int i = 0; i < listOfStrokes.Size(); ++i)
                        {
                            PenStroke newStroke = new PenStroke();
                            SDF.Obj strokeList = listOfStrokes.GetAt(i);
                            if (strokeList.Size() > 2)
                            {
                                for (int j = 0; j < strokeList.Size(); j += 2)
                                {
                                    newStroke.AddPoint(new PDFPoint(strokeList.GetAt(j).GetNumber(), strokeList.GetAt(j + 1).GetNumber()));
                                }
                                addedStrokes.Add(newStroke);
                            }
                        }
                    }
                }
                foreach (StrokeBundle removedBundle in removedBundles)
                {
                    if (mStrokeList.Contains(removedBundle))
                    {
                        mStrokeList.Remove(removedBundle);
                    }
                    if (removedBundle.mShape != null && mStrokeCanvas.Children.Contains(removedBundle.mShape))
                    {
                        mStrokeCanvas.Children.Remove(removedBundle.mShape);
                    }
                    removedStrokes.Add(removedBundle.mPenStroke);
                }
                foreach (PenStroke addedStroke in addedStrokes)
                {
                    AddPath(addedStroke);
                }

                return somethingWasModified;
            }

            public PDFRect GetBoundingRectInPageSpace()
            {
                PDFRect bounds = null;
                foreach (StrokeBundle stroke in mStrokeList)
                {
                    if (bounds == null)
                    {
                        bounds = new PDFRect(stroke.mPenStroke.Bounds.x1, stroke.mPenStroke.Bounds.y1, stroke.mPenStroke.Bounds.x2, stroke.mPenStroke.Bounds.y2);
                    }
                    else
                    {
                        bounds.x1 = Math.Min(bounds.x1, stroke.mPenStroke.Bounds.x1);
                        bounds.y1 = Math.Min(bounds.y1, stroke.mPenStroke.Bounds.y1);
                        bounds.x2 = Math.Max(bounds.x2, stroke.mPenStroke.Bounds.x2);
                        bounds.y2 = Math.Max(bounds.y2, stroke.mPenStroke.Bounds.y2);
                    }
                }
                if (bounds == null)
                {
                    bounds = new PDFRect();
                }
                return bounds;
            }

            private Path CreateNewShape()
            {
                PathFigureCollection collection = new PathFigureCollection();
                PathGeometry geom = new PathGeometry();
                Path shape = new Path();
                shape.Data = geom;
                geom.Figures = collection;
                shape.StrokeLineJoin = PenLineJoin.Round;
                shape.StrokeEndLineCap = PenLineCap.Round;
                shape.StrokeStartLineCap = PenLineCap.Round;
                shape.Opacity = mStrokeAppearance.Opacity;
                shape.StrokeThickness = mStrokeAppearance.Thickness * mPDFView.GetZoom();
                shape.Stroke = mStrokeAppearance.Brush;
                return shape;
            }

            public override string ToString()
            {
                int sequenceNumber = 0;
                StringBuilder sb = new StringBuilder();
                sb.AppendLine(string.Format("Collection {0} has {1} visual children, and {1} listed children", sequenceNumber++, mStrokeCanvas.Children.Count(), mStrokeList.Count));
                sb.AppendLine("Visual Children:");
                foreach (UIElement elem in mStrokeCanvas.Children)
                {
                    Path childPath = elem as Path;
                    sb.AppendLine(string.Format("Path with {0} points", (childPath.Data as PathGeometry).Figures.Count > 0 ? ((childPath.Data as PathGeometry).Figures[0].Segments.Count + 1) : 0));
                }
                sb.Append(Environment.NewLine);
                sb.AppendLine("Listed Children:");
                foreach (StrokeBundle strokeBundle in mStrokeList)
                {
                    sb.AppendLine(string.Format("list with {0} points", strokeBundle.mPenStroke.StrokePoints.Count));
                }

                return sb.ToString();
            }
        }

        public delegate void StrokeAddedDelegate();
        public delegate void StylusIsDetectedDelegate();

        protected class UndoRedoSnapshot
        {
            public class UndoRedoStroke
            {
                public StrokeAppearance Appearance;
                public List<PenStroke> PenStrokes;

                public UndoRedoStroke(StrokeCollection strokeCollection)
                {
                    Appearance = new StrokeAppearance(strokeCollection.mStrokeAppearance);
                    PenStrokes = new List<PenStroke>();
                    foreach (StrokeCollection.StrokeBundle strokeBundle in strokeCollection.mStrokeList)
                    {
                        PenStrokes.Add(strokeBundle.mPenStroke);
                    }
                }
            }


            private List<UndoRedoStroke> mUndoRedoStrokes;

            public UndoRedoSnapshot(List<StrokeCollection> penCollection)
            {
                mUndoRedoStrokes = new List<UndoRedoStroke>();
                foreach (StrokeCollection strokeCollection in penCollection)
                {
                    mUndoRedoStrokes.Add(new UndoRedoStroke(strokeCollection));
                }
            }

            public List<StrokeCollection> RestoreStrokes(FreehandCreate host, PDFViewCtrl ctrl)
            {
                List<StrokeCollection> penCollection = new List<StrokeCollection>();
                foreach (UndoRedoStroke undoRedoState in mUndoRedoStrokes)
                {
                    StrokeCollection strokeCollection = new StrokeCollection(undoRedoState.Appearance, ctrl);

                    try
                    {
                        ctrl.DocLockRead();
                        PDFRect rect = ctrl.GetDoc().GetPage(strokeCollection.mStrokeAppearance.PageNumber).GetBox(ctrl.GetPageBox());
                        rect = UtilityFunctions.ConvertFromPageRectToCanvasRect(rect, strokeCollection.mStrokeAppearance.PageNumber, ctrl);
                        rect.Normalize();
                        strokeCollection.mPageBoundaries = rect;
                    }
                    catch (Exception) { }
                    finally
                    {
                        ctrl.DocUnlockRead();
                    }
                    foreach (PenStroke penStroke in undoRedoState.PenStrokes)
                    {
                        strokeCollection.AddPath(penStroke);
                    }
                    host.Children.Add(strokeCollection.StrokeWidget);
                    penCollection.Add(strokeCollection);
                }
                return penCollection;
            }
        }

        protected List<UndoRedoSnapshot> mUndoneSnapshots;
        protected List<UndoRedoSnapshot> mUndoSnapshots;

        public delegate void StrokeMadeVisibleDelegate(StrokeAppearance appearance, PenStroke stroke);
        public delegate void StrokeMadeInvisibleDelegate(PenStroke stroke);
        public delegate void CurrentStrokeModifedDelegate(PenStroke stroke);
        public delegate void AllStrokesMadeInvisibleDelagate();

        protected bool mShapeHasBeenCreated = false;
        protected PathFigureBuilder mCurrentPathFigure;
        protected bool mFirstTime = false; // Whether or not a pointer has been pressed already.
        protected bool mMustCreateNewStroke = true;
        protected bool mMustAddPath = true;
        protected double mLeftmost, mRightmost, mTopmost, mBottommost;
        protected bool mMultiStrokeMode = false;
        protected UIPoint mLastPoint;
        protected PDFPoint mLastEndPoit;

        protected Path mCurrentPath;
        protected PenStroke mCurrentStroke;
        protected StrokeCollection mCurrentStrokeSet;
        protected List<StrokeCollection> mAllStrokes;

        // for error correcting.
        private UIPoint m_cnCurrentPoint;
        private UIPoint mTwoPointsBack;
        private UIPoint mOnePointBack;
        private bool mStartedPathSmoothing;
        private bool mIsStylus = false;

        // Differentiates between regular ink and a marker
        private bool mIsMarker = false;

        // for timing the ink when spontaneously creating it
        private DispatcherTimer mCompletionTimer;
        private bool mIsInTimedMode;
        private bool mShouldSaveInk;
        private bool mSaveInkOnClose = false;
        private bool mRetainInkWhenScalingOrSizing = false;

        // Eraser
        protected bool mIsEraser = false;
        protected bool mRemainInEraserMode = false;
        protected bool mEraseThisInkOnly = true;
        protected StrokeCollection mEraserStrokeCollection;
        protected bool mEraserStartErasing = false;
        protected PenStroke mFullEraserStroke;
        protected PenStroke mPartialEraserStroke;
        protected PDFPoint mLastErasedPoint;

        protected bool mEraserPropertiesSet = false;
        protected double mEraserThickness = -1;
        protected double mEraserOpacity = -1;
        protected SolidColorBrush mEraserColor;

        public const int DEFAULT_INK_PRESET_LABEL_FOR_ERASER = 100;
        public const double DEFAULT_INK_ERASER_OPACITY = 0.5;
        public static readonly Windows.UI.Color DEFAULT_INK_ERASER_COLOR = Windows.UI.Colors.LightGray;

        public bool IsEraser
        {
            get { return mIsEraser; }
            set
            {
                if (mIsEraser != value)
                {
                    if (mIsEraser)
                    {
                        if (this.Children.Contains(mEraserStrokeCollection.StrokeWidget))
                        {
                            this.Children.Remove(mEraserStrokeCollection.StrokeWidget);
                        }
                    }
                    mIsEraser = value;
                }
            }
        } 

        //protected 

        /// <summary>
        /// Gets whether the tool is currently in timed mode, meaning it will wait for a while and combine the next stroke
        /// into one annotation if it is close enough.
        /// </summary>
        public bool IsInTimedMode
        {
            get { return mIsInTimedMode; }
        }


        /// <summary>
        /// Sets whether the tool will push back each stroke once the pointer is released, 
        /// or if it should continue to accumulate strokes until the user explicitly says it's done
        /// 
        /// When done, call CommitAnnotation()
        /// </summary>
        public bool MultiStrokeMode
        {
            get
            {
                return mMultiStrokeMode;
            }
            set
            {
                mMultiStrokeMode = value;
            }
        }

        /// <summary>
        /// Gets or set whether ink should be saved if the tool is closed
        /// </summary>
        public bool SaveInkOnClose
        {
            get { return mSaveInkOnClose; }
            set { mSaveInkOnClose = value; }
        }

        /// <summary>
        /// Gets or sets whether ink should be saved when the PDFViewCtrl is zoomed or resized or it's layout is otherwise changed.
        /// </summary>
        public bool RetainInkWhenScalingOrSizing
        {
            get { return mRetainInkWhenScalingOrSizing; }
            set { mRetainInkWhenScalingOrSizing = value; }
        }

        /// <summary>
        /// Gets or sets whether this is a stylus annotation
        /// </summary>
        public bool IsUsingSylus
        {
            get { return mIsStylus; }
            set { mIsStylus = value; }
        }

        /// <summary>
        /// Gets or sets whether this is a marker 
        /// </summary>
        public bool IsMarker
        {
            get { return mIsMarker; }
            private set { mIsMarker = value; }
        }

        /// <summary>
        /// Gets the list of all strokes in page space.
        /// </summary>
        public List<StrokeCollection> AllStrokes
        {
            get { return mAllStrokes; }
        }

        /// <summary>
        /// Gets whether or not the ink has any strokes, whether they've been undone or not.
        /// </summary>
        public bool HasStrokes { get { return AllStrokes.Count > 0; } }

        /// <summary>
        /// This event is invoked when the Tool detects a Stylus and is not already in 
        /// Stylus mode (i.e. IsUsingSylus is false).
        /// </summary>
        public event StylusIsDetectedDelegate StylusIsDetected = delegate { };

        /// <summary>
        /// This even is invoked when a path has been drawn for the ink annotation.
        /// Note: It is not invoked after a call to RedoStroke()
        /// </summary>
        public event StrokeAddedDelegate StrokeAdded;

        /// <summary>
        /// This event is raised whenever a whole path is visible (e.g. as a result of drawing or redo)
        /// </summary>
        public event StrokeMadeVisibleDelegate StrokeMadeVisible;

        /// <summary>
        /// This even is raised whenever a whole path is no longer visible (e.g. as a result of undo)
        /// </summary>
        public event StrokeMadeInvisibleDelegate StrokeMadeInvisible;

        /// <summary>
        /// This event is raised whenever the current stroke is modified
        /// </summary>
        public event CurrentStrokeModifedDelegate CurrentStrokeModifed;

        public event AllStrokesMadeInvisibleDelagate AllStrokesMadeInvisible;

        public FreehandCreate(PDFViewCtrl ctrl, ToolManager tMan, bool isMarker = false)
            : base(ctrl, tMan)
        {
            if (!isMarker)
            {
                mNextToolMode = ToolType.e_ink_create;
                mToolMode = ToolType.e_ink_create;
            }
            else
            {
                mNextToolMode = ToolType.e_ink_marker;
                mToolMode = ToolType.e_ink_marker;
            }
            mShouldSaveInk = true;

            mAllStrokes = new List<StrokeCollection>();
            mUndoneSnapshots = new List<UndoRedoSnapshot>();
            mUndoSnapshots = new List<UndoRedoSnapshot>();
            mUndoSnapshots.Add(new UndoRedoSnapshot(new List<StrokeCollection>()));

            mIsMarker = isMarker;
        }

        internal override void OnClose()
        {
            EndTimer();
            if (MultiStrokeMode && mShouldSaveInk && SaveInkOnClose && mStartDrawing)
            {
                Finish();
                mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
                mDelayedRemovalSet = true;
            }
            base.OnClose();
            mPDFView.RequestRendering();
        }

        internal override bool OnSize()
        {
            if (MultiStrokeMode)
            {
                if (RetainInkWhenScalingOrSizing)
                {
                    foreach (StrokeCollection penCollection in mAllStrokes)
                    {
                        penCollection.RedrawAllPaths();
                    }
                }
                else if (!ReturnToPanModeWhenFinished)
                {
                    CommitAnnotation();
                    if (mToolManager.CurrentTool.ToolMode == ToolType.e_ink_create)
                    {
                        (mToolManager.CurrentTool as FreehandCreate).MultiStrokeMode = true;
                    }
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                }
            }
            return false;
        }

        internal override bool OnScale()
        {
            if (MultiStrokeMode)
            {
                if (RetainInkWhenScalingOrSizing)
                {
                    foreach (StrokeCollection penCollection in mAllStrokes)
                    {
                        penCollection.RedrawAllPaths();
                    }
                }
                else if (!ReturnToPanModeWhenFinished)
                {
                    CommitAnnotation();
                    EndCurrentTool(ToolType.e_ink_create);
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                }
            }
            return base.OnScale();
        }

        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            AddPointer(e.Pointer);
            if (mIsEraser && !mRemainInEraserMode)
            {
                mIsEraser = false;
                mMustCreateNewStroke = true;
            }
            if (mToolManager.ContactPoints.Count > 1)
            {
                mPointerID = -1;
                mToolManager.EnableScrollByManipulation = true;
                if (mEraserStrokeCollection != null && this.Children.Contains(mEraserStrokeCollection.StrokeWidget))
                {
                    this.Children.Remove(mEraserStrokeCollection.StrokeWidget);
                }
                if (mCurrentStrokeSet != null && this.Children.Contains(mCurrentStrokeSet.StrokeWidget))
                {
                    StrokeMadeInvisible?.Invoke(mCurrentStroke);
                    mCurrentStrokeSet.CancelCurrentPath();
                }
                if (mMultiStrokeMode && mCurrentPathFigure != null && !mIsStylus)
                {
                    mCurrentPathFigure = null;
                }
                if (mPageCropOnClient != null)
                {
                    mRectToKeepOnScreenWhileManipulating = new Rect(mPageCropOnClient.x1, mPageCropOnClient.y1, mPageCropOnClient.x2, mPageCropOnClient.y2);
                    mRectToKeepOnScreenWhileManipulating.Inflate(35);
                }
                return false;
            }
            if (IsUsingSylus && e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Pen && !IsEraser)
            {
                if (mIsInTimedMode)
                {
                    EnableScrolling();
                    EndTimer();
                    mNextToolMode = ToolType.e_pan;
                    return true;
                }
                else 
                {
                    if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Touch)
                    {
                        mToolManager.EnableScrollByManipulation = true;
                        mToolManager.EnableOneFingerScroll = true;
                        if (mPageCropOnClient != null)
                        {
                            mRectToKeepOnScreenWhileManipulating = new Rect(mPageCropOnClient.x1, mPageCropOnClient.y1, mPageCropOnClient.x2, mPageCropOnClient.y2);
                            mRectToKeepOnScreenWhileManipulating.Inflate(35);
                        }
                    }
                    return true;
                }
            }
            PointerPoint pointerPoint = e.GetCurrentPoint(mPDFView);
            if (pointerPoint.Properties.IsEraser)
            {
                if (mMultiStrokeMode && !mIsInTimedMode)
                {
                    if (!mEraserPropertiesSet)
                    {
                        mEraserPropertiesSet = false;
                        double eraserThickness = Settings.GetThickness(ToolType.e_ink_eraser, DEFAULT_INK_PRESET_LABEL_FOR_ERASER);
                        double eraserOpacity = DEFAULT_INK_ERASER_OPACITY;
                        Windows.UI.Color eraserColor = DEFAULT_INK_ERASER_COLOR;

                        SetEraserPropertiesInternal(true, eraserThickness, eraserOpacity, eraserColor);
                    }
                    mIsEraser = true;
                }
                else
                {
                    if (mIsInTimedMode)
                    {
                        mNextToolMode = ToolType.e_ink_eraser;
                    }
                    return true;
                }
            }
            if (!mIsStylus && pointerPoint.PointerDevice.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Pen)
            {
                if (mJustSwitchedFromAnotherTool)
                {
                    mIsInTimedMode = true;
                    mShouldSaveInk = true;
                    mMultiStrokeMode = true;
                }
                mIsStylus = true;
                StylusIsDetected();
            }
            if (mIsInTimedMode)
            {
                StopTimer();
            }

            UIPoint screenPoint = pointerPoint.Position;
            if (mDownPageNumber > 0) // This means each free hand annotation has more than one stroke
            {
                
                int newDownPageNumber = mPDFView.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);
                if (newDownPageNumber != mDownPageNumber)
                {
                    if (mIsInTimedMode)
                    {
                        mNextToolMode = ToolType.e_pan;
                    }
                    else if (MultiStrokeMode)
                    {
                        if (!ReturnToPanModeWhenFinished)
                        {
                            EndCurrentTool(ToolType.e_ink_create);
                            (mToolManager.CurrentTool as FreehandCreate).PointerPressedHandler(sender, e);
                            mPDFView.RequestRendering();
                        }
                    }
                    return false;
                }
            }

            base.PointerPressedHandler(sender, e); // gets us page, point, crop box and pointer capture
            if (mSimpleShapeCanceledAtPress)
            {
                return false;
            }


            if (mAllStrokes.Count > 0 && mIsInTimedMode && mToolManager.INK_MARGIN_FOR_STARTING_NEW_ANNOTATION > 0)
            {
                // check distance of mouse down from current bounding box.
                PDFRect rect = new PDFRect(mLeftmost, mTopmost, mRightmost, mBottommost);
                rect.Inflate(mToolManager.INK_MARGIN_FOR_STARTING_NEW_ANNOTATION);
                if (!rect.Contains(m_cnPt1.x, m_cnPt1.y))
                {
                    mNextToolMode = ToolType.e_pan;
                    return true;
                }
            }

            if (!mIsStylus)
            {
                mPDFView.CancelRendering();
            }

            mZoomLevel = mPDFView.GetZoom();
            if (mIsEraser)
            {
                mDrawThickness = mEraserThickness * mZoomLevel;
            }
            else
            {
                mDrawThickness = mStrokeThickness * mZoomLevel;
            }

            // Narrow the crop box to account for the thickness of the line we're drawing
            double halfThickness = mDrawThickness / 2;
            if (mPageCropOnClient != null)
            {
                if (m_cnPt1.x < mPageCropOnClient.x1 + halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x1 + halfThickness;
                }
                if (m_cnPt1.x > mPageCropOnClient.x2 - halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x2 - halfThickness;
                }
                if (m_cnPt1.y < mPageCropOnClient.y1 + halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y1 + halfThickness;
                }
                if (m_cnPt1.y > mPageCropOnClient.y2 - halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y2 - halfThickness;
                }
            }

            if (!mFirstTime)
            {
                mFirstTime = true;
                // set bounds
                mLeftmost = m_cnPt1.x;
                mRightmost = m_cnPt1.x;
                mTopmost = m_cnPt1.y;
                mBottommost = m_cnPt1.y;
            }

            mMustAddPath = true;
            if (mIsEraser)
            {
                mCurrentStrokeSet = null;
                SetUpEraserStroke(sender, e);
            }
            else
            {
                mEraserStrokeCollection = null;
                SetUpPenStroke(sender, e);
            }
           
            return true;
        }

        private void SetUpPenStroke(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            bool useSmoothing = !mIsStylus && !mIsEraser;
            StrokeAppearance appearance = new StrokeAppearance(mDownPageNumber, mOpacity, mStrokeThickness, mStrokeBrush, useSmoothing);
            if (mMustCreateNewStroke)
            {
                mMustCreateNewStroke = false;
              
                mCurrentStrokeSet = new StrokeCollection(appearance, mPDFView);
                mCurrentStrokeSet.mPageBoundaries = mPageCropOnClient;

                this.Children.Add(mCurrentStrokeSet.StrokeWidget);
            }

            PathFigureCollection newFigureCollection = mCurrentStrokeSet.StartNewPath();

            mCurrentPathFigure = new PathFigureBuilder(!mIsStylus && !mIsEraser);
            mCurrentPathFigure.AddPointToPath(new UIPoint(m_cnPt1.x - mPageCropOnClient.x1, m_cnPt1.y - mPageCropOnClient.y1));
            newFigureCollection.Add(mCurrentPathFigure.Figure);

            mCurrentStroke = mCurrentStrokeSet.CurrentPenStroke;
            StrokeMadeVisible?.Invoke(appearance, mCurrentStroke);
            mStartedPathSmoothing = false;
            mTwoPointsBack = new UIPoint(m_cnPt1.x, m_cnPt1.y);

            mLastPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);
            m_cnCurrentPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);

            mCurrentStroke.AddPoint(GetPagePointFromCanvasPosition(m_cnPt1.x, m_cnPt1.y, mDownPageNumber));
            CurrentStrokeModifed?.Invoke(mCurrentStroke);

            CheckIfShouldStartDrawing();
        }

        private void SetUpEraserStroke(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            mFullEraserStroke = new PenStroke();
            mPartialEraserStroke = new PenStroke();
            mLastErasedPoint = null;

            mCurrentPathFigure = new PathFigureBuilder(false);
            mCurrentPathFigure.AddPointToPath(new UIPoint(m_cnPt1.x - mPageCropOnClient.x1, m_cnPt1.y - mPageCropOnClient.y1));
            mCurrentPathFigure.AddPointToPath(new UIPoint(m_cnPt1.x - mPageCropOnClient.x1 + 0.5, m_cnPt1.y - mPageCropOnClient.y1 + 0.5));

            PDFPoint pagePoint = GetPagePointFromCanvasPosition(m_cnPt1.x, m_cnPt1.y, mDownPageNumber);
            mPartialEraserStroke.AddPoint(pagePoint);
            mFullEraserStroke.AddPoint(pagePoint);

            pagePoint = GetPagePointFromCanvasPosition(m_cnPt1.x + 0.5, m_cnPt1.y + 0.5, mDownPageNumber);
            mPartialEraserStroke.AddPoint(pagePoint);
            mFullEraserStroke.AddPoint(pagePoint);

            StrokeAppearance appearance = new StrokeAppearance(mDownPageNumber, mEraserOpacity, mEraserThickness, mEraserColor, false);
            mEraserStrokeCollection = new StrokeCollection(appearance, mPDFView);
            mEraserStrokeCollection.mPageBoundaries = mPageCropOnClient;
            PathFigureCollection collection = mEraserStrokeCollection.StartNewPath();
            collection.Add(mCurrentPathFigure.Figure);

            mStartDrawing = true;
            mEraserStartErasing = false;
            this.Children.Add(mEraserStrokeCollection.StrokeWidget);
            AddToolToCanvasIfNeeded();
            mLastPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);
        }

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (IsUsingSylus && e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Pen && !mIsEraser)
            {
                return true;
            }
            if (mPointerDown)
            {
                if (e.Pointer.PointerId != mPointerID)
                {
                    return true;
                }

                double samePointThreshold = 0.1;
                if (mIsEraser)
                {
                    samePointThreshold = 2.0;
                }
                // GetIntermediatePoints returns in REVERSE order.
                IList<PointerPoint> cnIntermediatePoints = e.GetIntermediatePoints(mViewerCanvas);
                for (int i = cnIntermediatePoints.Count - 1; i >= 0; i--)
                {
                    m_cnCurrentPoint = cnIntermediatePoints[i].Position;

                    // want to make sure we're within our bounding box.
                    m_cnPt2.x = m_cnCurrentPoint.X;
                    m_cnPt2.y = m_cnCurrentPoint.Y;
                    double halfThickness = mDrawThickness / 2;
                    if (mPageCropOnClient != null)
                    {
                        m_cnPt2.x = Math.Max(m_cnPt2.x, mPageCropOnClient.x1 + halfThickness);
                        m_cnPt2.x = Math.Min(m_cnPt2.x, mPageCropOnClient.x2 - halfThickness);
                        m_cnPt2.y = Math.Max(m_cnPt2.y, mPageCropOnClient.y1 + halfThickness);
                        m_cnPt2.y = Math.Min(m_cnPt2.y, mPageCropOnClient.y2 - halfThickness);
                    }

                    // Sometimes, we will get 2 point in the same, or at least a very similar position. Also, if we're out of bounds, this can happen.
                    if (Math.Abs(m_cnPt2.x - mLastPoint.X) < samePointThreshold && Math.Abs(m_cnPt2.y - mLastPoint.Y) < samePointThreshold)
                    {
                        continue;
                    }

                    if (mIsEraser)
                    {
                        mCurrentPathFigure.AddPointToPath(new UIPoint(m_cnPt2.x - mPageCropOnClient.x1, m_cnPt2.y - mPageCropOnClient.y1));
                        PDFPoint pagePoint = GetPagePointFromCanvasPosition(m_cnPt2.x, m_cnPt2.y, mDownPageNumber);
                        mPartialEraserStroke.AddPoint(pagePoint);
                        mFullEraserStroke.AddPoint(pagePoint);
                        mLastPoint = new UIPoint(m_cnPt2.x, m_cnPt2.y);
                    }
                    else if (mIsStylus)
                    {
                        AddPointToPath(new UIPoint(m_cnPt2.x, m_cnPt2.y), false);
                    }
                    else 
                    {
                        if (!mStartedPathSmoothing)
                        {
                            mStartedPathSmoothing = true;
                            mOnePointBack = new UIPoint(m_cnPt2.x, m_cnPt2.y);
                        }
                        else
                        {
                            UIPoint newPoint = UtilityFunctions.GetPointCloserToLine(mTwoPointsBack, new UIPoint(m_cnPt2.x, m_cnPt2.y), mOnePointBack);
                            mTwoPointsBack.X = mOnePointBack.X;
                            mTwoPointsBack.Y = mOnePointBack.Y;
                            mOnePointBack.X = m_cnPt2.x;
                            mOnePointBack.Y = m_cnPt2.y;

                            AddPointToPath(newPoint);
                        }
                    }
                }
                if (!mIsEraser)
                {
                    CurrentStrokeModifed?.Invoke(mCurrentStroke);

                    if (!mStartDrawing) // We don't want to start drawing until we have a bit of a distance
                    {
                        CheckIfShouldStartDrawing();
                    }
                }
            }
            return false;
        }

        protected void CheckIfShouldStartDrawing()
        {
            // wait until you're at least START_DRAWING_THRESHOLD distance from the start
            // unless we're in multi stroke mode
            if (mMultiStrokeMode
                || (Math.Abs(m_cnPt2.x - m_cnPt1.x) > START_DRAWING_THRESHOLD)
                || (Math.Abs(m_cnPt2.y - m_cnPt1.y) > START_DRAWING_THRESHOLD))
            {
                mStartDrawing = true;
                AddToolToCanvasIfNeeded();
            }
        }

        protected void AddToolToCanvasIfNeeded()
        {
            if (!mShapeHasBeenCreated) // path is only created once. Subsequent draws add to the PathfigureCollection
            {
                mShapeHasBeenCreated = true;

                mViewerCanvas.Children.Add(this);
                this.SetValue(Canvas.LeftProperty, mPageCropOnClient.x1);
                this.SetValue(Canvas.TopProperty, mPageCropOnClient.y1);
                this.Width = Math.Abs(mPageCropOnClient.Width());
                this.Height = Math.Abs(mPageCropOnClient.Height());
                // lower opacity when using marker tool to allow text to show through 
                if (IsMarker)
                {
                    this.Opacity = 0.65; // the selected opacity will be added to the stroke itself.
                }
            }
        }

        protected void AddPointToPath(UIPoint cnPoint, bool smooth = true)
        {
            // add to points going to PDF
            mCurrentStroke.AddPoint(GetPagePointFromCanvasPosition(cnPoint.X, cnPoint.Y, mDownPageNumber));

            mCurrentPathFigure.AddPointToPath(new UIPoint(cnPoint.X - mPageCropOnClient.x1, cnPoint.Y - mPageCropOnClient.y1));

            mLastPoint = new UIPoint(cnPoint.X, cnPoint.Y);

            // Update bounds
            if (cnPoint.X < mLeftmost)
            {
                mLeftmost = cnPoint.X;
            }
            if (cnPoint.X > mRightmost)
            {
                mRightmost = cnPoint.X;
            }
            if (cnPoint.Y < mTopmost)
            {
                mTopmost = cnPoint.Y;
            }
            if (cnPoint.Y > mBottommost)
            {
                mBottommost = cnPoint.Y;
            }
        }

        internal PDFPoint GetPagePointFromCanvasPosition(double canvasX, double canvasY, int pageNumber)
        {
            PDFDouble xPos = new PDFDouble(canvasX);
            PDFDouble yPos = new PDFDouble(canvasY);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(xPos, yPos);
            mPDFView.ConvScreenPtToPagePt(xPos, yPos, pageNumber);
            return new PDFPoint(xPos.Value, yPos.Value);
        }


        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            RemovePointer(e.Pointer);
            if (IsUsingSylus && e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Pen && !IsEraser)
            {
                return true;
            }
            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                return false; // ignore any secondary pointer
            }
            if (e.Pointer.PointerId == mPointerID)
            {
                if (mIsEraser)
                {
                    ReleaseEraserPointer(sender, e);
                }
                else
                {
                    ReleaseInkPointer(sender, e);
                }

                mCurrentPathFigure = null;
                mCurrentStroke = null;
                mPointerID = -1;
                mViewerCanvas.ReleasePointerCaptures();
                if (mIsInTimedMode)
                {
                    StartTimer();
                }
            }
            return false;
        }

        protected void ReleaseInkPointer(object sender, PointerRoutedEventArgs e)
        {
            mPointerDown = false;

            m_cnPt2.x = m_cnCurrentPoint.X;
            m_cnPt2.y = m_cnCurrentPoint.Y;
            double halfThickness = mDrawThickness / 2;
            if (mPageCropOnClient != null)
            {
                m_cnPt2.x = Math.Max(m_cnPt2.x, mPageCropOnClient.x1 + halfThickness);
                m_cnPt2.x = Math.Min(m_cnPt2.x, mPageCropOnClient.x2 - halfThickness);
                m_cnPt2.y = Math.Max(m_cnPt2.y, mPageCropOnClient.y1 + halfThickness);
                m_cnPt2.y = Math.Min(m_cnPt2.y, mPageCropOnClient.y2 - halfThickness);
            }
            AddPointToPath(new UIPoint(m_cnPt2.x, m_cnPt2.y));
            if (!mIsStylus)
            {
                mCurrentPathFigure.ClosePath();
            }
            CurrentStrokeModifed?.Invoke(mCurrentStroke);

            if (!mAllStrokes.Contains(mCurrentStrokeSet))
            {
                mAllStrokes.Add(mCurrentStrokeSet);
            }

            if (!mMultiStrokeMode)
            {
                if (mStartDrawing) // don't want to add any annotations resulting from a tap
                {
                    EnableScrolling();
                    Finish();

                    mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
                    mDelayedRemovalSet = true;

                    EndCurrentTool(ToolType.e_pan);
                }
            }
            else
            {
                mUndoSnapshots.Add(new UndoRedoSnapshot(mAllStrokes));
                ClearAllUndoneStrokes();
                if (StrokeAdded != null)
                {
                    StrokeAdded();
                }
            }
        }

        protected void ReleaseEraserPointer(object sender, PointerRoutedEventArgs e)
        {
            if (!mEraserStartErasing)
            {
                if (ApplyEraser(mPartialEraserStroke.StrokePoints))
                {
                    mUndoSnapshots.Add(new UndoRedoSnapshot(mAllStrokes));
                    ClearAllUndoneStrokes();
                    StrokeAdded?.Invoke();
                }
            }

            if (mEraserStrokeCollection != null && this.Children.Contains(mEraserStrokeCollection.StrokeWidget))
            {
                this.Children.Remove(mEraserStrokeCollection.StrokeWidget);
            }
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (!mMultiStrokeMode)
            {
                base.TappedHandler(sender, e);
            }
            return true;
        }


        internal override void Finish()
        {
            if (!mShouldSaveInk)
            {
                return;
            }
            try
            {
                mPDFView.DocLock(true);
                double halfThickness = mDrawThickness / 2;
                m_cnPt1.x = mLeftmost - halfThickness;
                m_cnPt1.y = mTopmost - halfThickness;
                m_cnPt2.x = mRightmost + halfThickness;
                m_cnPt2.y = mBottommost + halfThickness;
                Rect rect = GetShapeBBox();
                if (rect != null)
                {
                    List<Tuple<IAnnot, int>> addedAnnots = new List<Tuple<IAnnot, int>>();

                    pdftron.PDF.Annots.Ink ink = null;
                    StrokeAppearance previousInkAppearance = null;
                    int i = 0;

                    // Go through each StrokeCollection (based on ink color)
                    foreach (StrokeCollection strokeCollection in mAllStrokes)
                    {
                        bool createInk = ink == null;
                        if (ink != null)
                        {
                            createInk = !previousInkAppearance.IsEqual(strokeCollection.mStrokeAppearance);
                        }
                        if (createInk)
                        {
                            if (ink != null)
                            {
                                ink.RefreshAppearance();

                                UtilityFunctions.SetUniqueId(ink, Guid.NewGuid().ToString());

                                pdftron.PDF.Page page = mPDFView.GetDoc().GetPage(strokeCollection.mStrokeAppearance.PageNumber);
                                page.AnnotPushBack(ink);
                                mPDFView.UpdateWithAnnot(ink, mAnnotPageNum);
                                mToolManager.RaiseAnnotationAddedEvent(ink, mAnnotPageNum);
                                
                                mAnnotPageNum = strokeCollection.mStrokeAppearance.PageNumber;
                                mAnnot = ink;
                                addedAnnots.Add(new Tuple<IAnnot, int>(mAnnot, strokeCollection.mStrokeAppearance.PageNumber));
                            }

                            ink = pdftron.PDF.Annots.Ink.Create(mPDFView.GetDoc().GetSDFDoc(), rect);
                            previousInkAppearance = strokeCollection.mStrokeAppearance;
                            mOpacity = strokeCollection.mStrokeAppearance.Opacity;
                            mStrokeThickness = strokeCollection.mStrokeAppearance.Thickness;
                            mStrokeBrush = strokeCollection.mStrokeAppearance.Brush;
                            SetStyle(ink);
                            i = 0;
                        }

                        foreach (StrokeCollection.StrokeBundle strokeBundle in strokeCollection.mStrokeList)
                        {
                            int j = 0;
                            foreach (Point p in strokeBundle.mPenStroke.StrokePoints)
                            {
                                ink.SetPoint(i, j, p);
                                j++;
                            }
                            i++;
                        }

                        if (mToolManager.InkSmoothingBehaviour == ToolManager.InkSmoothingOptions.NoSmoothing ||
                            (mToolManager.InkSmoothingBehaviour == ToolManager.InkSmoothingOptions.AllButStylus && mIsStylus))
                        {
                            ink.SetSmoothing(false);
                        }

                        if (mIsMarker)
                        {
                            ink.SetHighlightIntent(true);
                        }

                        mAnnotPageNum = strokeCollection.mStrokeAppearance.PageNumber;
                    }
                    if (ink != null)
                    {
                        ink.RefreshAppearance();

                        UtilityFunctions.SetUniqueId(ink, Guid.NewGuid().ToString());

                        pdftron.PDF.Page page = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                        page.AnnotPushBack(ink);
                        mPDFView.UpdateWithAnnot(ink, mAnnotPageNum);
                        mToolManager.RaiseAnnotationAddedEvent(ink, mAnnotPageNum);

                        
                        mAnnot = ink;
                        addedAnnots.Add(new Tuple<IAnnot, int>(mAnnot, mAnnotPageNum));
                    }
                    //foreach (Tuple<IAnnot, int> annot in addedAnnots)
                    //{
                        // TODO: this is being called for every annot after all annot have already being added
                        // so it does not work if UndoRedoManager is listening. Need to change it in a way we can take a snapshot
                        // for every annotation

                        //mPDFView.UpdateWithAnnot(annot.Item1, annot.Item2);
                        // mToolManager.RaiseAnnotationAddedEvent(annot.Item1, annot.Item2);
                    //}
                }                
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mShouldSaveInk = false;
                mPDFView.DocUnlock();
            }
        }

        private void ClearAllUndoneStrokes()
        {
            mUndoneSnapshots.Clear();
        }

        #region MultiStrokeToolInterface

        /// <summary>
        /// Saves the ink annotation to the PDF
        /// </summary>
        public void CommitAnnotation()
        {
            if (mStartDrawing) // don't want to add any annotations resulting from a tap
            {
                EnableScrolling();
                Finish();
            }
            mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
            mDelayedRemovalSet = true;
            EndCurrentTool(ToolType.e_pan);
        }

        /// <summary>
        /// Undo the last stroke
        /// </summary>
        public void UndoStroke()
        {
            try
            {
                if (CanUndoStroke())
                {
                    AllStrokesMadeInvisible?.Invoke();
                    this.Children.Clear();
                    UndoRedoSnapshot snapshot = mUndoSnapshots[mUndoSnapshots.Count - 1];
                    mUndoneSnapshots.Add(snapshot);
                    mUndoSnapshots.RemoveAt(mUndoSnapshots.Count - 1);
                    mAllStrokes = mUndoSnapshots[mUndoSnapshots.Count - 1].RestoreStrokes(this, mPDFView);
                    foreach (StrokeCollection penCollection in mAllStrokes)
                    {
                        StrokeAppearance appearance = new StrokeAppearance(penCollection.mStrokeAppearance);
                        foreach (StrokeCollection.StrokeBundle stroke in penCollection.mStrokeList)
                        {
                            StrokeMadeVisible?.Invoke(appearance, stroke.mPenStroke);
                        }
                    }
                    mMustCreateNewStroke = true;
                    mMustAddPath = true;
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        /// <summary>
        /// Redo the last undone stroke
        /// </summary>
        public void RedoStroke()
        {
            try
            {
                if (CanRedoStroke())
                {
                    AllStrokesMadeInvisible?.Invoke();
                    this.Children.Clear();
                    UndoRedoSnapshot snapshot = mUndoneSnapshots[mUndoneSnapshots.Count - 1];
                    mUndoneSnapshots.RemoveAt(mUndoneSnapshots.Count - 1);
                    mUndoSnapshots.Add(snapshot);
                    mAllStrokes = snapshot.RestoreStrokes(this, mPDFView);
                    foreach (StrokeCollection penCollection in mAllStrokes)
                    {
                        StrokeAppearance appearance = new StrokeAppearance(penCollection.mStrokeAppearance);
                        foreach (StrokeCollection.StrokeBundle stroke in penCollection.mStrokeList)
                        {
                            StrokeMadeVisible?.Invoke(appearance, stroke.mPenStroke);
                        }
                    }
                    mMustCreateNewStroke = true;
                    mMustAddPath = true;
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        /// <summary>
        /// Returns true if it is possible to undo a stoke
        /// </summary>
        public bool CanUndoStroke()
        {
            return mUndoSnapshots.Count > 1;
        }

        /// <summary>
        /// Returns true if it is possible to redo a stroke
        /// </summary>
        public bool CanRedoStroke()
        {
            return mUndoneSnapshots.Count > 0;
        }

        /// <summary>
        /// Lets you change the brush properties while between strokes, though it will change past strokes as well.
        /// </summary>
        /// <param name="thickness"></param>
        /// <param name="opacity"></param>
        /// <param name="color"></param>
        public void SetInkProperties(double thickness, double opacity, Windows.UI.Color color)
        {
            mRemainInEraserMode = false;
            mSimpleShapeCreateSubclassHandlesProperties = true;
            mIsEraser = false;
            mStrokeThickness = thickness;
            mZoomLevel = mPDFView.GetZoom();
            mDrawThickness = mStrokeThickness * mZoomLevel;
            mOpacity = opacity;
            mStrokeBrush = new SolidColorBrush(color);

            mMustCreateNewStroke = true;
            mMustAddPath = true;
        }



        public void SetEraserProperties(bool eraseOnlyCurrentInk, double thickness, double opacity, Windows.UI.Color color)
        {
            mRemainInEraserMode = true;
            SetEraserPropertiesInternal(eraseOnlyCurrentInk, thickness, opacity, color);
        }

        protected void SetEraserPropertiesInternal(bool eraseOnlyCurrentInk, double thickness, double opacity, Windows.UI.Color color)
        {
            mSimpleShapeCreateSubclassHandlesProperties = true;
            mEraserThickness = thickness;
            mZoomLevel = mPDFView.GetZoom();
            mDrawThickness = mEraserThickness * mZoomLevel;
            mEraserOpacity = opacity;
            mEraserColor = new SolidColorBrush(color);

            mEraseThisInkOnly = eraseOnlyCurrentInk;
        }

        protected bool ApplyEraser(IReadOnlyCollection<PDFPoint> eraserPagePoints)
        {
            bool somethingErased = false;
            List<PDFPoint> eraserPagePointList = new List<Point>(); ;

            if (mLastErasedPoint != null)
            {
                eraserPagePointList.Add(mLastErasedPoint);
            }
            foreach (Point pt in eraserPagePoints)
            {
                eraserPagePointList.Add(pt);
            }

            if (eraserPagePoints.Count > 0)
            {
                mLastErasedPoint = eraserPagePointList[eraserPagePoints.Count - 1];
            }

            if (eraserPagePoints.Count > 1)
            {
                List<StrokeCollection> erasedStrokeGroups = new List<StrokeCollection>();
                List<StrokeCollection> addedStrokeGroups = new List<StrokeCollection>();
                foreach (StrokeCollection strokeGroup in mAllStrokes)
                {
                    List<PenStroke> addedStrokes = new List<PenStroke>();
                    List<PenStroke> removedStrokes = new List<PenStroke>();
                    somethingErased |= strokeGroup.Erase(eraserPagePointList, strokeGroup.GetBoundingRectInPageSpace(), mEraserThickness / 2.0, ref addedStrokes, ref removedStrokes);
                    if (somethingErased)
                    {
                        StrokeAppearance appearance = new StrokeAppearance(strokeGroup.mStrokeAppearance);
                        foreach (PenStroke stroke in removedStrokes)
                        {
                            StrokeMadeInvisible?.Invoke(stroke);
                        }
                        foreach (PenStroke stroke in addedStrokes)
                        {
                            StrokeMadeVisible?.Invoke(appearance, stroke);
                        }
                    }
                }
                foreach (StrokeCollection erasedStroke in erasedStrokeGroups)
                {
                    mAllStrokes.Remove(erasedStroke);
                }
                foreach (StrokeCollection addedStroke in addedStrokeGroups)
                {
                    mAllStrokes.Add(addedStroke);
                }
            }
            return somethingErased;
        }

        protected void ApplyEraserToAllAnnots(List<PDFPoint> eraserPagePoints)
        {

        }

        #endregion MultiStrokeToolInterface



        #region Timer specific utilities

        private void StartTimer()
        {
            if (mCompletionTimer == null && mToolManager.INK_TIME_BEFORE_INK_SAVES_ANNOTATION_IN_MILLISECONDS > 0)
            {
                mCompletionTimer = new DispatcherTimer();
                mCompletionTimer.Interval = TimeSpan.FromMilliseconds(mToolManager.INK_TIME_BEFORE_INK_SAVES_ANNOTATION_IN_MILLISECONDS);
                mCompletionTimer.Tick += mCompletionTimer_Tick;
            }
            mCompletionTimer.Stop();
            mCompletionTimer.Start();
        }

        private void StopTimer()
        {
            if (mCompletionTimer != null)
            {
                mCompletionTimer.Stop();
            }
        }

        private void EndTimer()
        {
            if (mCompletionTimer != null)
            {
                mCompletionTimer.Stop();
                mCompletionTimer = null;

                if (mShouldSaveInk)
                {
                    Finish();
                }

                mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
                mDelayedRemovalSet = true;
            }
        }

        void mCompletionTimer_Tick(object sender, object e)
        {
            mToolManager.CreateTool(ToolType.e_pan, null);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendLine(string.Format("FreeHandCreate with {0} separate inks", mAllStrokes.Count));

            int i = 0;
            foreach (StrokeCollection collection in mAllStrokes)
            {
                sb.AppendLine(string.Format("Stroke {0}:", i++));
                sb.AppendLine(collection.ToString());
            }

            return sb.ToString();
        }

        #endregion Timer specific utilities

    }
}
